/*======================================================================*\

Copyright (c) 1990-2002  Paul Vojta and others

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
PAUL VOJTA OR ANY OTHER AUTHOR OF THIS SOFTWARE BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

NOTE:
	xdvi is based on prior work, as noted in the modification history
	in xdvi.c.

\*========================================================================*/

#include "xdvi-config.h"

#ifdef STDC_HEADERS
# include <unistd.h>
# include <fcntl.h>
#endif
#include <signal.h>
#include <sys/file.h>	/* this defines FASYNC */
#ifdef HAVE_SYS_FCNTL_H
# include <sys/fcntl.h>  /* Or this might define FASYNC */
#endif
#include <sys/ioctl.h>	/* this defines SIOCSPGRP and FIOASYNC */
#include <sys/wait.h>	/* ignore HAVE_SYS_WAIT_H -- we always need WNOHANG */

#include "xdvi.h" /* this includes Xlib and Xutil are already included */
#include "xdvi-debug.h"


#ifndef MOTIF
#include <X11/Xaw/Form.h> /* for XtNresizable */
#endif

#include <setjmp.h>

#include <X11/Xatom.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>	/* needed for def. of XtNiconX(??) */

#include "pagesel.h"
#include "special.h"
#include "psgs.h"


#include <errno.h>

#include <ctype.h>

#include "util.h"
#include "x_util.h"
#include "string-utils.h"
#include "print-dialog.h"
#include "search-dialog.h"
#include "sfSelFile.h"
#include "mag.h"
#include "help-window.h"
#include "message-window.h"
#include "dvi-draw.h"
#include "statusline.h"
#include "hypertex.h"
#include "dvi-init.h"
#include "Tip.h"
#include "browser.h"
#include "search-internal.h"

#include "events.h"
#include "xm_toolbar.h"
#include "xaw_menu.h"
#include "xm_menu.h"

#ifdef	X_NOT_STDC_ENV
extern int errno;
#endif /* X_NOT_STDC_ENV */

/* Linux prefers O_ASYNC over FASYNC; SGI IRIX does the opposite.  */
#if !defined(FASYNC) && defined(O_ASYNC)
# define FASYNC	O_ASYNC
#endif

#if !defined(FLAKY_SIGPOLL) && !HAVE_STREAMS && !defined(FASYNC)
# if !defined(SIOCSPGRP) || !defined(FIOASYNC)
#  define FLAKY_SIGPOLL	1
# endif
#endif

#ifndef FLAKY_SIGPOLL

# ifndef SIGPOLL
#  define SIGPOLL	SIGIO
# endif

# ifndef SA_RESTART
#  define SA_RESTART 0
# endif

# if HAVE_STREAMS
#  include <stropts.h>

#  ifndef S_RDNORM
#   define S_RDNORM S_INPUT
#  endif

#  ifndef S_RDBAND
#   define S_RDBAND  0
#  endif

#  ifndef S_HANGUP
#   define S_HANGUP  0
#  endif

#  ifndef S_WRNORM
#   define S_WRNORM  S_OUTPUT
#  endif
# endif /* HAVE_STREAMS */

#endif /* not FLAKY_SIGPOLL */

#if HAVE_SIGACTION && !defined SA_RESETHAND
# ifdef SA_ONESHOT
#  define SA_RESETHAND SA_ONESHOT
# else
#  undef HAVE_SIGACTION         /* Needed for Mac OS X < 10.2 (9/2002) */
# endif
#endif

#if HAVE_POLL
# include <poll.h>
#else
# if HAVE_SYS_SELECT_H
#  include <sys/select.h>
# else
#  if HAVE_SELECT_H
#   include <select.h>
#  endif
# endif
# define XIO_IN 1
# define XIO_OUT 2
#endif /* HAVE_POLL */

/* Used to convert real pagenumbers to C (0-based) numbers,
   or as offset when user sets pagenumber via Act_declare_page_number */
int pageno_correct = 1;

/*
 * Mechanism for reducing repeated warning about specials, lost characters,
 * etc.
 */
Boolean	warn_spec_now;

static sigset_t all_signals;

/* globals for event handling */
unsigned int ev_flags = EV_IDLE;
VOLATILE int event_counter = 0;
jmp_buf canit_env;


/* globals for `pause' feature */
int pause_num = 0;
int *pause_num_save = NULL;
Boolean pause_flag = False;

/* even more globals, probably something to do with window geometry ... */
int min_x;
int max_x;
int min_y;
int max_y;

static void xdvi_normal_exit(void);

/* to remember the scrollbar positions so that we can restore them
   when `keep position' is active and window is resized to full size
   (removing scrollbars) and then back (#810501) */
static int m_x_scroll = 0, m_y_scroll = 0;

/* for communication with forward search in dvi-draw.c */
int source_fwd_box_page	= -1;	/* -1 means no box */
_Xconst char *source_forward_string = NULL;

static int source_reverse_x, source_reverse_y;
static int source_show_all;

/* globals for color stuff */
Pixel plane_masks[4];
XColor color_data[2];
struct pagecolor_info page_colors = { 0, NULL };

_Xconst struct rgb *color_bottom;
unsigned int color_bot_size;		/* number of entries */
struct colorframe *rcs_top;
struct rgb fg_initial;			/* Initial fg (from command line) */
struct rgb bg_initial;			/* Initial bg */
struct bgrec *bg_head = NULL;		/* head of list */
struct bgrec *bg_current = NULL;	/* current bg value */
struct fgrec *fg_current;		/* current fg value */
struct fgrec *fg_active = NULL;		/* where the GCs are */
Pixel *color_list;			/* list of colors */
unsigned int color_list_len = 0;	/* current len of list*/
unsigned int color_list_max = 0;	/* allocated size */
Boolean color_warned = False;
/*
  FIXME: the following file-static flag is used to control whether
  a mouse-related Act_* should block further actions that have
  been registered for the same event.
  
  It is used for the Act_href_* routines.
  
  This used to be a
     longjmp(somewehre_in_the_event_loop);
  but that was pretty dreadful as well. Surely there must
  be some other (X(t)-specific?) way to do this?? (Apart from
  de-registering the action and re-registering it again.)
*/
static Boolean block_mouse_event_flag = False; 

/* for calling it from mag.c */
void
block_next_mouse_event(void) {
    block_mouse_event_flag = True;
}

Boolean
block_this_mouse_event(void) {
    if (block_mouse_event_flag == True) {
	block_mouse_event_flag = False;
	return True;
    }
    return False;
}

/*
 *	The cursor shape in the magnifying glass is determined by which
 *	window received the button press event.  Under XAW, that is a parent
 *	of mane.win, so we have to define the cursor of that window instead
 *	of mane.win in that case.
 */


#ifdef MOTIF
# define CURSORWIN	mane.win
#else
# define CURSORWIN	mane.win
/*
  SU: FIXME: this is broken. With the uncommented stuff, cursor will never
  get the drag shape; however even without it, the cursor drops back to the
  magnifier during the dragging operation, which is not wanted.
 */
/* # if BUTTONS */
/* #  define CURSORWIN	XtWindow(form_widget) */
/* # else */
/* #  define CURSORWIN	XtWindow(vport_widget) */
/* # endif */
#endif



void null_mouse(XEvent *event)
{
    UNUSED(event);
}

mouse_proc mouse_motion = null_mouse;
mouse_proc mouse_release = null_mouse;

static void Act_digit(Widget, XEvent *, String *, Cardinal *);
static void Act_find(Widget, XEvent *, String *, Cardinal *);
static void Act_minus(Widget, XEvent *, String *, Cardinal *);
static void Act_quit(Widget, XEvent *, String *, Cardinal *);
static void Act_print(Widget, XEvent *, String *, Cardinal *);
static void Act_save(Widget, XEvent *, String *, Cardinal *);
static void Act_help(Widget, XEvent *, String *, Cardinal *);
static void Act_close_help(Widget, XEvent *, String *, Cardinal *);
static void Act_goto_page(Widget, XEvent *, String *, Cardinal *);
static void Act_declare_page_number(Widget, XEvent *, String *, Cardinal *);
static void Act_toggle_mark(Widget, XEvent *, String *, Cardinal *);
static void Act_home(Widget, XEvent *, String *, Cardinal *);
static void Act_center(Widget, XEvent *, String *, Cardinal *);
static void Act_left(Widget, XEvent *, String *, Cardinal *);
static void Act_right(Widget, XEvent *, String *, Cardinal *);
static void Act_up(Widget, XEvent *, String *, Cardinal *);
static void Act_down(Widget, XEvent *, String *, Cardinal *);
static void Act_up_or_previous(Widget, XEvent *, String *, Cardinal *);
static void Act_down_or_next(Widget, XEvent *, String *, Cardinal *);
static void Act_set_margins(Widget, XEvent *, String *, Cardinal *);
static void Act_show_display_attributes(Widget, XEvent *, String *, Cardinal *);
static void Act_set_density(Widget, XEvent *, String *, Cardinal *);
static void Act_change_density(Widget, XEvent *, String *, Cardinal *);
static void Act_fullscreen(Widget, XEvent *, String *, Cardinal *);
#ifdef GREY
static void Act_set_greyscaling(Widget, XEvent *, String *, Cardinal *);
#endif
#if COLOR
static void Act_set_color(Widget, XEvent *, String *, Cardinal *);
#endif

static void Act_htex_anchorinfo(Widget, XEvent *, String *, Cardinal *);

static void Act_reread_dvi_file(Widget, XEvent *, String *, Cardinal *);
#ifdef SELFILE
static void Act_select_dvi_file(Widget, XEvent *, String *, Cardinal *);
#endif
static void Act_discard_number(Widget, XEvent *, String *, Cardinal *);
static void Act_drag(Widget, XEvent *, String *, Cardinal *);
static void Act_wheel(Widget, XEvent *, String *, Cardinal *);
static void Act_motion(Widget, XEvent *, String *, Cardinal *);
static void Act_release(Widget, XEvent *, String *, Cardinal *);
static void Act_source_special(Widget, XEvent *, String *, Cardinal *);
static void Act_show_source_specials(Widget, XEvent *, String *, Cardinal *);
static void Act_source_what_special(Widget, XEvent *, String *, Cardinal *);
static void Act_unpause_or_next(Widget, XEvent *, String *, Cardinal *);
static void Act_ruler_snap_origin(Widget w, XEvent *event, String *params, Cardinal *num_params);
static void Act_load_url(Widget w, XEvent *event, String *params, Cardinal *num_params);

XtActionsRec actions[] = {
    {"digit", Act_digit},
    {"ruler-mode", Act_ruler_mode},
    {"minus", Act_minus},
    {"quit", Act_quit},
    {"print", Act_print},
    {"save", Act_save},
    {"find", Act_find},
    {"help", Act_help},
    {"close-help", Act_close_help},
    {"goto-page", Act_goto_page},
    {"use-tex-pages", Act_use_tex_pages},
    {"forward-page", Act_forward_page},
    {"back-page", Act_back_page},
    {"toggle-mark", Act_toggle_mark},
    {"declare-page-number", Act_declare_page_number},
    {"home", Act_home},
    {"center", Act_center},
    {"set-keep-flag", Act_set_keep_flag},
    {"ruler-snap-origin", Act_ruler_snap_origin},
    {"left", Act_left},
    {"right", Act_right},
    {"up", Act_up},
    {"down", Act_down},
    {"up-or-previous", Act_up_or_previous},
    {"down-or-next", Act_down_or_next},
    {"set-margins", Act_set_margins},
    {"show-display-attributes", Act_show_display_attributes},
    {"set-shrink-factor", Act_set_shrink_factor},
    {"shrink-to-dpi", Act_shrink_to_dpi},
    {"set-density", Act_set_density},
    {"change-density", Act_change_density},
    {"fullscreen", Act_fullscreen},
#ifdef GREY
    {"set-greyscaling", Act_set_greyscaling},
#endif
#if COLOR
    {"set-color", Act_set_color},
#endif
#ifdef PS
    {"set-ps", Act_set_ps},
#endif
#ifdef HTEX
    {"htex-back", Act_htex_back},
    {"htex-forward", Act_htex_forward},
    {"htex-anchorinfo", Act_htex_anchorinfo},
#endif
#ifdef PS_GS
    {"set-gs-alpha", Act_set_gs_alpha},
#endif
    {"set-expert-mode", Act_set_expert_mode},
    {"reread-dvi-file", Act_reread_dvi_file},
#ifdef SELFILE
    {"select-dvi-file", Act_select_dvi_file},
#endif /* SELFILE */
    {"discard-number", Act_discard_number},
    {"drag", Act_drag},
    {"wheel", Act_wheel},
    {"motion", Act_motion},
    {"release", Act_release},
    {"source-special", Act_source_special},
    {"show-source-specials", Act_show_source_specials},
    {"source-what-special", Act_source_what_special},
    {"unpause-or-next", Act_unpause_or_next},
#if defined(NEW_MENU_CREATION) || defined(MOTIF)
    {"set-papersize", Act_set_papersize},
    {"set-paper-landscape", Act_set_paper_landscape},
#endif /* NEW_MENU_CREATION */
    {"load-url", Act_load_url},
};

Cardinal num_actions = XtNumber(actions);

/*
 *	Data for buffered events.
 */

#ifndef FLAKY_SIGPOLL
static VOLATILE int event_freq = 70;
#else
#define	event_freq	70
#endif

static void can_exposures(struct WindowRec *windowrec);

/*
 * Event-handling routines.
 */
void
expose(struct WindowRec *windowrec,
       int x, int y,
       unsigned int w, unsigned int h)
{
    if (windowrec->min_x > x)
	windowrec->min_x = x;
    if (windowrec->max_x < (int)(x + w))
	windowrec->max_x = x + w;
    if (windowrec->min_y > y)
	windowrec->min_y = y;
    if (windowrec->max_y < (int)(y + h))
	windowrec->max_y = y + h;

    ev_flags |= EV_EXPOSE;
}

/* non-static for href */
void
clearexpose(struct WindowRec *windowrec,
	    int x, int y,
	    unsigned w, unsigned h)
{
    XClearArea(DISP, windowrec->win, x, y, w, h, False);
    expose(windowrec, x, y, w, h);
}

/*
 *	Routines for X11 toolkit.
 */

static Arg arg_wh[] = {
    {XtNwidth, (XtArgVal) &window_w},
    {XtNheight, (XtArgVal) &window_h},
};

static Position window_x, window_y;
static Arg arg_xy[] = {
    {XtNx, (XtArgVal) &window_x},
    {XtNy, (XtArgVal) &window_y},
};

#define	get_xy() XtGetValues(draw_widget, arg_xy, XtNumber(arg_xy))


struct xdvi_action *
compile_action(_Xconst char *str)
{
    _Xconst char *p, *p1, *p2;
    XtActionsRec *actp;
    struct xdvi_action *ap;

    while (*str == ' ' || *str == '\t')
	++str;

    if (*str == '\0' || *str == '\n')
	return NULL;

    p = str;

    /* find end of command name */
    while (isalnum(*p) || *p == '-' || *p == '_')
	++p;

    for (actp = actions;; ++actp) {
	if (actp >= actions + num_actions) {
	    _Xconst char *tmp = strchr(str, '\0');
	    if (tmp == NULL) {
		tmp = p;
	    }
	    XDVI_WARNING((stderr, "Cannot compile action \"%.*s\".", tmp - str, str));

	    return NULL;
	}
	if (memcmp(str, actp->string, p - str) == 0 && actp->string[p - str] == '\0')
	    break;
    }

    while (*p == ' ' || *p == '\t')
	++p;
    if (*p != '(') {
	while (*p != '\0' && *p != '\n')
	    ++p;
	XDVI_WARNING((stderr, "Syntax error in action %.*s.", p - str, str));

	return NULL;
    }
    ++p;
    while (*p == ' ' || *p == '\t')
	++p;
    for (p1 = p;; ++p1) {
	if (*p1 == '\0' || *p1 == '\n') {
	    XDVI_WARNING((stderr, "Syntax error in action %.*s.", p1 - str, str));
	    return NULL;
	}
	if (*p1 == ')')
	    break;
    }

    ap = xmalloc(sizeof *ap);
    ap->proc = actp->proc;
    for (p2 = p1;; --p2) {
	if (p2 <= p) {	/* if no args */
	    ap->num_params = 0;
	    ap->param = NULL;
	    break;
	}
	else if (p2[-1] != ' ' && p2[-1] != '\t') {
	    char *arg;

	    arg = xmalloc(p2 - p + 1);
	    bcopy(p, arg, p2 - p);
	    arg[p2 - p] = '\0';
	    ap->num_params = 1;
	    ap->param = arg;
	    break;
	}
    }
    
    ap->next = compile_action(p1 + 1);

    return ap;
}

void
handle_command(Widget widget, XtPointer client_data, XtPointer call_data)
{
    struct xdvi_action *actp;

    UNUSED(call_data);

    /* call all actions registered for this event */
    for (actp = (struct xdvi_action *)client_data; actp != NULL; actp = actp->next) {
	if (debug & DBG_EVENT)
	    fprintf(stderr, "calling action with param: %s\n", actp->param);
	(actp->proc) (widget, NULL, &actp->param, &actp->num_params);
    }
}

#ifdef MOTIF
int
set_bar_value(Widget bar, int value, int max)
{
    XmScrollBarCallbackStruct call_data;

#ifdef TEST_SCROLLING
    fprintf(stderr, "set_bar_value: val %d, max %d\n", value, max);
#endif
    if (value > max)
	value = max;
    if (value < 0)
	value = 0;
    call_data.value = value;
    XtVaSetValues(bar, XmNvalue, value, NULL);
    XtCallCallbacks(bar, XmNvalueChangedCallback, &call_data);
    return value;
}
#endif

void
home(wide_bool scrl)
{
    if (!scrl)
	XUnmapWindow(DISP, mane.win);
# ifdef MOTIF
    {
	int value;

	value = (page_w - mane.width) / 2;
	if (value > home_x / mane.shrinkfactor)
	    value = home_x / mane.shrinkfactor;
	(void)set_bar_value(x_bar, value, (int)(page_w - mane.width));

	value = (page_h - mane.height) / 2;
	if (value > home_y / mane.shrinkfactor)
	    value = home_y / mane.shrinkfactor;
	(void)set_bar_value(y_bar, value, (int)(page_h - mane.height));
    }
# else
    get_xy();
    if (x_bar != NULL) {
	int coord = (page_w - mane.width) / 2;

	if (coord > home_x / mane.shrinkfactor)
	    coord = home_x / mane.shrinkfactor;
	XtCallCallbacks(x_bar, XtNscrollProc, (XtPointer) (window_x + coord));
    }
    if (y_bar != NULL) {
	int coord = (page_h - mane.height) / 2;

	if (coord > home_y / mane.shrinkfactor)
	    coord = home_y / mane.shrinkfactor;
	XtCallCallbacks(y_bar, XtNscrollProc, (XtPointer) (window_y + coord));
    }
# endif /* MOTIF */
    if (!scrl) {
	XMapWindow(DISP, mane.win);
	/* Wait for the server to catch up---this eliminates flicker. */
	XSync(DISP, False);
    }
    handle_x_scroll(NULL, NULL, NULL, NULL);
    handle_y_scroll(NULL, NULL, NULL, NULL);
}

/*
 *	Same as home(), except move to the bottom of the page.
 */

static void
home_bottom(wide_bool scrl)
{
    UNUSED(scrl);
    XUnmapWindow(DISP, mane.win);
#ifdef MOTIF
    {
	int value;

	value = (page_w - mane.width) / 2;
	if (value > home_x / mane.shrinkfactor)
	    value = home_x / mane.shrinkfactor;
	(void)set_bar_value(x_bar, value, (int)(page_w - mane.width));

	(void)set_bar_value(y_bar, (int)(page_h - mane.height), (int)(page_h - mane.height));
    }
#else /* MOTIF */
    get_xy();
    if (x_bar != NULL) {
	int coord = (page_w - mane.width) / 2;

	if (coord > home_x / mane.shrinkfactor)
	    coord = home_x / mane.shrinkfactor;
	XtCallCallbacks(x_bar, XtNscrollProc, (XtPointer) (window_x + coord));
    }
    if (y_bar != NULL)
	XtCallCallbacks(y_bar, XtNscrollProc, (XtPointer)(window_y + (page_h - mane.height)));
#endif /* MOTIF */
    XMapWindow(DISP, mane.win);
    /* Wait for the server to catch up---this eliminates flicker. */
    XSync(DISP, False);

    handle_x_scroll(NULL, NULL, NULL, NULL);
    handle_y_scroll(NULL, NULL, NULL, NULL);
}


#ifndef MOTIF
static void
handle_destroy_bar(Widget w, XtPointer client_data, XtPointer call_data)
{
    UNUSED(w);
    UNUSED(call_data);
    *(Widget *) client_data = NULL;
}
#endif

static Boolean resized = False;

static void
get_geom(void)
{
    static Dimension new_clip_w, new_clip_h;
    
    static Arg arg_wh_clip[] = {
	{XtNwidth, (XtArgVal) & new_clip_w},
	{XtNheight, (XtArgVal) & new_clip_h},
    };
    int old_clip_w;

#ifdef MOTIF
    /* event handlers for Motif scrollbars have already been added
       in create_initialize_widgets(), xdvi.c */
    XtGetValues(main_window, arg_wh, XtNumber(arg_wh));
#else
    XtGetValues(vport_widget, arg_wh, XtNumber(arg_wh));
    /* Note:  widgets may be destroyed but not forgotten */
    if (x_bar == NULL) {
	x_bar = XtNameToWidget(vport_widget, "horizontal");
	if (x_bar != NULL) {
	    XtAddCallback(x_bar, XtNdestroyCallback, handle_destroy_bar, (XtPointer)&x_bar);
	    XtAddEventHandler(x_bar, ButtonPressMask | ButtonReleaseMask, False, handle_x_scroll, NULL);
	}
    }
    if (y_bar == NULL) {
	y_bar = XtNameToWidget(vport_widget, "vertical");
	if (y_bar != NULL) {
	    XtAddCallback(y_bar, XtNdestroyCallback, handle_destroy_bar, (XtPointer)&y_bar);
	    XtAddEventHandler(y_bar, ButtonPressMask | ButtonReleaseMask, False, handle_y_scroll, NULL);
	}
    }
#endif
    XtGetValues(clip_widget, arg_wh_clip, XtNumber(arg_wh_clip));

    old_clip_w = mane.width;

    /* we need to do this because 
       sizeof(Dimension) != sizeof(int)
    */
    mane.width = new_clip_w;
    mane.height = new_clip_h;
    if (old_clip_w == 0) {
	ev_flags |= EV_NEWPAGE;
    }

    if (resource.keep_flag) {
#ifndef MOTIF
	Dimension d;
	int curr_scroll;
	if ((x_bar != NULL && m_x_scroll != 0) || (y_bar != NULL && m_y_scroll != 0)) {
	    get_xy();
	}
#endif
	if (x_bar != NULL && m_x_scroll != 0) {
#ifdef MOTIF
	    if (m_x_scroll > 0)
		(void)set_bar_value(x_bar, m_x_scroll, (int)(page_w - mane.width));
#else
	    XtVaGetValues(clip_widget, XtNy, &d, NULL);
	    curr_scroll = d - window_x;
	    if (m_x_scroll > curr_scroll) {
		TRACE_GUI((stderr, "======== diff: %d", m_x_scroll - curr_scroll));
		XtCallCallbacks(x_bar, XtNscrollProc, (XtPointer)(m_x_scroll - curr_scroll));
	    }
#endif
	}
	if (y_bar != NULL && m_y_scroll != 0) {
#ifdef MOTIF
	    if (m_y_scroll > 0)
		(void)set_bar_value(y_bar, m_y_scroll, (int)(page_h - mane.height));
#else
	    XtVaGetValues(clip_widget, XtNy, &d, NULL);
	    curr_scroll = d - window_y;
	    if (m_y_scroll > curr_scroll) {
		TRACE_GUI((stderr, "======== diff: %d", m_y_scroll - curr_scroll));
		XtCallCallbacks(y_bar, XtNscrollProc, (XtPointer)(m_y_scroll - curr_scroll));
	    }
#endif
	}
    }
    /*     home(False); */
    resized = False;
}

/*
 * Callback routines
 */

void
handle_resize(Widget widget, XtPointer junk, XEvent *event, Boolean *cont)
{
    UNUSED(widget);
    UNUSED(junk);
    UNUSED(event);
    UNUSED(cont);
    
    resized = True;
#ifndef MOTIF
    handle_statusline_resize();
    handle_pagelist_resize();
#endif
}

void
reconfig(void)
{
/*      Dimension x, y; */

    if (g_bak_fp == NULL)
	return;
    
#ifndef MOTIF
    XtVaSetValues(vport_widget, XtNresizable, (XtArgVal)False, NULL);
#endif
    TRACE_GUI((stderr, "draw_widget: w %d, h %d", page_w, page_h));
    XtVaSetValues(draw_widget, XtNwidth, (XtArgVal)page_w, XtNheight, (XtArgVal)page_h, NULL);
    
#ifdef TEST_SCROLLING
/*     XtVaSetValues(draw_background, XtNwidth, (XtArgVal)page_w, XtNheight, (XtArgVal)page_h, NULL); */
#endif
    
#ifndef MOTIF
    handle_statusline_resize(); /* without this, statusline will disappear */
    /* following not needed? */
    /*     handle_pagelist_resize(); */
#endif
    
    get_geom();

/*     set_windowsize(&x, &y */
/* #ifndef MOTIF */
/* 		   , get_panel_width() */
/* #endif */
/* 		   ); */
/*     XResizeWindow(DISP, XtWindow(top_level), x, y); */
/*      reconfigure_window(False, x, y, True); */
/*      reconfig_window(); */
}


int
check_goto_page(int pageno)
{
    if (pageno < 0) {
	XBell(DISP, 0);
	statusline_print(STATUS_SHORT, "Can't go to page %d, going to first page instead", pageno + 1);
	return 0;
    }
    else if (pageno >= total_pages) {
	XBell(DISP, 0);
	statusline_print(STATUS_SHORT,
			 "Can't go to page %d, going to last page (%d) instead",
			 pageno + 1, total_pages);
	return total_pages - 1;
    }
    else
	return pageno;
}


static int
check_goto_tex_page(int pageno)
{
    /*
      Translate from TeX page number to `real' page number if
      needed. Note that pageno is a C-style 0-based number, hence we
      add 1 for the argument of pageinfo_get_index_of_number().
    */
    if (resource.use_tex_pages) {
	int res = pageinfo_get_index_of_number(pageno + 1);
	if (res >= 0)
	    return res;
	else {
	    XBell(DISP, 0);
	    if (pageno < 1) {
		statusline_print(STATUS_SHORT, "Can't go to page %d, going to first page instead", pageno + 1);
		return 0;
	    }
	    else {
		/* there is no quick way to determine the last page number in the TeX page index */
		statusline_print(STATUS_SHORT,
				 "Can't go to page %d, going to last page instead",
				 pageno + 1);
		return total_pages - 1;
	    }
	}
    }
    else
	return check_goto_page(pageno);
}

/* |||
 *	Currently the event handler does not coordinate XCopyArea requests
 *	with GraphicsExpose events.  This can lead to problems if the window
 *	is partially obscured and one, for example, drags a scrollbar.
 */

/*
 *	Actions for the translation mechanism.
 */

Boolean have_arg = False; /* flag whether next action will have a `prefix argument' (e.g. 1g) */
int number = 0;
int sign = 1;

/* if there are global prefixes, return them in res and reset them to defaults */
static Boolean
get_prefix_arg(int *res)
{
    Boolean ret;

    *res = sign * number;
    ret = have_arg;
    /* reset global flags */
    have_arg = False;
    number = 0;
    sign = 1;
    /* erase `prefix' text */
    statusline_print(STATUS_SHORT, "");
    return ret;
}

Boolean
get_int_arg(String *param, Cardinal *num_params, int *res)
{
    if (*num_params > 0) {
	*res = atoi(*param);
	return True;
    }
    else {
	if (get_prefix_arg(res)) { /* prefix argument? */
	    return True;
	}
    }
    return False;
}

#if 0
//UNUSED static Boolean
//UNUSED get_float_arg(String *param, Cardinal *num_params, double *res)
//UNUSED {
//UNUSED     if (*num_params > 0) {
//UNUSED 	*res = strtod(*param, (char **)NULL);
//UNUSED 	return True;
//UNUSED     }
//UNUSED     return False;
//UNUSED }
#endif /* 0 */

Boolean
toggle_arg(int arg, String *param, Cardinal *num_params)
{
    if (*num_params > 0) {
	if (**param != 't' && (atoi(*param) != 0) == arg)
	    return False;
    }
    else {
	if (have_arg) {
	    int	tmparg = number;

	    have_arg = False;
	    number = 0;
	    sign = 1;

	    if ((tmparg != 0) == arg)
		return False;
	}
    }
    return True;
}

#if defined(NEW_MENU_CREATION) || defined(MOTIF)
Boolean
check_expert_flag(void *val, const char *param)
{
    int j = strtol(param, (char **)NULL, 10);
    /* check if the j-1th bit is set: */
    return (*(int *)val >> (j - 1)) & 1;
}

Boolean
check_papersize(void *val, const char *param)
{
#ifdef __GNUC__
#warning FIXME: implement check_papersize
#endif
    UNUSED(val);
    UNUSED(param);
    return False; /* TODO */
}

Boolean
check_paper_landscape(void *val, const char *param)
{
#ifdef __GNUC__
#warning FIXME: implement check_paper_landscape
#endif
    UNUSED(val);
    UNUSED(param);
    return False; /* TODO */
}

/* comparison functions for the menu setting code */
Boolean
check_toggle(void *val, const char *param)
{
    Boolean *on = val;
    if (strcmp(param, "toggle") == 0) {
	return *on;
    }
    else {
	fprintf(stderr, "TODO: check_toggle: arg |%s|, curr: %d\n", param, *(int *)val);
	return *on;
    }
}

Boolean
check_int(void *val, const char *param)
{
    int i = strtol(param, (char **)NULL, 10);
    return i == *(int *)val;
}
#endif /* NEW_MENU_CREATION */


static void
Act_digit(Widget w, XEvent *event,
	  String *params, Cardinal *num_params)
{
    unsigned int digit;

    UNUSED(w);
    UNUSED(event);

    if (*num_params != 1 || (digit = **params - '0') > 9) {
	XBell(DISP, 0);
	return;
    }
    have_arg = True;
    number = number * 10 + digit;
    if (expert_flag & XPRT_SHOW_STATUSLINE) /* too distracting for stdout */
	statusline_print(STATUS_SHORT, "numerical prefix: %s%d", sign < 0 ? "-" : "", number);
}

static void
Act_minus(Widget w, XEvent *event,
	  String *params, Cardinal *num_params)
{
    UNUSED(w);
    UNUSED(event);
    UNUSED(params);
    UNUSED(num_params);
    
    have_arg = True;
    sign = -sign;
    if (number > 0) {
	if (expert_flag & XPRT_SHOW_STATUSLINE) /* too distracting for stdout */
	    statusline_print(STATUS_SHORT, "numerical prefix: %s%d", sign < 0 ? "-" : "", number);
    }
    else {
	if (expert_flag & XPRT_SHOW_STATUSLINE) /* too distracting for stdout */
	    statusline_print(STATUS_SHORT, "numerical prefix: %s", sign < 0 ? "-" : "");
    }
}

static void
Act_quit(Widget w, XEvent *event,
	 String *params, Cardinal *num_params)
{
    UNUSED(w);
    UNUSED(event);
    UNUSED(params);
    UNUSED(num_params);
    
#ifndef FLAKY_SIGPOLL
    if (debug & DBG_EVENT)
	puts(event_freq < 0
	     ? "SIGPOLL is working"
	     : "no SIGPOLL signals received");
#endif
    xdvi_normal_exit();
}

static void
Act_load_url(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    UNUSED(w);
    UNUSED(event);
    UNUSED(num_params);
    launch_browser(*params);
}

static void
Act_print(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    UNUSED(w);
    UNUSED(event);
    UNUSED(params);
    UNUSED(num_params);
    save_or_print_callback(FILE_PRINT);
}

static void
Act_save(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    UNUSED(w);
    UNUSED(event);
    UNUSED(params);
    UNUSED(num_params);
    save_or_print_callback(FILE_SAVE);
}

static void
Act_find(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    UNUSED(w);
    UNUSED(event);
    UNUSED(params);
    UNUSED(num_params);
    dvi_search_dialog();
}

static void
Act_help(Widget w, XEvent *event,
	 String *params, Cardinal *num_params)
{
    UNUSED(w);
    UNUSED(event);
    
    show_help(top_level);

    if (*num_params > 0) {
	char *arg = *params;
	select_help_topic(arg);
    }
}

static void
Act_close_help(Widget w, XEvent *event,
	 String *params, Cardinal *num_params)
{
    UNUSED(event);
    UNUSED(params);
    UNUSED(num_params);
    
    popdown_help_callback(w, NULL, NULL);
}

static home_proc home_action = NULL;

void
goto_page(int new_page, home_proc proc)
{
    if (g_bak_fp == NULL) {
	current_page = new_page; /* so that xdvi gets the page change */
	return;
    }

    ASSERT(new_page >= 0 && new_page < total_pages, "new_page in goto_page() out of range");   
    if (current_page != new_page) {
	cursor_flags &= ~LINK_CURSOR; /* disable link cursor if needed */

	current_page = new_page;
	home_action = proc;
	warn_spec_now = resource.warn_spec;
    /* this seems unneccessary */
/* 	if (!resource.keep_flag) */
/* 	    home(False); */
#if defined(MOTIF) && HAVE_XPM
	tb_check_navigation_sensitivity(current_page);
#endif
	maybe_scroll_pagelist(current_page, False);

	/* SU: added clearing the window here, else old window content
	   will survive switching pages as long as pause_flag is active
	   (i.e. when inadvertedly changing page instead of unpausing)
	 */
	if (pause_flag) {
	    XClearWindow(DISP, mane.win);
	}
	    
	if (pause_num_save)
	    pause_num = pause_num_save[new_page];
	else
	    pause_num = 0;
	/* Control-L (and changing the page) clears this box */
	source_fwd_box_page = -1;

	ev_flags |= EV_NEWPAGE;
	XFlush(DISP);       
    }
}


static void
Act_goto_page(Widget w, XEvent *event,
	      String *params, Cardinal *num_params)
{
    int arg;

    UNUSED(w);
    UNUSED(event);

    if (*num_params > 0) {
	if (**params == 'e') {
	    arg = total_pages - 1;
	}
	else
	    arg = atoi(*params) - pageno_correct;
    }
    else {
	if (get_prefix_arg(&arg)) {
	    arg -= pageno_correct;
	}
	else {
	    arg = total_pages - 1;
	}
    }
    if (arg == total_pages - 1) {
	/* with TeX numbers, total_pages - 1 might not be found in the page list,
	   so don't use check_goto_tex_page() in this case */
	goto_page(check_goto_page(arg), resource.keep_flag ? NULL : home);
    }
    else {
	goto_page(check_goto_tex_page(arg), resource.keep_flag ? NULL : home);
    }
    search_signal_page_changed();
}

void
Act_use_tex_pages(Widget w, XEvent *event,
		  String *params, Cardinal *num_params)
{
    UNUSED(w);
    UNUSED(event);

    if (*num_params == 0) {
	if (have_arg) {
	    resource.use_tex_pages = (number != 0);
	    have_arg = False;
	    number = 0;
	    sign = 1;
	}
	else {
	    resource.use_tex_pages = !resource.use_tex_pages;
	}
    }
    else
	resource.use_tex_pages = (**params == 't' ? !resource.use_tex_pages : atoi(*params));
    
    if (resource.use_tex_pages) {
	statusline_print(STATUS_SHORT, "Using TeX page numbers for \"g\", goto-page()");
    }
    else {
	statusline_print(STATUS_SHORT, "Using physical page numbers for \"g\", goto-page()");
    }
#if defined(NEW_MENU_CREATION) || defined(MOTIF)
    set_menu(&resource.use_tex_pages, Act_use_tex_pages, check_toggle);
#else
#ifdef MOTIF
   set_use_tex_option();
#else
    toggle_menu(resource.use_tex_pages,  Act_use_tex_pages);
#endif
#endif /* NEW_MENU_CREATION */
}

void
Act_forward_page(Widget w, XEvent *event,
		 String *params, Cardinal *num_params)
{
    int arg;

    UNUSED(w);
    UNUSED(event);

    if (!get_int_arg(params, num_params, &arg)) {
	arg = 1;
    }
    arg += current_page;

    if (arg == current_page) { /* zero argument -> redraw page */
	source_fwd_box_page = -1;
	search_reset_info();
	ev_flags |= EV_NEWPAGE;
	XFlush(DISP);
	statusline_print(STATUS_SHORT, "Page redrawn");
	return;
    }
    else if (current_page >= total_pages - 1) {
	XBell(DISP, 0);
	statusline_print(STATUS_SHORT, "Last page of DVI file");
	return;
    }
    
    goto_page(check_goto_page(arg), resource.keep_flag ? NULL : home);
    search_signal_page_changed();
}

void
Act_back_page(Widget w, XEvent *event,
	      String *params, Cardinal *num_params)
{
    int arg;
    
    UNUSED(w);
    UNUSED(event);

    if (!get_int_arg(params, num_params, &arg)) {
	arg = 1;
    }
    arg = current_page - arg;

    if (current_page == 0) {
	XBell(DISP, 0);
	statusline_print(STATUS_SHORT, "First page of DVI file");
	return;
    }

    goto_page(check_goto_page(arg), resource.keep_flag ? NULL : home);
    search_signal_page_changed();
}

static void
Act_declare_page_number(Widget w, XEvent *event,
			String *params, Cardinal *num_params)
{
    int arg;

    UNUSED(w);
    UNUSED(event);

    if (resource.ruler_mode) {
	show_distance_from_ruler(event, True);
	return;
    }
    
    if (!get_int_arg(params, num_params, &arg)) {
	arg = 0;
    }
    pageno_correct = arg - current_page;
    statusline_print(STATUS_SHORT, "Current page number set to %d", arg);
}

static void
Act_toggle_mark(Widget w, XEvent *event,
		String *params, Cardinal *num_params)
{
    int arg;

    UNUSED(w);
    UNUSED(event);

    if (g_bak_fp == NULL)
	return;
    
    if (
#ifdef MOTIF
	(expert_flag & XPRT_SHOW_PAGELIST) == 0
#else
	(expert_flag & XPRT_SHOW_BUTTONS) == 0
#endif
	)
	return;
    
    if (*num_params > 0) {
	arg = atoi(*params);
	if (arg < -1 || arg > 2) {
	    XBell(DISP, 0);
	    statusline_print(STATUS_SHORT,
			     "Possible arguments: none (toggle current), "
			     "-1 (mark all), 0 (unmark all), 1 (toggle odd), 2 (toggle even)");
	}
	list_toggle_marks(arg);
    }
    else if (get_prefix_arg(&arg)) {
	list_toggle_marks(arg);
    }
    else {
	arg = current_page;
	list_toggle_current(arg);
    }
}

static void
Act_home(Widget w, XEvent *event,
	 String *params, Cardinal *num_params)
{
    UNUSED(w);
    UNUSED(event);
    UNUSED(params);
    UNUSED(num_params);

    home(True);
}

static void
Act_center(Widget w, XEvent *event,
	   String *params, Cardinal *num_params)
{
    int x, y;

    UNUSED(w);
    UNUSED(params);
    UNUSED(num_params);

#ifdef MOTIF
    /* The clip widget gives a more exact value. */
    x = event->xkey.x - mane.width / 2;
    y = event->xkey.y - mane.height / 2;

    x = set_bar_value(x_bar, x, (int)(page_w - mane.width));
    y = set_bar_value(y_bar, y, (int)(page_h - mane.height));
    get_xy();
    XWarpPointer(DISP, None, None, 0, 0, 0, 0, -x - window_x, -y - window_y);
#else

    if (event == NULL)
	return;	/* button actions do not provide events */

    x = event->xkey.x - mane.width / 2;
    y = event->xkey.y - mane.height / 2;
    /* The clip widget gives a more exact value. */
    if (x_bar != NULL)
	XtCallCallbacks(x_bar, XtNscrollProc, (XtPointer) x);
    if (y_bar != NULL)
	XtCallCallbacks(y_bar, XtNscrollProc, (XtPointer) y);
    XWarpPointer(DISP, None, None, 0, 0, 0, 0, -x, -y);
#endif
    handle_x_scroll(NULL, NULL, NULL, NULL);
    handle_y_scroll(NULL, NULL, NULL, NULL);
}

void
Act_ruler_snap_origin(Widget w, XEvent *event,
		      String *params, Cardinal *num_params)
{
    UNUSED(w);
    UNUSED(params);
    UNUSED(num_params);

    if (resource.ruler_mode)
	ruler_snap_origin(event);
}

#if defined(NEW_MENU_CREATION) || defined(MOTIF)
void
Act_set_papersize(Widget w, XEvent *event,
		  String *params, Cardinal *num_params)
{
    UNUSED(w);
    UNUSED(event);
    UNUSED(num_params);

    fprintf(stderr, "set_paperformat: %s\n", *params);
    resource.paper = *params;
    set_menu((void *)resource.paper, Act_set_papersize, check_papersize);
}

void
Act_set_paper_landscape(Widget w, XEvent *event,
			String *params, Cardinal *num_params)
{
    UNUSED(w);
    UNUSED(event);
    UNUSED(num_params);

    resource.paper_landscape = !resource.paper_landscape;
    fprintf(stderr, "set_paper_landscape: %s\n", *params);

    set_menu((char *)resource.paper,  Act_set_paper_landscape, check_paper_landscape);
}
#endif /* NEW_MENU_CREATION */

void
Act_set_keep_flag(Widget w, XEvent *event,
		  String *params, Cardinal *num_params)
{
    UNUSED(event);
    UNUSED(w);
    
    if (*num_params == 0) {
	if (have_arg) {
	    resource.keep_flag = (number != 0);
	    have_arg = False;
	    number = 0;
	    sign = 1;
	}
	else {
	    resource.keep_flag = !resource.keep_flag;
	}
    }
    else
	resource.keep_flag = (**params == 't'
			      ? !resource.keep_flag : atoi(*params));
    if (resource.keep_flag) {
	statusline_print(STATUS_SHORT, "Keeping position when switching pages");
    }
    else {
	statusline_print(STATUS_SHORT,
			 "Not keeping position when switching pages");
    }
#if defined(NEW_MENU_CREATION) || defined(MOTIF)
    set_menu(&resource.keep_flag, Act_set_keep_flag, check_toggle);
#else
#ifdef MOTIF
   set_keep_option();
#else
    toggle_menu(resource.keep_flag,  Act_set_keep_flag);
#endif
#endif /* NEW_MENU_CREATION */
}

static void
Act_left(Widget w, XEvent *event,
	 String *params, Cardinal *num_params)
{
    UNUSED(w);
    UNUSED(event);

#ifdef MOTIF
    get_xy();
#ifdef TEST_SCROLLING
    fprintf(stderr, "left for %p\n", w);
    fprintf(stderr, "window x: %d, y: %d\n", window_x, window_y);
#endif

    (void)set_bar_value(x_bar, (*num_params == 0 ? (-2 * (int)mane.width / 3)
				: (int)(-atof(*params) * mane.width)) - window_x,
			(int)(page_w - mane.width));
#else
    if (x_bar != NULL)
	XtCallCallbacks(x_bar, XtNscrollProc,
			(XtPointer) (*num_params == 0 ? (-2 * (int)mane.width / 3)
				     : (int)(-atof(*params) * mane.width)));
    else {
	XBell(DISP, 0);
	statusline_print(STATUS_SHORT, "Horizontal scrolling not possible");
    }
#endif
    handle_x_scroll(NULL, NULL, NULL, NULL);
}

static void
Act_right(Widget w, XEvent *event,
	  String *params, Cardinal *num_params)
{
    UNUSED(w);
    UNUSED(event);

#ifdef MOTIF
    get_xy();
#ifdef TEST_SCROLLING
    fprintf(stderr, "right for %p\n", w);
    fprintf(stderr, "window x: %d, y: %d\n", window_x, window_y);
#endif

    (void)set_bar_value(x_bar, (*num_params == 0 ? (2 * (int)mane.width / 3)
				: (int)(atof(*params) * mane.width)) - window_x,
			(int)(page_w - mane.width));
#else
    if (x_bar != NULL)
	XtCallCallbacks(x_bar, XtNscrollProc,
			(XtPointer) (*num_params == 0 ? (2 * (int)mane.width / 3)
				     : (int)(atof(*params) * mane.width)));
    else {
	XBell(DISP, 0);
	statusline_print(STATUS_SHORT, "Horizontal scrolling not possible");
    }
#endif
    handle_x_scroll(NULL, NULL, NULL, NULL);
}

static void
Act_up(Widget w, XEvent *event,
       String *params, Cardinal *num_params)
{
    UNUSED(w);
    UNUSED(event);

    do_autoscroll = False;
#ifdef MOTIF
#ifdef TEST_SCROLLING
    fprintf(stderr, "up for %p\n", w);
#endif
    get_xy();
    (void)set_bar_value(y_bar, (*num_params == 0 ? (-2 * (int)mane.height / 3)
				: (int)(-atof(*params) * mane.height)) - window_y,
			(int)(page_h - mane.height));
#else
    if (y_bar != NULL)
	XtCallCallbacks(y_bar, XtNscrollProc,
			(XtPointer) (*num_params == 0 ? (-2 * (int)mane.height / 3)
				     : (int)(-atof(*params) * mane.height)));
    else {
	XBell(DISP, 0);
	statusline_print(STATUS_SHORT, "Vertical scrolling not possible");
    }
#endif
    handle_y_scroll(NULL, NULL, NULL, NULL);
}

static void
Act_down(Widget w, XEvent *event,
	 String *params, Cardinal *num_params)
{
    UNUSED(w);
    UNUSED(event);

    do_autoscroll = False;
#ifdef MOTIF
#ifdef TEST_SCROLLING
    fprintf(stderr, "down for %p\n", w);
#endif
    get_xy();
    (void)set_bar_value(y_bar, (*num_params == 0 ? (2 * (int)mane.height / 3)
				: (int)(atof(*params) * mane.height)) - window_y,
			(int)(page_h - mane.height));
#else
    if (y_bar != NULL)
	XtCallCallbacks(y_bar, XtNscrollProc,
			(XtPointer) (*num_params == 0 ? (2 * (int)mane.height / 3)
				     : (int)(atof(*params) * mane.height)));
    else {
	XBell(DISP, 0);
	statusline_print(STATUS_SHORT, "Vertical scrolling not possible");
    }
#endif
    handle_y_scroll(NULL, NULL, NULL, NULL);
}

static void
Act_down_or_next(Widget w, XEvent *event,
		 String *params, Cardinal *num_params)
{
    UNUSED(w);
    UNUSED(event);

#ifdef TEST_SCROLLING
    fprintf(stderr, "down-or-next for %p\n", w);
#endif
    do_autoscroll = False;

#ifdef MOTIF
    get_xy();
    if (window_y > (int)mane.height - (int)page_h) {
	(void)set_bar_value(y_bar, (*num_params == 0 ? (2 * (int)mane.height / 3)
				    : (int)(atof(*params) * mane.height)) -
			    window_y, (int)(page_h - mane.height));
	return;
    }
#else
    if (y_bar != NULL) {
	get_xy();
	if (window_y > (int)mane.height - (int)page_h) {
	    XtCallCallbacks(y_bar, XtNscrollProc,
			    (XtPointer) (*num_params ==
					 0 ? (2 * (int)mane.height / 3)
					 : (int)(atof(*params) * mane.height)));
	    return;
	}
    }
#endif

    if (current_page < total_pages - 1) {
	goto_page(current_page + 1, home);
	search_signal_page_changed();
    }
    else {
	XBell(DISP, 0);
	statusline_print(STATUS_SHORT, "At bottom of last page");
    }
    handle_y_scroll(NULL, NULL, NULL, NULL);
}


static void
Act_up_or_previous(Widget w, XEvent *event,
		   String *params, Cardinal *num_params)
{
    UNUSED(w);
    UNUSED(event);

#ifdef TEST_SCROLLING
    fprintf(stderr, "up-or-previous for %p\n", w);
#endif
    do_autoscroll = False;

#ifdef MOTIF
    get_xy();
    if (window_y < 0) {
	(void)set_bar_value(y_bar,
			    (*num_params == 0 ? (-2 * (int)mane.height / 3)
			     : (int)(-atof(*params) * mane.height)) - window_y,
			    (int)(page_h - mane.height));
	return;
    }
#else
    if (y_bar != NULL) {
	get_xy();
	if (window_y < 0) {
	    XtCallCallbacks(y_bar, XtNscrollProc,
			    (XtPointer) (*num_params ==
					 0 ? (-2 * (int)mane.height / 3)
					 : (int)(-atof(*params) * mane.height)));
	    return;
	}
    }
#endif

    if (current_page > 0) {
	goto_page(current_page - 1, home_bottom);
	search_signal_page_changed();
    }
    else {
	XBell(DISP, 0);
	statusline_print(STATUS_SHORT, "At top of first page");
    }
    handle_y_scroll(NULL, NULL, NULL, NULL);
}


static void
Act_set_margins(Widget w, XEvent *event,
		String *params, Cardinal *num_params)
{
    Window ww;

    UNUSED(w);
    UNUSED(params);
    UNUSED(num_params);

#ifndef MOTIF
    if (event == NULL)
	return;	/* button actions do not provide events */
#endif

    (void)XTranslateCoordinates(DISP, event->xkey.window, mane.win,
				event->xkey.x, event->xkey.y, &home_x, &home_y, &ww);	/* throw away last argument */
    statusline_print(STATUS_SHORT, "Margins set to cursor position (%d, %d)", home_x, home_y);
    home_x *= mane.shrinkfactor;
    home_y *= mane.shrinkfactor;
}

static void
Act_show_display_attributes(Widget w, XEvent *event,
			    String *params, Cardinal *num_params)
{
    UNUSED(w);
    UNUSED(event);
    UNUSED(params);
    UNUSED(num_params);

    statusline_print(STATUS_SHORT, "Unit = %d, bitord = %d, byteord = %d",
		     BitmapUnit(DISP), BitmapBitOrder(DISP),
		     ImageByteOrder(DISP));
}

int
shrink_to_fit(void)
{
    int value1;
    int value2;

    value1 = ROUNDUP(unshrunk_page_w, window_w - 2);

#ifdef  MOTIF
    {	/* account for menubar */
	static Dimension new_h;

	/* get rid of scrollbar */
	XtVaSetValues(draw_widget, XtNwidth, (XtArgVal)1, XtNheight, (XtArgVal)1, NULL);
	XtVaGetValues(clip_widget, XtNheight, &new_h, NULL);
	value2 = ROUNDUP(unshrunk_page_h, new_h - 2);
    }
#else
    /* FIXME: value seems too small here! */
    value2 = ROUNDUP(unshrunk_page_h, window_h - 2);
#endif

    return value1 > value2 ? value1 : value2;
}

void
Act_set_shrink_factor(Widget w, XEvent *event,
		      String *params, Cardinal *num_params)
{
    int arg;

    UNUSED(w);
    UNUSED(event);

    if (*num_params > 0) {
	if (**params == 'a')
	    arg = shrink_to_fit();
	else if (**params == '-')
	    arg =  mane.shrinkfactor + 1;
	else if (**params == '+')
	    arg = mane.shrinkfactor - 1;
	else
	    arg = atoi(*params);
    }
    else if (!get_prefix_arg(&arg)) {
	arg = shrink_to_fit();
    }

    if (arg <= 0) {
	XBell(DISP, 0);
	statusline_print(STATUS_SHORT,
			 "No more enlarging possible");
	return;
    }
    
    if (arg > 1000) { /* this would consume too much memory
			 for the subpixel lookup table pixeltbl, dvi-draw.c
			 FIXME: A better way would be to recognize that at a certain
			 shrink, the only pixel for a character will be always set
			 anyway, so that we can do without the lookup table in the
			 first place.
		      */
	XBell(DISP, 0);
	statusline_print(STATUS_SHORT,
			 "Shrink factor %d too large", arg);
	return;
    }
    

    statusline_print(STATUS_SHORT, "shrink factor: %d", arg);
    if (arg == mane.shrinkfactor) {
	/* Uncommented this, otherwise selecting `shrink to fit' with same value as
	   current mane.shrinkfactor will clear the canvas (only fixed by changing shrink
	   factor again). That bug is also present in xdvik-22.48-alpha3. */
	/* return; */
    }
    mane.shrinkfactor = arg;
#if defined(NEW_MENU_CREATION) || defined(MOTIF)
    set_menu(&arg, Act_set_shrink_factor, check_int);
#else
#ifdef MOTIF
   set_shrink_factor();
#else
    toggle_menu(arg, Act_set_shrink_factor);
#endif
#endif /* NEW_MENU_CREATION */
    if (arg != 1 && arg != bak_shrink) {
	bak_shrink = arg;
#if GREY
#if COLOR
	if (resource.use_grey)
	    fg_active = NULL;
#else 
	if (resource.use_grey)
	    init_pix();
#endif
#endif /* GREY */
	reset_fonts();
    }
    init_page();
    reconfig();
#ifdef HTEX
    htex_resize_page();
#endif /* HTEX */
    /* this seems unneccessary */
/*     if (!resource.keep_flag) */
/* 	home(False); */
    ev_flags |= EV_NEWPAGE;
    XFlush(DISP);
}

void
Act_shrink_to_dpi(Widget w, XEvent *event,
		  String *params, Cardinal *num_params)
{
    int arg, arg_bak;

    UNUSED(w);
    UNUSED(event);

    if (!get_int_arg(params, num_params, &arg)) {
	arg = 0;
    }
    arg_bak = arg;
    
    if (arg > 0)
	arg = (double)resource.pixels_per_inch / arg + 0.5;

    if (arg <= 0) {
	XBell(DISP, 0);
	statusline_print(STATUS_SHORT,
			 "shrink-to-dpi requires a positive argument");
	return;
    }

    if (arg == mane.shrinkfactor)
	return;
#if defined(NEW_MENU_CREATION) || defined(MOTIF)
    mane.shrinkfactor = arg;
    set_menu(&arg_bak, Act_shrink_to_dpi, check_int);
#else
#ifdef MOTIF
   set_shrink_factor();
#else
    mane.shrinkfactor = arg;
    toggle_menu(arg_bak, Act_shrink_to_dpi);
#endif
#endif /* NEW_MENU_CREATION */
    
    if (arg != 1 && arg != bak_shrink) {
	bak_shrink = arg;
#if GREY && !COLOR
	if (resource.use_grey)
	    init_pix();
#endif
	reset_fonts();
    }
    init_page();
    reconfig();
/*     if (!resource.keep_flag) */
/* 	home(False); */
    ev_flags |= EV_NEWPAGE;
    XFlush(DISP);
}

void
do_set_density(double newgamma)
{
    
#ifdef GREY
    if (resource.use_grey) {
	if (newgamma == resource.gamma) {
	    statusline_print(STATUS_SHORT, "density value: %.3f", newgamma);
	    return;
	}
	resource.gamma = newgamma;
#if COLOR
	fg_active = NULL;
	reset_colors();
#else
	init_pix();
	if (our_visual->class != TrueColor) {
	    return;
	}
	reset_fonts();
#endif /* COLOR */
	statusline_print(STATUS_SHORT, "density value: %.3f", newgamma);
    } else
#endif /* GREY */
    {
	reset_fonts();
	if (mane.shrinkfactor == 1) {
	    statusline_print(STATUS_SHORT,
		    "set-density ignored at magnification 1");
	    return;
	}
	statusline_print(STATUS_SHORT, "density value: %.3f", newgamma);
    }
    ev_flags |= EV_NEWPAGE;
    XFlush(DISP);
}

static void
Act_set_density(Widget w, XEvent *event,
		String *params, Cardinal *num_params)
{
    int arg;

    UNUSED(w);
    UNUSED(event);

    if (!get_int_arg(params, num_params, &arg)) {
	XBell(DISP, 0);
	return;
    }
    if (arg < 0) {
	XBell(DISP, 0);
	statusline_print(STATUS_SHORT,
			 "set-density requires a positive value");
	return;
    }
    do_set_density(arg != 0 ? arg / 100.0 : 1.0);
}


static void
Act_change_density(Widget w, XEvent *event,
		   String *params, Cardinal *num_params)
{
    int arg;
    double diff;
    double oldgamma = resource.gamma;
    UNUSED(w);
    UNUSED(event);

    
    if (!get_int_arg(params, num_params, &arg)) {
	XBell(DISP, 0);
	return;
    }
    diff = oldgamma / arg;
    if (oldgamma + diff <= 0.0)
	oldgamma = 0.0;
    else
	oldgamma += diff;
    do_set_density(oldgamma);
}

static void
Act_fullscreen(Widget w, XEvent *event,
	       String *params, Cardinal *num_params)
{
    Dimension main_win_w, main_win_h;
    static int orig_x = 0, orig_y = 0;
#ifndef MOTIF
    static Dimension get_x, get_y;
#endif
    static Dimension old_w = 0, old_h = 0;
    Dimension panel_width = 0;
#if 0
    static Dimension w_old = 0, h_old = 0;
#endif
    UNUSED(w);
    UNUSED(event);
    
    if (!toggle_arg(resource.fullscreen, params, num_params)) {
	return;
    }
    
    resource.fullscreen = !resource.fullscreen;
    
#ifndef MOTIF
    panel_width = get_panel_width();
#endif
    
    if (resource.fullscreen) {
	/* when toggling to fullscreen, save current scrollbar values
	   and window geometry */
	XtVaGetValues(top_level, XtNwidth, &old_w, XtNheight, &old_h, NULL);
	/*  	fprintf(stderr, "saved geometry: %dx%d\n", old_w, old_h); */
#ifdef MOTIF
	if (x_bar != NULL)
	    XtVaGetValues(x_bar, XmNvalue, &orig_x, NULL);
	if (y_bar != NULL)
	    XtVaGetValues(y_bar, XmNvalue, &orig_y, NULL);
#else
	get_xy();
	if (x_bar != NULL)
	    XtVaGetValues(clip_widget, XtNx, &get_x, NULL);
	if (y_bar != NULL)
	    XtVaGetValues(clip_widget, XtNy, &get_y, NULL);
	orig_x = -(get_x - window_x);
	orig_y = -(get_y - window_y);
#endif /* MOTIF */
	/*  	fprintf(stderr, "Current offsets: %d, %d, %d, %d\n", orig_x, window_x, orig_y, window_y); */
    }
#if 0 /* TODO: don't hardcode w/h values -
	 what's wrong with the set_windowsize() call below? */    
    if (resource.fullscreen) {
	XWindowAttributes attr;
	if (XGetWindowAttributes(DISP, XtWindow(top_level), &attr)) {
	    w_old = attr.width;
	    h_old = attr.height;
	}
	main_win_w = WidthOfScreen(SCRN);
	main_win_h = HeightOfScreen(SCRN);
    }
    else if (w_old > 0 && h_old > 0) {
	main_win_w = w_old;
	main_win_h = h_old;
    }
    else {
	main_win_w = 800;
	main_win_h = 600;
    }
#else
    if (resource.fullscreen)
	set_windowsize(&main_win_w, &main_win_h, panel_width, 0, False);
    else { /* re-use old window sizes */
	set_windowsize(&old_w, &old_h, panel_width, 0, True);
    }
#endif
    
/*      fprintf(stderr, "reconfigure window!\n"); */
    if (resource.fullscreen)
	reconfigure_window(resource.fullscreen, main_win_w, main_win_h, True);
    else {
	reconfigure_window(resource.fullscreen, old_w, old_h, True);
    }

    /*      fprintf(stderr, "orig values: %d, %d\n", orig_x, orig_y); */
    /* restore horizontal scroll position */
    if (resource.keep_flag && orig_x != 0 && x_bar != NULL) {
#ifdef MOTIF
	if (!resource.fullscreen) {
	    /* only scroll back in non-fullscreen mode. This was:
	       if (resource.fullscreen)
	           (void)set_bar_value(y_bar, orig_y, (int)(page_h - mane.height));
	       else ...
	       but that was broken (bogus values in page_h - mane.height).
	    */
	    (void)set_bar_value(x_bar, orig_x, INT_MAX);
	}
#else
	int curr_x;
	get_xy();
	XtVaGetValues(clip_widget, XtNx, &get_x, NULL);
	curr_x = -(get_x - window_x);
	if (curr_x - orig_x > 0)
	    XtCallCallbacks(x_bar, XtNscrollProc, (XtPointer)(curr_x - orig_x));
#endif /* MOTIF */
    }

    /* restore vertical scroll position */
    if (resource.keep_flag && orig_y != 0 && y_bar != NULL) {
#ifdef MOTIF
	if (!resource.fullscreen) {
	    /* only scroll back in non-fullscreen mode. This was:
	       if (resource.fullscreen)
	           (void)set_bar_value(y_bar, orig_y, (int)(page_h - mane.height));
	       else ...
	       but that was broken (bogus values in page_h - mane.height).
	    */
	    (void)set_bar_value(y_bar, orig_y, INT_MAX);
	}
#else
	int curr_y;
	get_xy();
	XtVaGetValues(clip_widget, XtNy, &get_y, NULL);
	curr_y = -(get_y - window_y);
	if (curr_y - orig_y > 0)
	    XtCallCallbacks(x_bar, XtNscrollProc, (XtPointer)(curr_y - orig_y));
#endif /* MOTIF */
    }
    handle_x_scroll(NULL, NULL, NULL, NULL);
    handle_y_scroll(NULL, NULL, NULL, NULL);
}

#ifdef GREY

static void
Act_set_greyscaling(Widget w, XEvent *event,
		    String *params, Cardinal *num_params)
{
    int arg;

    UNUSED(w);
    UNUSED(event);

    if (!get_int_arg(params, num_params, &arg)) { /* no arg, toggle */
	if (!toggle_arg(resource.use_grey, params, num_params)) {
	    return;
	}
	resource.use_grey = !resource.use_grey;
	if (resource.use_grey) {
	    statusline_print(STATUS_SHORT, "greyscaling on");
	}
	else {
	    statusline_print(STATUS_SHORT, "greyscaling off");
	}
	ev_flags |= EV_NEWPAGE;
	XFlush(DISP);
	return;
    }

    switch (arg) {
    case 0:
	resource.use_grey = False;
	statusline_print(STATUS_SHORT, "greyscaling off");
	break;
    case 1:
	resource.use_grey = True;
	statusline_print(STATUS_SHORT, "greyscaling on");
	break;
    default:
	{
	    float newgamma = arg != 0 ? arg / 100.0 : 1.0;
	    resource.use_grey = newgamma;
	    statusline_print(STATUS_SHORT, "greyscale value: %.1f", newgamma);
	}
    }
    if (resource.use_grey) {
	if (our_visual->class != TrueColor)
	    init_plane_masks();
#if COLOR
	fg_active = NULL;
#else
	init_pix();
#endif
    }
    reset_fonts();
    ev_flags |= EV_NEWPAGE;
    XFlush(DISP);
}

#endif

#if COLOR

static void
Act_set_color(Widget w, XEvent *event,
	      String *params, Cardinal *num_params)
{
    UNUSED(w);
    UNUSED(event);

    if (!toggle_arg(resource.use_color, params, num_params)) {
	return;
    };
	
    if (resource.use_color) {
	resource.use_color = False;
	full_reset_colors();
	scanned_page_color = total_pages;
#if PS
	if (ignore_papersize_specials || scanned_page_ps <= total_pages) {
	    scanned_page = scanned_page_ps;
	}
#endif
	statusline_print(STATUS_SHORT, "color specials off");
    }
    else {
	resource.use_color = True;
	scanned_page = scanned_page_color = scanned_page_reset;
	statusline_print(STATUS_SHORT, "color specials on");
    }
    ev_flags |= EV_NEWPAGE;
}

#endif /* COLOR */

#if PS

void
Act_set_ps(Widget w, XEvent *event,
	   String *params, Cardinal *num_params)
{
    int arg;
    UNUSED(event);
    UNUSED(w);
    
    if (!get_int_arg(params, num_params, &arg))
	resource.postscript++;
    else
	resource.postscript = arg;

    if (resource.postscript > 2)
	resource.postscript = 0;

#ifdef PS_GS
    if (!resource.useGS) {
	if (resource.postscript > 0) {
	    popup_message(MSG_WARN,
			     "This version of xdvi depends on ghostscript for rendering Postscript images. "
			     "Postscript rendering cannot be activated if the option ``-noghostscript'' is used "
			     "or if the resource ``Ghostscript'' is set to false.",
			     "Option ``-noghostscript'' is active; "
			     "cannot enable Postscript rendering without ghostscript.");
	}
	resource.postscript = 0;
    }
#endif
    if (resource.postscript > 0) {
	scanned_page_ps = scanned_page_ps_bak;
	if (scanned_page > scanned_page_ps)
	    scanned_page = scanned_page_ps;
	if (resource.postscript == 1) {
	    statusline_print(STATUS_SHORT, "Postscript rendering on");
	}
	else {
	    statusline_print(STATUS_SHORT, "Postscript rendering on (with bounding box)");
	}
    }
    else {
	scanned_page_ps_bak = scanned_page_ps;
	scanned_page_ps = total_pages;
#if COLOR
	if (ignore_papersize_specials || scanned_page_color <= total_pages) {
	    scanned_page = scanned_page_color;
	}
#endif
	statusline_print(STATUS_SHORT, "Postscript rendering off; displaying bounding box instead");
    }
    
    psp.toggle(resource.postscript);
#if defined(NEW_MENU_CREATION) || defined(MOTIF)
    set_menu(&resource.postscript, Act_set_ps, check_int);
#else
# ifdef MOTIF
   set_ps_option();
# else
    toggle_menu(resource.postscript, Act_set_ps);
# endif
#endif /* NEW_MENU_CREATION */
    ev_flags |= EV_PS_TOGGLE;
    XFlush(DISP);
}

#endif /* PS */

#ifdef HTEX
void
Act_htex_back(Widget w, XEvent *event,
	      String *params, Cardinal *num_params)
{
    UNUSED(w);
    UNUSED(event);
    UNUSED(params);
    UNUSED(num_params);

    htex_back();
}

void
Act_htex_forward(Widget w, XEvent *event,
		 String *params, Cardinal *num_params)
{
    UNUSED(w);
    UNUSED(event);
    UNUSED(params);
    UNUSED(num_params);

    htex_forward();
}

static void
Act_htex_anchorinfo(Widget w, XEvent *event,
		    String *params, Cardinal *num_params)
{
    int x, y;
    Window ww;

    UNUSED(w);
    UNUSED(params);
    UNUSED(num_params);

    (void)XTranslateCoordinates(DISP, event->xkey.window, mane.win,
				event->xkey.x, event->xkey.y, &x, &y, &ww);
    htex_displayanchor(x, y);
}
#endif /* HTEX */

#ifdef PS_GS
void
Act_set_gs_alpha(Widget w, XEvent *event,
		 String *params, Cardinal *num_params)
{
    UNUSED(w);
    UNUSED(event);

    if (!toggle_arg(resource.gs_alpha, params, num_params)) {
	return;
    }
    resource.gs_alpha = !resource.gs_alpha;

    if (resource.gs_alpha) {
	statusline_print(STATUS_SHORT, "ghostscript alpha active");
    }
    else {
	statusline_print(STATUS_SHORT, "ghostscript alpha inactive");
    }

#if defined(NEW_MENU_CREATION) || defined(MOTIF)
    set_menu(&resource.gs_alpha, Act_set_gs_alpha, check_toggle);
#else
# ifdef MOTIF
   set_gs_option();
# else
    toggle_menu(resource.gs_alpha, Act_set_gs_alpha);
# endif
#endif /* NEW_MENU_CREATION */
    ev_flags |= EV_PS_TOGGLE;
    XFlush(DISP);
}
#endif

static void
Act_reread_dvi_file(Widget w, XEvent *event,
		    String *params, Cardinal *num_params)
{
    UNUSED(w);
    UNUSED(event);
    UNUSED(params);
    UNUSED(num_params);

#if PS
    ps_clear_cache();
#if PS_GS
    if (resource.gs_alpha) {
	/* restart gs so that user has a method for fixing GS artifacts with gs_alpha
	   by using `reload' (see also GS_PIXMAP_CLEARING_HACK) */
	ps_destroy();
    }
#endif
#endif
/*      fprintf(stderr, "reread file!\n"); */
    ev_flags |= EV_RELOAD;
}

static void
Act_discard_number(Widget w, XEvent *event,
		   String *params, Cardinal *num_params)
{
    UNUSED(w);
    UNUSED(event);
    UNUSED(params);
    UNUSED(num_params);

    have_arg = False;
    number = 0;
    sign = 1;
    if (expert_flag & XPRT_SHOW_STATUSLINE) /* too distracting for stdout */
	statusline_print(STATUS_SHORT, "numerical prefix discarded");
}

/*
 * Actions to support dragging the image.
 */

static int drag_last_x, drag_last_y;	/* last position of cursor */

static void drag_motion(XEvent *);
static void drag_release(XEvent *);

static void
Act_drag(Widget w, XEvent *event,
	 String *params, Cardinal *num_params)
{
    UNUSED(w);

    if (block_this_mouse_event())
	return;
    
    if (mouse_release != null_mouse && mouse_release != drag_release)
	return;

    if (*num_params != 1)
	return;
    switch (**params) {
    case '|':
	cursor_flags |= DRAG_VERTICAL;
	break;
    case '-':
	cursor_flags |= DRAG_HORIZONTAL;
	break;
    case '+':
	cursor_flags |= DRAG_ALL;
	break;
    default:
	return;
    }
    
    ev_flags |= EV_CURSOR;
    
    if (mouse_release == null_mouse) {
	mouse_motion = drag_motion;
	mouse_release = drag_release;
	drag_last_x = event->xbutton.x_root;
	drag_last_y = event->xbutton.y_root;
    }
    else
	drag_motion(event);

    XFlush(DISP);
}


static void
drag_motion(XEvent * event)
{
#ifdef MOTIF
    get_xy();
#endif

    if (cursor_flags & (DRAG_HORIZONTAL | DRAG_ALL)) { /* horizontal motion */
#ifdef MOTIF
	(void)set_bar_value(x_bar,
			    drag_last_x - event->xbutton.x_root - window_x,
			    (int)(page_w - mane.width));
#else
	if (x_bar != NULL) {
	    XtCallCallbacks(x_bar, XtNscrollProc,
			    (XtPointer) (drag_last_x - event->xbutton.x_root));
	}
#endif
	drag_last_x = event->xbutton.x_root;
    }

    if (cursor_flags & (DRAG_VERTICAL | DRAG_ALL)) { /* vertical motion */
#ifdef MOTIF
	(void)set_bar_value(y_bar,
			    drag_last_y - event->xbutton.y_root - window_y,
			    (int)(page_h - mane.height));
#else
	if (y_bar != NULL) {
	    XtCallCallbacks(y_bar, XtNscrollProc,
			    (XtPointer) (drag_last_y - event->xbutton.y_root));
	}
#endif
	drag_last_y = event->xbutton.y_root;
    }
    handle_x_scroll(NULL, NULL, NULL, NULL);
    handle_y_scroll(NULL, NULL, NULL, NULL);
}


static void
drag_release(XEvent * event)
{
    drag_motion(event);
    mouse_motion = mouse_release = null_mouse;

    cursor_flags &= ~(DRAG_HORIZONTAL | DRAG_VERTICAL | DRAG_ALL);
    ev_flags |= EV_CURSOR;
    XFlush(DISP);
}



/* Wheel mouse support.  */

static int wheel_button = -1;

static void
Act_wheel(Widget w, XEvent *event,
	  String *params, Cardinal *num_params)
{
    int dist;

    UNUSED(w);

    if (block_this_mouse_event())
	return;
    
    if (*num_params == 0) {
	XBell(DISP, 0);
	return;
    }
    dist = (strchr(*params, '.') == NULL) ? atoi(*params)
	: (int)(atof(*params) * resource.wheel_unit);
#ifdef MOTIF
    get_xy();
    set_bar_value(y_bar, dist - window_y, (int)(page_h - mane.height));
#else
    if (y_bar != NULL)
	XtCallCallbacks(y_bar, XtNscrollProc, (XtPointer) dist);
#endif

    wheel_button = event->xbutton.button;

    handle_y_scroll(NULL, NULL, NULL, NULL);
}


/* Internal mouse actions.  */

static void
Act_motion(Widget w, XEvent *event,
	   String *params, Cardinal *num_params)
{
#ifdef HTEX
    /* remember last position, to do this only when pointer is actually moved */
    static int old_x = -1, old_y = -1;
    int x, y;
#endif

    UNUSED(w);
    UNUSED(params);
    UNUSED(num_params);

    if (block_this_mouse_event())
	return;

    if (resource.ruler_mode) {
	show_distance_from_ruler(event, False);
    }
#ifdef HTEX
    /* This used to be:
          (abs(x - old_x) > x_threshold || abs(y - old_y) > y_threshold))
       but that wasn't really convincing. Just change it whenever user
       moves the mouse. */       
    if (!MAGNIFIER_ACTIVE && pointerlocate(&x, &y) && (x != old_x || y != old_y)) {
	htex_displayanchor(x, y);
	old_x = x;
	old_y = y;
    }
#endif
    if ((int)(event->xbutton.button) != wheel_button) {
	mouse_motion(event);
    }
}


static void
Act_release(Widget w, XEvent *event,
	    String *params, Cardinal *num_params)
{
    UNUSED(w);
    UNUSED(params);
    UNUSED(num_params);
    if ((int)(event->xbutton.button) == wheel_button) {
	wheel_button = -1;
	return;
    }

    mouse_release(event);
}

void
Act_ruler_mode(Widget w, XEvent *event,
	       String *params, Cardinal *num_params)
{
    UNUSED(w);
    UNUSED(params);
    UNUSED(num_params);

    if (!toggle_arg(resource.ruler_mode, params, num_params)) {
	return;
    }
    
    resource.ruler_mode = !resource.ruler_mode;
    if (resource.ruler_mode) {
	XDefineCursor(DISP, mane.win, rule_cursor);
	statusline_print(STATUS_SHORT, "Ruler mode on; use Mouse-1 to set/drag ruler");
	show_ruler(event);
	show_distance_from_ruler(event, False);
    }
    else {
	if (cursor_flags & LINK_CURSOR)
	    XDefineCursor(DISP, mane.win, link_cursor);
	else
	    XDefineCursor(DISP, mane.win, ready_cursor);
	statusline_print(STATUS_SHORT, "Ruler mode off");
	clear_ruler();
	mouse_motion = mouse_release = null_mouse;
    }
#if defined(NEW_MENU_CREATION) || defined(MOTIF)
    set_menu(&resource.ruler_mode, Act_ruler_mode, check_toggle);
#else
#ifdef MOTIF
   set_toggle_ruler();
#else
    toggle_menu(resource.ruler_mode, Act_ruler_mode);
#endif
#endif /* NEW_MENU_CREATION */
}


/* Actions for source specials.  */

void
Act_source_special(Widget w, XEvent *event,
		   String *params, Cardinal *num_params)
{
    Window ww;

    UNUSED(w);
    UNUSED(params);
    UNUSED(num_params);

    if (block_this_mouse_event())
	return;
    
    if ((event->type == ButtonPress && mouse_release != null_mouse)
	|| MAGNIFIER_ACTIVE) {
	XBell(DISP, 0);
	return;
    }

    source_reverse_x = event->xbutton.x;
    source_reverse_y = event->xbutton.y;
    if (event->xbutton.window != mane.win)
	(void)XTranslateCoordinates(DISP,
				    RootWindowOfScreen(SCRN), mane.win,
				    event->xbutton.x_root,
				    event->xbutton.y_root,
				    &source_reverse_x,
				    &source_reverse_y,
				    &ww);	/* throw away last argument */

    source_reverse_x = (source_reverse_x + mane_base_x) * mane.shrinkfactor;
    source_reverse_y = (source_reverse_y + mane_base_y) * mane.shrinkfactor;

    ev_flags |= EV_SRC;
}

void
Act_show_source_specials(Widget w, XEvent *event,
			 String *params, Cardinal *num_params)
{
    int arg;

    UNUSED(w);

    if (!get_int_arg(params, num_params, &arg)) {
	arg = -1;
    }

    if ((event->type == ButtonPress && mouse_release != null_mouse)
	|| magnifier.win != (Window) 0) {
	XBell(DISP, 0);
	return;
    }

    if (!(ev_flags & EV_SRC)) {
	source_reverse_x = -1;
	source_show_all = (arg == 1 ? True : False);

	ev_flags |= EV_SRC;
    }
}

void
Act_source_what_special(Widget w, XEvent *event,
			String *params, Cardinal *num_params)
{
    int my_x, my_y;
    Window ww;

    UNUSED(w);
    UNUSED(params);
    UNUSED(num_params);

    if (block_this_mouse_event())
	return;
    
    (void)XTranslateCoordinates(DISP, event->xkey.window, mane.win,
				event->xkey.x, event->xkey.y, &my_x, &my_y, &ww);	/* throw away last argument */
    my_x = (my_x + mane_base_x) * mane.shrinkfactor;
    my_y = (my_y + mane_base_y) * mane.shrinkfactor;
    source_reverse_search(my_x, my_y, False);
}

#ifdef SELFILE
static void
Act_select_dvi_file(Widget w, XEvent *event,
		    String *params, Cardinal *num_params)
{
    UNUSED(w);
    UNUSED(event);
    UNUSED(params);
    UNUSED(num_params);

    if (select_filename(dvi_name, True))
	ev_flags |= EV_NEWDOC;
}
#endif

/*
 * If all GUI elements have been turned on/off, make this synonymous
 * with expert mode off/on, so that next `x' keystroke does something
 * reasonable.
 */
void
check_expert_mode(void)
{
    if ((expert_flag & (XPRT_SHOW_STATUSLINE | XPRT_SHOW_SCROLLBARS
#ifdef MOTIF
			| XPRT_SHOW_PAGELIST | XPRT_SHOW_TOOLBAR | XPRT_SHOW_MENUBAR
#else
			| XPRT_SHOW_BUTTONS
#endif
			)) == XPRT_SHOW_ALL) {
	resource.expert = False;
    }
    else if ((expert_flag & (XPRT_SHOW_STATUSLINE
#ifdef MOTIF
			     | XPRT_SHOW_SCROLLBARS | XPRT_SHOW_PAGELIST | XPRT_SHOW_TOOLBAR | XPRT_SHOW_MENUBAR
#else
			     | BROKEN_RECONFIG ? 0 : XPRT_SHOW_SCROLLBARS | XPRT_SHOW_BUTTONS
#endif
			     )) == XPRT_SHOW_NONE) {
	resource.expert = True;
    }
}

void
Act_set_expert_mode(Widget w, XEvent *event,
		    String *params, Cardinal *num_params)
{
    int arg;

    UNUSED(w);
    UNUSED(event);

    if (!get_int_arg(params, num_params, &arg)) {
	arg = -1;
    }
    
    switch(arg) {
    case 1:
	expert_flag ^= XPRT_SHOW_STATUSLINE;
	toggle_statusline();
	check_expert_mode();
	break;
    case 2:
#ifndef MOTIF
	/* show this warning only when not toggling global expert mode
	   (that's why it can't be inside toggle_scrollbars) */
	if (BROKEN_RECONFIG) {
	    popup_message(MSG_WARN,
			     NULL,
			     "Sorry - cannot toggle scrollbars with this X Version.\n"
			     "This version of XFree has a broken implementation of the viewportWidget, "
			     "which would break the layout if the scrollbars are toggled. "
			     "Versions that are known to work have a VendorRelease version below 4000 or above 4002. "
			     "You will need to update your XFree server to fix this.");
	    return;
	}
#endif
	expert_flag ^= XPRT_SHOW_SCROLLBARS;
	toggle_scrollbars();
	check_expert_mode();
	break;
	
#ifdef MOTIF
    case 3:
	expert_flag ^= XPRT_SHOW_PAGELIST;
	toggle_pagelist();
	check_expert_mode();
	break;
	
    case 4: /* toolbar */
	expert_flag ^= XPRT_SHOW_TOOLBAR;
	toggle_toolbar();
	check_expert_mode();
	break;
	
    case 5:
	expert_flag ^= XPRT_SHOW_MENUBAR;
	toggle_menubar();
	check_expert_mode();
	break;
#else
    case 3:
/*  	fprintf(stderr, "sidebar\n"); */
	expert_flag ^= XPRT_SHOW_BUTTONS;
	toggle_buttons();
	check_expert_mode();
	break;
#endif
    default:
	/* warn 'em */
	if (
#ifdef MOTIF
	    arg > 5
#else
	    arg > 3
#endif
	    ) {
	    statusline_print(STATUS_SHORT, "Number %d too large for `set-expert-mode', using 0 (= toggle) instead.",
			     arg);
	}
	/* toggle all items */
	resource.expert = !resource.expert;
	if (resource.expert)
	    expert_flag = XPRT_SHOW_NONE;
	else
	    expert_flag = XPRT_SHOW_ALL;
	
	toggle_statusline();
#ifndef MOTIF
	if (!BROKEN_RECONFIG)
	    toggle_scrollbars();
#else
	toggle_scrollbars();
#endif
	    
#ifdef MOTIF
	toggle_pagelist();
	toggle_toolbar();
	toggle_menubar();
#else
	toggle_buttons();
#endif
    }
}

Boolean have_src_specials = False;
static Boolean do_update_property = False;

void
handle_x_scroll(Widget w, XtPointer closure, XEvent *ev, Boolean *cont)
{
#ifndef MOTIF
    Dimension get_x = 0;
#endif

    UNUSED(w);
    UNUSED(closure);
    UNUSED(ev);
    UNUSED(cont);
    
    if (!resource.keep_flag || x_bar == NULL)
	return;
    
#ifdef MOTIF
    XtVaGetValues(x_bar, XmNvalue, &m_x_scroll, NULL);
#else
    get_xy();
    XtVaGetValues(clip_widget, XtNx, &get_x, NULL);
    m_x_scroll = get_x - window_x;
#endif
}

void
handle_y_scroll(Widget w, XtPointer closure, XEvent *ev, Boolean *cont)
{
#ifndef MOTIF
    Dimension get_y = 0;
#endif
    
    UNUSED(w);
    UNUSED(closure);
    UNUSED(ev);
    UNUSED(cont);
    
    if (!resource.keep_flag || y_bar == NULL)
	return;
    
#ifdef MOTIF
    XtVaGetValues(y_bar, XmNvalue, &m_y_scroll, NULL);
#else
    get_xy();
    XtVaGetValues(clip_widget, XtNy, &get_y, NULL);
    m_y_scroll = get_y - window_y;
#endif
}

void
handle_expose(Widget w, XtPointer closure, XEvent *ev, Boolean *cont)
{
    struct WindowRec *windowrec = (struct WindowRec *)closure;

    UNUSED(w);
    UNUSED(cont);

    if (windowrec == &magnifier) {
	if (magnifier_stat < 0) { /* destroy upon exposure */
	    magnifier_stat = 0;
	    mag_release(ev);
	    return;
	}
	else
	    magnifier_stat = 0;
    }

    do_update_property = False;

    if (have_src_specials && !MAGNIFIER_ACTIVE) {
	do_update_property = True;
    }

    expose(windowrec, (&(ev->xexpose))->x, (&(ev->xexpose))->y,
	   (unsigned int)(&(ev->xexpose))->width, (unsigned int)(&(ev->xexpose))->height);
}


void
handle_property_change(Widget w, XtPointer junk,
		       XEvent *ev, Boolean *cont)
{
    char *prop_ret;
    size_t prop_len;

    UNUSED(w);
    UNUSED(junk);
    UNUSED(cont);
    
    if ((&(ev->xproperty))->window != XtWindow(top_level)) /* if spurious event */
	return;
    
    if ((&(ev->xproperty))->atom == atom_src_goto()) {
	/* forward search requested */
	if ((prop_len = property_get_data(XtWindow(top_level), atom_src_goto(),
					  (unsigned char **)&prop_ret,
					  XGetWindowProperty)) == 0) {
	    TRACE_CLIENT((stderr, "property_get_data() failed for atom_src_goto()!"));
	    return;
	}
	TRACE_CLIENT((stderr, "got back atom_src_goto: |%s|", prop_ret));
	source_forward_string = prop_ret;
	ev_flags |= EV_SRC;
    }
    else if ((&(ev->xproperty))->atom == atom_newdoc()) {
	/* loading a new file */
	FILE *new_fp;
	if ((prop_len = property_get_data(XtWindow(top_level), atom_newdoc(),
					  (unsigned char **)&prop_ret,
					  XGetWindowProperty)) == 0) {
	    TRACE_CLIENT((stderr, "property_get_data() returned zero length for atom_newdoc()"));
	    /* just raise it in this case */
	    XMapRaised(XtDisplay(top_level), XtWindow(top_level));
	    return;
	}
	TRACE_CLIENT((stderr, "got back atom_newdoc: |%s|", prop_ret));
	if ((new_fp = XFOPEN(prop_ret, "r")) == NULL) {
	    popup_message(MSG_ERR, NULL, "Loading %s failed: %s",
			     prop_ret, strerror(errno));
	    return;
	}
	set_dvi_name(prop_ret);
	ev_flags |= EV_NEWDOC;
    }
    else if ((&(ev->xproperty))->atom == atom_newpage()) {
	/* jumping to a different page */
	int newpage;
	char *testptr;
	if ((prop_len = property_get_data(XtWindow(top_level), atom_newpage(),
					  (unsigned char **)&prop_ret,
					  XGetWindowProperty)) == 0) {
	    TRACE_CLIENT((stderr, "property_get_data() failed for atom_newpage(): |%s|", prop_ret));
	    return;
	}
	TRACE_CLIENT((stderr, "got back atom_newpage: |%s|", prop_ret));
	if (strcmp(prop_ret, "+") == 0) { /* special case: treat `+' as last page */
	    newpage = total_pages - 1;
	}
	else {
	    newpage = strtol(prop_ret, &testptr, 10) - 1;
	    if (*testptr != '\0') {
		XDVI_FATAL((stderr, "Invalid page number: `%s'.", prop_ret));
	    }
	}

	if (newpage == total_pages - 1) { /* as in Act_goto_page() */
	    goto_page(check_goto_page(newpage), resource.keep_flag ? NULL : home);
	    search_signal_page_changed();
	}
	else {
	    goto_page(check_goto_tex_page(newpage), resource.keep_flag ? NULL : home);
	    search_signal_page_changed();
	}
    }
    else if ((&(ev->xproperty))->atom == atom_raise()) {
	XMapRaised(XtDisplay(top_level), XtWindow(top_level));
    }
}

/*
 *	Interrupt system for receiving events.  The program sets a flag
 *	whenever an event comes in, so that at the proper time (i.e., when
 *	reading a new dvi item), we can check incoming events to see if we
 *	still want to go on printing this page.  This way, one can stop
 *	displaying a page if it is about to be erased anyway.  We try to read
 *	as many events as possible before doing anything and base the next
 *	action on all events read.
 *	Note that the Xlib and Xt routines are not reentrant, so the most we
 *	can do is set a flag in the interrupt routine and check it later.
 *	Also, sometimes the interrupts are not generated (some systems only
 *	guarantee that SIGIO is generated for terminal files, and on the system
 *	I use, the interrupts are not generated if I use "(xdvi foo &)" instead
 *	of "xdvi foo").  Therefore, there is also a mechanism to check the
 *	event queue every 70 drawing operations or so.  This mechanism is
 *	disabled if it turns out that the interrupts do work.
 *	For a fuller discussion of some of the above, see xlife in
 *	comp.sources.x.
 */

/*
 *	Signal flags
 */

/* This could be static volatile, but we want to avoid any optimizer bugs.  */
VOLATILE unsigned int sig_flags	= 0;

#define	SF_USR	1
#define	SF_ALRM	2
#define	SF_POLL	4
#define	SF_CHLD	8
#define	SF_TERM	16
#define	SF_SEGV	32

static void do_sigusr(void);
static void do_sigalrm(void);
static void do_sigpoll(void);
static void do_sigchld(void);
static void do_sigterm(void);
static void do_sigsegv(void);

/* these must be in the same order as SF_*.
   The higher flags have higher priority, since these are
   checked first when resolving flags_to_sigproc[sig_flags].

   Example: flag := SF_TERM | SF_CHLD = 24
   
   flags_to_sigproc[24] =>
       do_sigterm, which sets flag to 24 & ~SF_TERM == 8
   then, in next check:       
   flags_to_sigproc[8] =>
       do_sigchld, which sets flag 8 & ~SF_CHLD == 0.
 */
#define	SP0	do_sigusr
#define	SP1	do_sigalrm
#define	SP2	do_sigpoll
#define	SP3	do_sigchld
#define	SP4	do_sigterm
#define	SP5	do_sigsegv

typedef	void	(*signalproc)(void);

static	_Xconst signalproc flags_to_sigproc[64] = {
    NULL,
    /* 1 */
    SP0,
    /* 2 - 3 */
    SP1, SP1,
    /* 4 - 7 */
    SP2, SP2, SP2, SP2,
    /* 8 - 15 */
    SP3, SP3, SP3, SP3, SP3, SP3, SP3, SP3,
    /* 16 - 31 */
    SP4, SP4, SP4, SP4, SP4, SP4, SP4, SP4,
    SP4, SP4, SP4, SP4, SP4, SP4, SP4, SP4,
    /* 32 - 63 */
    SP5, SP5, SP5, SP5, SP5, SP5, SP5, SP5,
    SP5, SP5, SP5, SP5, SP5, SP5, SP5, SP5,
    SP5, SP5, SP5, SP5, SP5, SP5, SP5, SP5,
    SP5, SP5, SP5, SP5, SP5, SP5, SP5, SP5
};

#undef	SP0
#undef	SP2
#undef	SP3
#undef	SP4
#undef	SP5

/*
 *	Signal routines.  At the signal level, all we do is set flags.
 */

#ifndef FLAKY_SIGPOLL
static RETSIGTYPE
handle_sigpoll(int signo)
{
    UNUSED(signo);
    
    event_counter = 1;
    event_freq = -1;	/* forget Plan B */
    sig_flags |= SF_POLL;
# ifndef HAVE_SIGACTION
    (void) signal(SIGPOLL, handle_sigpoll);	/* reset the signal */
# endif
}
#endif

static RETSIGTYPE
handle_sigterm(int signo)
{
    UNUSED(signo);
    
    sig_flags |= SF_TERM;
}

static RETSIGTYPE
handle_sigchld(int signo)
{
    UNUSED(signo);

    sig_flags |= SF_CHLD;
}

static RETSIGTYPE
handle_sigalrm(int signo)
{
    UNUSED(signo);

    sig_flags |= SF_ALRM;
}

static RETSIGTYPE
handle_sigusr(int signo)
{
    UNUSED(signo);
    
    event_counter = 1;
    sig_flags |= SF_USR;
}

static RETSIGTYPE
handle_sigsegv(int signo)
{
    UNUSED(signo);
    
    XDVI_ABORT((stderr, "Segmentation fault - trying to clean up and aborting ..."));
}

static Boolean sigalarm_initialized = False;

void
setup_sigalarm(void)
{
#if HAVE_SIGACTION
    struct sigaction a;

    a.sa_handler = handle_sigalrm;
    (void) sigemptyset(&a.sa_mask);
    (void) sigaddset(&a.sa_mask, SIGALRM);
    a.sa_flags = 0;
    sigaction(SIGALRM, &a, NULL);
#else /* not HAVE_SIGACTION */
    (void) signal(SIGALRM, handle_sigalrm);
#endif /* not HAVE_SIGACTION */

    sigalarm_initialized = True;
}

/*
 * Called from main to set up the signal handlers.
 */
void
setup_signal_handlers(void)
{
#ifndef FLAKY_SIGPOLL
    int	sock_fd	= ConnectionNumber(DISP);
#endif
#if HAVE_SIGACTION
    struct sigaction a;
#endif

#ifndef FLAKY_SIGPOLL
#if HAVE_SIGACTION
    /* Subprocess handling, e.g., MakeTeXPK, fails on the Alpha without
       this, because SIGPOLL interrupts the call of system(3), since OSF/1
       doesn't retry interrupted wait calls by default.  From code by
       maj@cl.cam.ac.uk.  */
    a.sa_handler = handle_sigpoll;
    (void) sigemptyset(&a.sa_mask);
    (void) sigaddset(&a.sa_mask, SIGPOLL);
    a.sa_flags = SA_RESTART;
    sigaction(SIGPOLL, &a, NULL);
#else /* not HAVE_SIGACTION */
    (void) signal(SIGPOLL, handle_sigpoll);
#endif /* not HAVE_SIGACTION */

    prep_fd(sock_fd, False);
#endif	/* not FLAKY_SIGPOLL */

#if HAVE_SIGACTION
    a.sa_handler = handle_sigterm;
    (void) sigemptyset(&a.sa_mask);
    (void) sigaddset(&a.sa_mask, SIGINT);
    (void) sigaddset(&a.sa_mask, SIGQUIT);
    (void) sigaddset(&a.sa_mask, SIGTERM);
    (void) sigaddset(&a.sa_mask, SIGHUP);
    a.sa_flags = SA_RESETHAND;
    sigaction(SIGINT, &a, NULL);
    sigaction(SIGQUIT, &a, NULL);
    sigaction(SIGTERM, &a, NULL);
    sigaction(SIGHUP, &a, NULL);
    a.sa_handler = handle_sigsegv;
    (void)sigemptyset(&a.sa_mask);
    (void)sigaddset(&a.sa_mask, SIGSEGV);
    a.sa_flags = 0;
    sigaction(SIGSEGV, &a, NULL);
#else /* not HAVE_SIGACTION */
    (void) signal(SIGINT, handle_sigterm);
    (void) signal(SIGQUIT, handle_sigterm);
    (void) signal(SIGTERM, handle_sigterm);
    (void) signal(SIGHUP, handle_sigterm);
    (void)signal(SIGSEGV, handle_sigsegv);
#endif /* not HAVE_SIGACTION */

#if HAVE_SIGACTION
    a.sa_handler = handle_sigchld;
    (void) sigemptyset(&a.sa_mask);
    (void) sigaddset(&a.sa_mask, SIGCHLD);
    a.sa_flags = 0;
    sigaction(SIGCHLD, &a, NULL);
#else /* not HAVE_SIGACTION */
    (void) signal(SIGCHLD, handle_sigchld);
#endif /* not HAVE_SIGACTION */

#if HAVE_SIGACTION
    a.sa_handler = handle_sigusr;
    (void) sigemptyset(&a.sa_mask);
    (void) sigaddset(&a.sa_mask, SIGUSR1);
    a.sa_flags = 0;
    sigaction(SIGUSR1, &a, NULL);
#else /* not HAVE_SIGACTION */
    (void) signal(SIGUSR1, handle_sigusr);
#endif /* not HAVE_SIGACTION */

    (void)sigemptyset(&all_signals);
    (void)sigaddset(&all_signals, SIGPOLL);
    (void)sigaddset(&all_signals, SIGINT);
    (void)sigaddset(&all_signals, SIGQUIT);
    (void)sigaddset(&all_signals, SIGTERM);
    (void)sigaddset(&all_signals, SIGHUP);
    (void)sigaddset(&all_signals, SIGCHLD);
    (void)sigaddset(&all_signals, SIGALRM);
    (void)sigaddset(&all_signals, SIGUSR1);
    (void)sigaddset(&all_signals, SIGSEGV);

}


/*
 *	Mid-level signal handlers.  These are called from within read_events(),
 *	and do the actual work appropriate for the given signal.
 */

/*
 *	Process-related routines.  Call set_chld(xchild *) to indicate
 *	that a given child process should be watched for when it
 *	terminates.  Call clear_chld() to remove the process from the
 *	list.  When the child terminates, the record is removed from
 *	the list and xchild->proc is called.
 *	The caller can set the `io' member in xchild to a pointer to
 *	an xio structure for reading from the file descriptor; see
 *	psgs.c and util.c for examples.
 */

static struct xchild *child_recs = NULL;	/* head of child process list */

void
set_chld(struct xchild *cp)
{
    cp->next = child_recs;
    child_recs = cp;
}

void
clear_chld(struct xchild *cp)
{
    struct xchild	**cpp;
    struct xchild	*cp2;

    if (child_recs == NULL) {
	if (debug & DBG_EVENT)
	    fprintf(stderr, "child_recs: %p\n", (void *)child_recs);
	return;
    }
    for (cpp = &child_recs;;) {
	cp2 = *cpp;
	if (cp2 == cp)
	    break;
	cpp = &cp2->next;
    }
    *cpp = cp->next;
}

static void
do_sigchld(void)
{
    pid_t	pid;
    int	status;

    sig_flags &= ~SF_CHLD;

#if ! HAVE_SIGACTION
    (void) signal(SIGCHLD, handle_sigchld);	/* reset the signal */
#endif
    for (;;) {
#if HAVE_WAITPID
	pid = waitpid(-1, &status, WNOHANG);
#else
	pid = wait3(&status, WNOHANG, NULL);
#endif
	if (pid == 0) break;

	if (pid != -1) {
	    struct xchild	**cpp;
	    struct xchild	*cp;

	    for (cpp = &child_recs;;) {
		cp = *cpp;
		if (cp == NULL)
		    break;
		if (cp->pid == pid) {
		    *cpp = cp->next;	/* unlink it */
		    /* call exit reporting procedure for this child */
		    (cp->proc)(status, cp);
		    break;
		}
		cpp = &cp->next;
	    }
	    break;
	}

	if (errno == EINTR) continue;
	if (errno == ECHILD) break;
#if HAVE_WAITPID
	perror("xdvi: waitpid");
#else
	perror("xdvi: wait3");
#endif
	break;
    }
}


/*
 *	File-related routines.  Call set_io() to indicate that a given fd
 *	should be watched for ability to input or output data.  Call clear_io()
 *	to remove it from the list.  When poll()/select() indicates that the fd
 *	is available for the indicated type of i/o, the corresponding routine
 *	is called.  Call clear_io() to remove an fd from the list.
 *	Both set_io() and clear_io() can be called from within read_proc or
 *	write_proc (although turning an io descriptor on or off is better
 *	accomplished by setting the events flag in the xio structure, and
 *	in the corresponding pollfd structure if the pfd pointer is not NULL
 *	(it is always non-NULL when read_proc and write_proc are called)).
 *	We allocate space for one additional record in the pollfd array, to
 *	accommodate the fd for the X connection; this is done by initializing
 *	num_fds to 1 instead of zero.
 */

static	struct xio	*iorecs	= NULL;	/* head of xio list */

#if HAVE_POLL
static	struct pollfd	*fds	= NULL;
static	int		num_fds	= 1;	/* current number of fds */
static	int		max_fds	= 0;	/* max allocated number of fds */
static	Boolean		io_dirty= True;	/* need to recompute fds[] array */
#else
static	int		numfds	= 0;
static	fd_set		readfds;
static	fd_set		writefds;
#endif

void
set_io(struct xio *ip)
{
    ip->next = iorecs;
    iorecs = ip;

#if HAVE_POLL
    ++num_fds;
    if (!io_dirty && num_fds <= max_fds) {
	fds[num_fds - 1].fd = ip->fd;
	fds[num_fds - 1].events = ip->xio_events;
	ip->pfd = &fds[num_fds - 1];
    }
    else {
	ip->pfd = NULL;
	io_dirty = True;
    }
#else
    if (numfds <= ip->fd) numfds = ip->fd + 1;
#endif
}

void
clear_io(struct xio *ip)
{
    struct xio	**ipp;

    for (ipp = &iorecs;;) {
	struct xio *ip2;

	ip2 = *ipp;
	if (ip2 == ip) break;
	ipp = &ip2->next;
    }
    *ipp = ip->next;

#if HAVE_POLL
    --num_fds;
    io_dirty = True;
#else
# if FLAKY_SIGPOLL
    numfds = ConnectionNumber(DISP);
# else
    numfds = (event_freq < 0 ? -1 : ConnectionNumber(DISP));
# endif
    for (ip = iorecs; ip != NULL; ip = ip->next)
	if (ip->fd > numfds)
	    numfds = ip->fd;
    ++numfds;
#endif /* !HAVE_POLL */
}

static void
do_sigpoll(void)
{
#if FLAKY_SIGPOLL
    sig_flags &= ~SF_POLL;
#else
    struct xio	*ip;

    sig_flags &= ~SF_POLL;

# if HAVE_POLL

    if (io_dirty) {
	struct pollfd *fp;

	if (num_fds > max_fds) {
	    if (fds != NULL) free(fds);
	    fds = xmalloc(num_fds * sizeof *fds);
	    max_fds = num_fds;
	    fds->fd = ConnectionNumber(DISP);
	    fds->events = POLLIN;
	}
	fp = fds + 1;
	for (ip = iorecs; ip != NULL; ip = ip->next) {
	    fp->fd = ip->fd;
	    fp->events = ip->xio_events;
	    ip->pfd = fp;
	    ++fp;
	}
	io_dirty = False;
    }

    for (;;) {
	if (poll(fds + 1, num_fds - 1, 0) >= 0)
	    break;

	if (errno != EAGAIN && errno != EINTR) {
	    perror("xdvi: poll");
	    return;
	}
    }

    for (ip = iorecs; ip != NULL; ip = ip->next) {
	int revents = ip->pfd->revents;
	if (revents & POLLIN && ip->read_proc != NULL)
	    (void)(ip->read_proc)(ip->fd);
	if (revents & POLLOUT && ip->write_proc != NULL)
	    (ip->write_proc)(ip->fd);
    }

# else

    FD_ZERO(&readfds);
    FD_ZERO(&writefds);
    for (ip = iorecs; ip != NULL; ip = ip->next) {
	if (ip->xio_events & XIO_IN)
	    FD_SET(ip->fd, &readfds);
	if (ip->xio_events & XIO_OUT)
	    FD_SET(ip->fd, &writefds);
    }

    for (;;) {
	struct timeval tv;

	tv.tv_sec = tv.tv_usec = 0;
	if (select(numfds, &readfds, &writefds, (fd_set *) NULL, &tv) >= 0)
	    break;

	if (errno != EAGAIN && errno != EINTR) {
	    perror("select (xdvi read_events)");
	    return;
	}
    }

    for (ip = iorecs; ip != NULL; ip = ip->next) {
	if (FD_ISSET(ip->fd, &readfds) && ip->read_proc != NULL)
	    (void)(ip->read_proc)(ip->fd);
	if (FD_ISSET(ip->fd, &writefds) && ip->write_proc != NULL)
	    (ip->write_proc)(ip->fd);
    }

# endif
#endif /* not FLAKY_SIGPOLL */
}


/*
 *	Timer-related routines.  Call set_timer() to set a timer a given number
 *	of milliseconds in the future.  At that time, the timer will be cleared
 *	and the given procedure will be called with argument set to the struct
 *	passed to set_timer().  The timer routine may call set_timer() or
 *	cancel_timer().
 */


static	struct xtimer		*timers	= NULL;	/* head of timer list */

static	struct itimerval	itv	= {{0, 0}, {0, 0}};

#ifndef	timercmp
#define	timercmp(a, b, cmp)	((a)->tv_sec cmp (b)->tv_sec || \
	((a)->tv_sec == (b)->tv_sec && (a)->tv_usec cmp (b)->tv_usec))
#endif	/* timercmp */


static void
show_timers(char *what)
{
    struct xtimer *tp;
    fprintf(stderr, "=======%s; timers:\n", what);
    for (tp = timers; tp != NULL; tp = tp->next) {
	fprintf(stderr, "timer %p: %lu\n", (void *)tp, tp->when.tv_sec);
    }
    fprintf(stderr, "=======\n");
}

void
set_timer(struct xtimer *tp, int ms)
{
    struct xtimer	**tpp;
    struct xtimer	*tp2;

    if (debug & DBG_EVENT)
	fprintf(stderr, "%s:%d: set_timer\n", __FILE__, __LINE__);

    gettimeofday(&tp->when, NULL);
    itv.it_value.tv_sec = ms / 1000;
    itv.it_value.tv_usec = (ms % 1000) * 1000;
    tp->when.tv_sec += itv.it_value.tv_sec;
    tp->when.tv_usec += itv.it_value.tv_usec;
    if (tp->when.tv_usec >= 1000000) {
	tp->when.tv_usec -= 1000000;
	++tp->when.tv_sec;
    }

    for (tpp = &timers;;) {		/* add timer to list */
	tp2 = *tpp;
	if (tp2 == NULL || timercmp(&tp->when, &tp2->when, <))
	    break;
	tpp = &tp2->next;
    }
    tp->next = tp2;
    *tpp = tp;

    if (tpp == &timers) {
	setitimer(ITIMER_REAL, &itv, NULL);
	if (ms == 0)
	    sig_flags |= SF_ALRM;
    }
    if (debug & DBG_EVENT)
	show_timers("after set_timer");
}

void
cancel_timer(struct xtimer *tp)
{
    struct xtimer	**tpp;

    if (debug & DBG_EVENT)
	show_timers("beginning of cancel_timer");
    
    if (timers == NULL) {
	fprintf(stderr, "%s:%d: BUG? timers == NULL!\n", __FILE__, __LINE__);
	return;
    }

    if (debug & DBG_EVENT)
	fprintf(stderr, "%s:%d: cancel_timer %p from %p\n", __FILE__, __LINE__, &timers, (void *)tp);

    ASSERT(timers != NULL, "timers in cancel_timer() mustn't be NULL");
    for (tpp = &timers; ; ) {		/* remove from list */
	
	if (*tpp == tp)
	    break;
	tpp = &(*tpp)->next;
    }

    *tpp = (*tpp)->next;	/* unlink it */

    if (timers == NULL) {	/* cancel SIGALRM */
	itv.it_value.tv_sec = itv.it_value.tv_usec = 0;
	setitimer(ITIMER_REAL, &itv, NULL);
    }
}

#ifdef TIMER_HACK
/*
 * Original comment by Paul:
 *	Newer versions of the Motif toolkit use the timer facility
 *	(XtAppAddTimeOut(), etc.) in the X Toolkit.  Proper functioning of
 *	this mechanism, however, requires that the X Toolkit be in charge of
 *	blocking.  Since xdvi does its own blocking, this means that we need
 *	to provide working alternatives to these X Toolkit routines (by
 *	redefining them ...).
 *	One symptom of the above-mentioned bug is that the printlog window
 *	eventually stops showing dvips progress until you move the mouse.
 *
 * Comment SU:
 *	Xdvik also uses XtAppAddTimeOut() in other places (hyperref/statusline/...),
 *	so we also need these redefinitions also in the Xaw version.
 *
 */

static void xt_alarm (struct xtimer *);

static struct xtimer *xt_free_timers = NULL;

XtIntervalId
XtAppAddTimeOut(XtAppContext app, unsigned long interval, XtTimerCallbackProc proc, XtPointer closure)
{
    struct xtimer *tp;
    
    UNUSED(app);

    /* FIXME: better way of checking this instead of static boolean sigalarm_initialized?
       The following doesn't work, even after
       sigaddset(&all_signals, SIGALRM);
       
          static sigset_t sig_set;
          (void)sigprocmask(0, NULL, &sig_set);
	  if (sigismember(&sig_set, SIGALRM))
	      ... OK ...
	  else
	      ... NOT OK ...
    */
    ASSERT(sigalarm_initialized, "Shouldn't invoke XtAppAddTimeOut() before setup_sigalarm()");
    
    if (debug & DBG_EVENT)
	fprintf(stderr, "XtAppAddTimeOut: %lu msecs\n", interval);
    
    if (xt_free_timers == NULL)
	tp = xmalloc(sizeof *tp);
    else {
	tp = xt_free_timers;
	xt_free_timers = xt_free_timers->next;
    }

    tp->proc = xt_alarm;
    tp->xt_proc = proc;
    tp->closure = closure;
    
    set_timer(tp, interval);

    if (debug & DBG_EVENT)
	show_timers("XtAppAddTimeOut");

    return (XtIntervalId)tp;
}

void
XtRemoveTimeOut(XtIntervalId id)
{
    struct xtimer *tp;

    ASSERT(id != 0, "XtIntervalId argument in XtRemoveTimeOut() mustn't be NULL");
    tp = (struct xtimer *)id;

    /* Motif (2.1 on Solaris 9, 2003) sometimes calls XtRemoveTimeOut() after
       the timer event has occurred, so we need to be sure not to remove
       the timer record twice.  */
    if (tp->proc == NULL)
	return;
    
    cancel_timer(tp);
    
    tp->next = xt_free_timers;
    xt_free_timers = tp;

    if (debug & DBG_EVENT)
	show_timers("XtRemoveTimeOut");
}

void
xt_alarm(struct xtimer *tp)
{
    XtIntervalId id;

    tp->proc = NULL;	/* flag timer as used-up */
    id = (XtIntervalId) tp;
    (tp->xt_proc)(tp->closure, &id);
    
    tp->next = xt_free_timers;
    xt_free_timers = tp;

    if (debug & DBG_EVENT)
	show_timers("xt_alarm");
}
#endif /* TIMER_HACK */

static	void
do_sigalrm(void)
{
    struct timeval	now;

    sig_flags &= ~SF_ALRM;
#ifndef HAVE_SIGACTION
    (void) signal(SIGALRM, handle_sigalrm);	/* reset the signal */
#endif

    gettimeofday(&now, NULL);

    while (timers != NULL && timercmp(&timers->when, &now, <=)) {
	struct xtimer *tp = timers;
	timers = timers->next;	/* unlink it _first_ */
	(tp->proc)(tp);
    }

    if (timers != NULL) {		/* set next timer */
	int i;
	itv.it_value.tv_sec = timers->when.tv_sec - now.tv_sec;
	i = timers->when.tv_usec - now.tv_usec;
	if (i < 0) {
	    --itv.it_value.tv_sec;
	    i += 1000000;
	}
	itv.it_value.tv_usec = i;

	setitimer(ITIMER_REAL, &itv, NULL);
    }
}



/*
 *	Handle SIGUSR1 signal.  Pretty straightforward.
 */

static void
do_sigusr(void)
{
    sig_flags &= ~SF_USR;
#ifndef HAVE_SIGACTION
    (void) signal(SIGUSR1, handle_sigusr);	/* reset the signal */
#endif
    ev_flags |= EV_RELOAD;
}


static void
do_sigsegv(void)
{
    sig_flags &= ~SF_SEGV;
#ifndef HAVE_SIGACTION
    (void) signal(SIGSEGV, handle_sigsegv);	/* reset the signal */
#endif
    handle_sigsegv(SIGSEGV);
}


/*
 *	Handle termination signals.  Kill child processes, if permitted.
 *	Otherwise, leave it up to the caller.  This is the only place where
 *	the EV_TERM event flag is set, and it can only happen if there's an
 *	active non-killable process.  This should only happen if read_events()
 *	is called from one of a very few select locations.
 */

static void
do_sigterm(void)
{
    struct xchild   *cp;
    Boolean	    all_killed	= True;

    sig_flags &= ~SF_TERM;

    /* loop through child processes */
    for (cp = child_recs; cp != NULL; cp = cp->next)
	if (cp->killable)
	    kill(cp->pid, SIGKILL);
	else
	    all_killed = False;

    if (all_killed)		/* if all processes killed */
	xdvi_exit(EXIT_SUCCESS);
    /* SU: added following else case, which is needed to allow children
       forked via fork_process() continue running, while still being
       able to exit xdvi.
     */
    else
	xdvi_exit(EXIT_SUCCESS);
    ev_flags |= EV_TERM;	/* otherwise, let the caller handle it */
}

/*
 *	Set the flag so that termination occurs, via the above routine.
 *	This should be used in place of xdvi_exit() when there may be a
 *	non-killable process running (e.g., anytime within read_events()).
 */

static void
xdvi_normal_exit(void)
{
    sig_flags |= SF_TERM;
}


/*
 *	Since redrawing the screen is (potentially) a slow task, xdvi checks
 *	for incoming events while this is occurring.  It does not register
 *	a work proc that draws and returns every so often, as the toolkit
 *	documentation suggests.  Instead, it checks for events periodically
 *	(or not, if SIGPOLL can be used instead) and processes them in
 *	a subroutine called by the page drawing routine.  This routine (below)
 *	checks to see if anything has happened and processes those events and
 *	signals.  (Or, if it is called when there is no redrawing that needs
 *	to be done, it blocks until something happens.)
 *
 *	Ultimately, the goal is to have this be the only place in xdvi where
 *	blocking occurs.
 *
 *	The argument to this function should be a mask of event types (EV_*)
 *	indicating which event types should cause read_events to return instead
 *	of waiting for more events.  This function will always process all
 *	pending events and signals before returning.
 *	The return value is the value of ev_flags.
 */

unsigned int
read_events(unsigned int ret_mask)
{
    XEvent event;

#if !HAVE_POLL
    if (numfds == 0) numfds = ConnectionNumber(DISP) + 1;
#endif

    if (debug & DBG_EVENT)
	fprintf(stderr, "%s:%d: read_events %u\n", __FILE__, __LINE__, ret_mask);
    for (;;) {
	event_counter = event_freq;
	/*
	 * The above line clears the flag indicating that an event is
	 * pending.  So if an event comes in right now, the flag will be
	 * set again needlessly, but we just end up making an extra call.
	 * Also, be careful about destroying the magnifying glass while
	 * drawing on it.
	 */

#if !FLAKY_SIGPOLL

	if (event_freq < 0) {	/* if SIGPOLL works */
	    if (!XtPending()) {
		sigset_t oldsig;

		(void) sigprocmask(SIG_BLOCK, &all_signals, &oldsig);
		for (;;) {
#ifdef SHOW_SIG_FLAGS
		    /* this gives HUGE output ... */
		    if (debug & DBG_EVENT)
			fprintf(stderr, "%s:%d: sig_flags = %d\n",
				__FILE__, __LINE__, sig_flags);
#endif
		    while (sig_flags) {
			flags_to_sigproc[sig_flags]();
		    }

		    if (XtPending())
			break;

		    if (ev_flags & ret_mask) {
			(void) sigprocmask(SIG_SETMASK, &oldsig, (sigset_t *) NULL);
			return ev_flags;
		    }
		    (void) sigsuspend(&oldsig);
		}
		(void) sigprocmask(SIG_SETMASK, &oldsig, (sigset_t *) NULL);
	    }
	}
	else

#endif /* not FLAKY_SIGPOLL */

	    {
		for (;;) {
		    struct xio	*ip;

		    if (debug & DBG_EVENT)
			fprintf(stderr, "%s:%d: (flaky) sig_flags = %d\n",
				__FILE__, __LINE__, sig_flags);
		    while (sig_flags) {
			sigset_t oldsig;

			(void) sigprocmask(SIG_BLOCK, &all_signals, &oldsig);

			while (sig_flags) {
			    flags_to_sigproc[sig_flags]();
			}

			(void) sigprocmask(SIG_SETMASK, &oldsig,
					   (sigset_t *) NULL);
		    }

		    if (XtPending())
			break;

		    if (ev_flags & ret_mask)
			return ev_flags;

		    /* If a SIGUSR1 signal comes right now, then it will wait
		       until an X event or another SIGUSR1 signal arrives. */

#if HAVE_POLL
		    if (debug & DBG_EVENT)
			fprintf(stderr, "%s:%d: have_poll!\n",
				__FILE__, __LINE__);
		    if (io_dirty) {
			struct pollfd *fp;

			if (num_fds > max_fds) {
			    if (fds != NULL) free(fds);
			    fds = xmalloc(num_fds * sizeof *fds);
			    max_fds = num_fds;
			    fds->fd = ConnectionNumber(DISP);
			    fds->events = POLLIN;
			}
			fp = fds + 1;
			for (ip = iorecs; ip != NULL; ip = ip->next) {
			    fp->fd = ip->fd;
			    fp->events = ip->xio_events;
			    ip->pfd = fp;
			    ++fp;
			}
			io_dirty = False;
		    }

		    for (;;) {
			if (poll(fds, num_fds, -1) >= 0) {
			    for (ip = iorecs; ip != NULL; ip = ip->next) {
				int revents = ip->pfd->revents;

				if (revents & POLLIN && ip->read_proc != NULL)
				    (ip->read_proc)(ip->fd);
				if (revents & POLLOUT && ip->write_proc != NULL)
				    (ip->write_proc)(ip->fd);
			    }
			    break;
			}

			if (errno == EINTR)
			    break;

			if (errno != EAGAIN) {
			    perror("xdvi: poll");
			    break;
			}
		    }
#else /* HAVE_POLL */
		    if (debug & DBG_EVENT)
			fprintf(stderr, "%s:%d: NOT have_poll!\n",
				__FILE__, __LINE__);
		    FD_ZERO(&readfds);
		    FD_ZERO(&writefds);
		    FD_SET(ConnectionNumber(DISP), &readfds);
		    for (ip = iorecs; ip != NULL; ip = ip->next) {
			if (ip->xio_events & XIO_IN)
			    FD_SET(ip->fd, &readfds);
			if (ip->xio_events & XIO_OUT)
			    FD_SET(ip->fd, &writefds);
		    }

		    for (;;) {
			if (select(numfds, &readfds, &writefds, (fd_set *) NULL,
				   (struct timeval *) NULL) >= 0) {
			    for (ip = iorecs; ip != NULL; ip = ip->next) {
				if (FD_ISSET(ip->fd, &readfds) && ip->read_proc != NULL) {
				    if (debug & DBG_EVENT)
					fprintf(stderr, "%s:%d: reading from %d\n",
						__FILE__, __LINE__, ip->fd);
				    (ip->read_proc)(ip->fd);
				}
				if (FD_ISSET(ip->fd, &writefds) && ip->write_proc != NULL) {
				    if (debug & DBG_EVENT)
					fprintf(stderr, "%s:%d: writing to %d\n",
						__FILE__, __LINE__, ip->fd);
				    (ip->write_proc)(ip->fd);
				}
			    }
			    break;
			}

			if (errno == EINTR)
			    break;

			if (errno != EAGAIN) {
			    perror("xdvi: select");
			    break;
			}
		    }
#endif /* HAVE_POLL */
		}
	    }
	XtAppNextEvent(app, &event);
#ifdef MOTIF
	if ((expert_flag & XPRT_SHOW_TOOLBAR) != 0)
	    TipAppHandle(app, &event);
#endif

	if (resized)
	    get_geom();

	if (event.xany.window == magnifier.win && event.type == Expose) {
	    handle_expose((Widget) NULL, (XtPointer) &magnifier, &event,
			  (Boolean *) NULL);
	    continue;
	}

#ifdef MOTIF
	if (event.xany.window == mane.win && event.type == EnterNotify) { /* kludge for #610206 */
	    static Boolean first_time = True;
	    if (first_time) {
		if (XtIsRealized(top_level)) {
		    /* top_level might not be realized yet when xdvi
		       starts up with no filename, and user selects
		       file via file dialog */
		    motif_translations_kludge();
		    first_time = False;
		}
	    }
	}
#endif
	XtDispatchEvent(&event);
    }
}


/*
 * Higher-level routines for managing events.
 */

static void
can_exposures(struct WindowRec *windowrec)
{
    windowrec->min_x = windowrec->min_y = MAXDIM;
    windowrec->max_x = windowrec->max_y = 0;
}

void
redraw(struct WindowRec *windowrec)
{
    currwin = *windowrec;
    min_x = currwin.min_x + currwin.base_x;
    min_y = currwin.min_y + currwin.base_y;
    max_x = currwin.max_x + currwin.base_x;
    max_y = currwin.max_y + currwin.base_y;
    can_exposures(windowrec);
    
    /* fix for bug #619070 - the complicated flags (and do_update_property)
       are needed to avoid too many updates at exposures, especially for
       a window of another xdvi instance when the magnifier intersects with
       that window.
    */
    if (have_src_specials && do_update_property
	&& min_x != 1 && max_y - min_y != 1
	&& currwin.base_x == 0 && currwin.base_y == 0) {
	update_window_property(XtWindow(top_level), True);
    }

    if (debug & DBG_EVENT)
	printf("Redraw %d x %d at (%d, %d) (base=%d,%d)\n", max_x - min_x,
	       max_y - min_y, min_x, min_y, currwin.base_x, currwin.base_y);

    /* can't use ev_cursor here, since the event loop might not see this change quick enough */
    if (!(ev_flags & EV_CURSOR)) {
	if (!(cursor_flags & (DRAG_HORIZONTAL | DRAG_VERTICAL | DRAG_ALL))) {
	    if (resource.ruler_mode)
		XDefineCursor(DISP, CURSORWIN, rule_cursor);
	    else
		XDefineCursor(DISP, CURSORWIN, wait_cursor);
	    XFlush(DISP);
	}
	ev_flags |= EV_CURSOR;
    }

    raise_message_windows();
    draw_page();
    warn_spec_now = False;
}


void
redraw_page(void)
{
#if COLOR
    _Xconst struct rgb *rgbp;
#endif
    if (debug & DBG_EVENT)
	fputs("Redraw page:  ", stdout);

    if (g_bak_fp == NULL)
	return;
    
    if (scanned_page < current_page) {
	prescan(g_bak_fp);

	if (ev_flags & EV_GE_NEWPAGE)	/* if we need to re-prescan */
	    return;
    }

    if (pageinfo_get_window_width(current_page) != unshrunk_page_w
	|| pageinfo_get_window_height(current_page) != unshrunk_page_h) {
	init_page();
	reconfig();
    }
    
    /* We can't call home() without proper unshrunk_page_*, which requires
     * prescan(), which can't be done from within read_events() */

    /* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!! BUG ALERT !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
       There's some complicated interaction with Postscript specials
       here: if home_action check comes before the gs stuff, psp.drawfile
       might not get initialized correctly, resulting in empty PS figures
       (bounding box instead of figure). This is different in xdvi, due to
       different handling of the home_action stuff, but at the moment I can't
       remember the reason for this ...
       !!!!!!!!!!!!!!!!!!!!!!!!!!!!!! BUG ALERT !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    */
#if PS_GS
    if (gs_postpone_prescan) {
	if (!setjmp(canit_env))
	    gs_resume_prescan();
	else
	    return;
    }
#endif
    if (home_action != NULL) {
	home_action(False);
	home_action = NULL;
	/* This discards the expose event generated by home()
	   (1 for each page) */
	if (read_events(EV_NOWAIT) & EV_GE_NEWPAGE) {
	    return;
	}

	can_exposures(&mane);
    }

#if COLOR
    rgbp = &bg_initial;
    if (page_colors.stack != NULL) {
	ASSERT(current_page < (int)page_colors.size, "page_colors.size too small");
	rgbp = &page_colors.stack[current_page].bg;
    }

    /* Set background color */
    if (bg_current == NULL
	|| rgbp->r != bg_current->color.r
	|| rgbp->g != bg_current->color.g
	|| rgbp->b != bg_current->color.b) {
	struct bgrec **bgpp;

	for (bgpp = &bg_head;;) {
	    bg_current = *bgpp;
	    if (bg_current == NULL) {	/* if bg is not in list */
		bg_current = *bgpp = xmalloc(sizeof *bg_current);
		bg_current->next = NULL;
		bg_current->color = *rgbp;
		bg_current->fg_head = NULL;
		bg_current->pixel_good = False;
		break;
	    }
	    if (bg_current->color.r == rgbp->r
		&& bg_current->color.g == rgbp->g
		&& bg_current->color.b == rgbp->b)
		break;
	    bgpp = &bg_current->next;
	}
	fg_current = NULL;	/* force change of foreground color */
	/* highGC is only used in XDrawRectangle, so its background color
	   doesn't need to be changed.  */
	if (debug & DBG_DVI)
	    printf("Changing background color to %5d %5d %5d\n",
		   bg_current->color.r, bg_current->color.g,
		   bg_current->color.b);

	if (!bg_current->pixel_good) {
	    bg_current->pixel = alloc_color(&bg_current->color,
					    back_color_data.pixel);
	    bg_current->pixel_good = True;
	}
	XSetWindowBackground(DISP, mane.win, bg_current->pixel);
/*  	XClearWindow(DISP, mane.win); */
#if 0 /* don't recolor the cursor - gives too low contrast on color backgrounds,
	 and bad appearance when part of the background is white and cursor mask
	 is colored */
	{
	    XColor	bg_Color;
	    bg_Color.pixel = bg_current->pixel;
	    XQueryColor(DISP, our_colormap, &bg_Color);
	    XRecolorCursor(DISP, ready_cursor, &cr_Color, &bg_Color);
	    XRecolorCursor(DISP, wait_cursor, &cr_Color, &bg_Color);
	}
#endif
    }
#endif /* COLOR */
    
    if (!pause_flag) {
	XClearWindow(DISP, mane.win);
    }

    if (backing_store != NotUseful) {
	mane.min_x = mane.min_y = 0;
	mane.max_x = page_w;
	mane.max_y = page_h;
    }
    else {
	get_xy();
	mane.min_x = -window_x;
	mane.max_x = -window_x + mane.width;
	mane.min_y = -window_y;
	mane.max_y = -window_y + mane.height;
    }

/*      update_TOC(); */
    redraw(&mane);
}

NORETURN void
do_pages(void)
{
    if (debug & DBG_BATCH) {
	
	(void)read_events(EV_GT_IDLE);
	for (current_page = 0; current_page < total_pages; ++current_page) {
	    if (resource.keep_flag) {
		home_action = NULL;
	    }
	    else {
		home_action = home;
	    }
	    warn_spec_now = resource.warn_spec;
#if PS_GS
	    for (;;) {
		redraw_page();
		(void) read_events(EV_NOWAIT);
		if (!(ev_flags & (EV_NEWPAGE | EV_NEWDOC | EV_RELOAD)))
		    break;
		ev_flags = EV_IDLE;
	    }
#else
	    redraw_page();
#endif
	}
	xdvi_exit(EXIT_SUCCESS);
    }
    else {
	for (;;) {	/* normal operation */
	    (void) read_events(EV_GT_IDLE);
/*  	    fprintf(stderr, "ev_flags: %d; ev_newpage: %d, ev_newdoc: %d\n", ev_flags, EV_NEWPAGE, EV_NEWDOC); */
	    /* NOTE: reloading must be checked first! */
	    if (ev_flags & (EV_NEWPAGE | EV_NEWDOC | EV_RELOAD | EV_PS_TOGGLE)) {
		ev_flags &= ~(EV_NEWPAGE | EV_EXPOSE | EV_PS_TOGGLE);
		if (ev_flags & EV_RELOAD) {
		    ev_flags &= ~EV_RELOAD;
		    load_dvi_file(True, True);
		}
		if (ev_flags & EV_NEWDOC) {
/*  		    fprintf(stderr, "newdoc!\n"); */
		    ev_flags &= ~EV_NEWDOC;
		    load_dvi_file(False, True);
		}

		can_exposures(&mane);
		can_exposures(&magnifier);

#if PS && PS_GS && GS_PIXMAP_CLEARING_HACK
		if (had_ps_specials && !MAGNIFIER_ACTIVE) {
		    erasepage_gs();
		    had_ps_specials = False;
		}
#endif /* PS && PS_GS && GS_PIXMAP_CLEARING_HACK */

		if (g_bak_fp != NULL)
		    redraw_page();
		else
		    (void)dvi_file_changed();
	    }
	    else if (ev_flags & EV_SEARCH_CANCEL) {
		/* NOTE: This must be done before checking for expose() */
		ev_flags &= ~EV_SEARCH_CANCEL;
	    }
#ifdef HTEX
	    else if (ev_flags & EV_ANCHOR) {
		/*
		 * Similar to forward search: search for a htex anchor.
		 * This needs to come before the next case which does the redraw_page(),
		 * otherwise anchors for the current page might not be drawn at all:
		 * anchor_search() sets the info later used by htex_draw_anchormarkers(),
		 * which is invoked by redraw_page().
		 */
		
		/* switch off the link cursor */
		cursor_flags &= ~LINK_CURSOR;

		if (dvi_file_changed())
		    continue;
		
		anchor_search(g_anchor_pos);
		ev_flags &= ~EV_ANCHOR;
	    }
#endif /* HTEX */
	    else if (ev_flags & EV_SRC) {
		/*
		 * Source special operations are deferred to here because
		 * they call geom_scan(), which may call define_font(),
		 * which may call makefont(), which may call read_events()
		 * recursively.
		 */
		if (source_forward_string != NULL) {
		    _Xconst char *s = source_forward_string;

		    if (dvi_file_changed())
			continue;
		    
		    source_forward_search(s);
		    ev_flags &= ~EV_SRC;
		    source_forward_string = NULL;

		    /* de-iconify window if needed, and raise it */
		    XMapRaised(XtDisplay(top_level), XtWindow(top_level));
		}
		else if (source_reverse_x != -1) {
		    if (dvi_file_changed())
			continue;
		    
		    source_reverse_search(source_reverse_x, source_reverse_y, True);
		    ev_flags &= ~EV_SRC;
		}
		else {
		    source_special_show(source_show_all);
		    ev_flags &= ~EV_SRC;
		}

	    }
	    else if (ev_flags & EV_MAG_MOVE) {
		move_magnifier();
	    }
	    else if (ev_flags & EV_EXPOSE) {
		if (magnifier.min_x < MAXDIM) {
/*  		    fprintf(stderr, "magnifier < maxdim!\n"); */
		    if (mane.min_x >= MAXDIM) {
/*  			fprintf(stderr, "mane >= maxdim!\n"); */
			ev_flags &= ~EV_EXPOSE;
		    }
		    redraw(&magnifier);
		}
		else {
		    /* see comment in mag.c */
		    ev_flags &= ~EV_EXPOSE;
		    if (mane.min_x < MAXDIM) {
			redraw(&mane);
		    }
		}
	    }
	    else if (ev_flags & EV_CURSOR) {
		/*
		 * This code eliminates unnecessary calls to XDefineCursor,
		 * since this is a slow operation on some hardware (e.g., S3
		 * chips).
		 */
		XSync(DISP, False);
		if (!XtPending()) {
		    Cursor curr;
		    if (cursor_flags & DRAG_VERTICAL)
			curr = drag_cursor_v;
		    else if (cursor_flags & DRAG_HORIZONTAL)
			curr = drag_cursor_h;
		    else if (cursor_flags & DRAG_ALL)
			curr = drag_cursor_a;
		    else if (resource.ruler_mode)
			curr = rule_cursor;
		    else if (cursor_flags & LINK_CURSOR)
			curr = link_cursor;
		    else if (cursor_flags & MAG_CURSOR)
			curr = mag_cursor;
		    else if (cursor_flags & CORRUPTED_CURSOR)
			curr = corrupted_cursor;
		    else if (pause_flag)
			curr = pause_cursor;
		    else
			curr = ready_cursor;
		    XDefineCursor(DISP, CURSORWIN, curr);
		    ev_flags &= ~EV_CURSOR;
		}
	    }
	    XFlush(DISP);
	}
    }
}

static void
Act_unpause_or_next(Widget w, XEvent *event,
         String *params, Cardinal *num_params)
{
    if (pause_flag) {
        pause_num++;
	if (pause_num_save)
	    pause_num_save[current_page] = pause_num; 
	redraw_page();
    }
    else {
	Act_down_or_next(w, event, params, num_params);
    }
}

/*
 * timer callback for watching the DVI file.
 */
void
watch_file_cb(XtPointer client_data, XtIntervalId *id)
{
    static XtIntervalId timer = 0;
    
    UNUSED(client_data);
    UNUSED(id);

    if (resource.watch_file > 0.0) {
	unsigned long watch_time_ms;

	check_watch_dvi_file();
    
	if (timer) {
	    XtRemoveTimeOut(timer);
	    timer = (XtIntervalId)(XtIntervalId)0;
	}

	watch_time_ms = (unsigned long)(resource.watch_file * 1000);
	timer = XtAppAddTimeOut(app, watch_time_ms, watch_file_cb, (XtPointer)NULL);
    }
}
