/*****************************************************************************
Help window for xdvi

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL PAUL
VOJTA OR ANY OTHER AUTHOR OF THIS SOFTWARE BE LIABLE FOR ANY CLAIM, DAMAGES OR
OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

*****************************************************************************/

#include "xdvi-config.h"
#include "xdvi.h"
#include "version.h"

#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#if MOTIF
# include <Xm/Xm.h>
# include <Xm/Form.h>
# include <Xm/DialogS.h>
# include <Xm/RowColumn.h>
# include <Xm/CascadeB.h>
# include <Xm/PushB.h>
# include <Xm/Frame.h>
# include <Xm/Text.h>
# include <Xm/PanedW.h>
# include <Xm/Label.h>
# include <Xm/LabelG.h>
# include <Xm/Protocols.h>
# include <Xm/List.h>
#else
# include <X11/Xaw/Paned.h>
# include <X11/Xaw/Box.h>
# include <X11/Xaw/MenuButton.h>
# include <X11/Xaw/SimpleMenu.h>
# include <X11/Xaw/Sme.h>
# include <X11/Xaw/Form.h>
# include <X11/Xaw/Cardinals.h>
# include <X11/Xaw/Command.h>
# include <X11/Xaw/SmeBSB.h>
# include <X11/Xaw/AsciiText.h>
# include <X11/Xaw/List.h>
#endif

#if MOTIF

const char *TAG_NORMAL = "NORMAL";
const char *TAG_TOPIC_HEADING = "TOPIC_HEADING";
const char *TAG_TOPIC_LABEL = "TOPIC_LABEL";

# define SHELL_WIDGET_CLASS	xmDialogShellWidgetClass
# define PANED_WIDGET_CLASS	xmPanedWindowWidgetClass
# define FORM_WIDGET_CLASS	xmFormWidgetClass
  
# define FORM_ARGS		XmNhorizontalSpacing, 10,	\
				NULL
# define HELP_SHELL_ARGS	XmNdeleteResponse, XmDO_NOTHING,	\
  				XtNmappedWhenManaged, False,		\
				NULL
# define HELP_PANED_ARGS	XmNsashWidth, 1,	\
				XmNsashHeight, 1,	\
				NULL
# define LEFT_FORM_ARGS		XmNtopAttachment, XmATTACH_FORM,	\
  				XmNleftAttachment, XmATTACH_FORM,	\
  				XmNbottomAttachment, XmATTACH_FORM,	\
				XmNtopOffset, 9,		\
				XmNbottomOffset, 10,		\
				NULL
# define RIGHT_FORM_ARGS	XmNtopAttachment, XmATTACH_OPPOSITE_WIDGET,	\
  				XmNtopWidget, left_form,			\
  				XmNleftAttachment, XmATTACH_WIDGET,		\
  				XmNleftWidget, left_form,			\
  				XmNrightAttachment, XmATTACH_FORM,		\
  				XmNbottomAttachment, XmATTACH_FORM,		\
				XmNbottomOffset, 10,		\
				NULL
# define ACTION_AREA_ARGS	NULL

#else /* MOTIF */

# define SHELL_WIDGET_CLASS	transientShellWidgetClass
# define PANED_WIDGET_CLASS	panedWidgetClass
# define FORM_WIDGET_CLASS	formWidgetClass
  
# define FORM_ARGS		XtNdefaultDistance, 6, NULL
# define HELP_SHELL_ARGS	XtNx, 60,				\
				XtNy, 80,				\
  				XtNtransientFor, top_level,		\
  				XtNtranslations, wm_translations,	\
  				XtNtransientFor, toplevel,		\
  				XtNallowShellResize, False,		\
				NULL
# define HELP_PANED_ARGS	NULL
# define LEFT_FORM_ARGS		XtNtop, XtChainTop,		\
  				XtNbottom, XtChainBottom,	\
  				XtNleft, XtChainLeft,		\
  				XtNright, XtChainLeft,		\
  				XtNborderWidth, 0,		\
				XtNdefaultDistance, 6,		\
  				NULL
# define RIGHT_FORM_ARGS	XtNfromHoriz, left_form,	\
  				XtNtop, XtChainTop,		\
  				XtNbottom, XtChainBottom,	\
  				XtNleft, XtChainLeft,		\
  				XtNright, XtChainRight,		\
  				XtNvertDistance, 16,		\
  				XtNborderWidth, 0,		\
  				NULL
# define ACTION_AREA_ARGS	XtNdefaultDistance, 6,	\
				XtNshowGrip, False,	\
				XtNskipAdjust, True,	\
				NULL
#endif /* MOTIF */

#include <sys/stat.h>
#include <stdio.h>
#include <unistd.h>

#include "message-window.h"
#include "util.h"
#include "x_util.h"
#include "help-window.h"


/*
 * store info on all widgets into array so that we can initialize them within a loop;
 * `name' == NULL indicates end of array
 */
#define WIDGET_INFO_LIST_LEN 16 /* number of help topics */

static struct widget_info {
    char *name;			/* name of widget */
    char *title;		/* string used for menu entry; NULL if widget is not a menu entry */
    char *summary;		/* if not NULL, a short summary of the topic */
    char *resource_string;	/* value of X resource */
    _Xconst char **default_resource_string;	/* fallback if resource is unspecified */
} widget_info_list[WIDGET_INFO_LIST_LEN];

#if MOTIF
static XmString
#else
static char *
#endif
create_label_string(_Xconst char *title, _Xconst char *subtitle);
/*
 * Global Variables for communication with callbacks
 */

static int help_active = 0;
static Widget help_text, help_shell, help_label, topics_list;


/*------------------------------------------------------------
 *  popdown_help_callback
 *
 *  Arguments:
 *	Widget w	- unused
 *	XtPointer a, b  - unused
 *
 *  Returns:
 *	void
 *
 *  Purpose:
 *	Help popdown callback function
 *------------------------------------------------------------*/

void
popdown_help_callback(Widget w, XtPointer a, XtPointer b)
{
    UNUSED(w);
    UNUSED(a);
    UNUSED(b);

    if (!help_active) {
	return;
    }
    help_active = 0;
    XtPopdown(help_shell);
}

/*
 * Print the contents of the help associated with
 * menu entry w into the helptext widget.
 */
void
select_help_topic(char *topic)
{
    int j;

    if (topic == NULL)
	return;

    for (j = 0; widget_info_list[j].name != NULL; ++j) {
	if (strcmp(widget_info_list[j].name, topic) == 0) {
#if MOTIF
	    XmString label = create_label_string(widget_info_list[j].title, widget_info_list[j].summary);
	    XtVaSetValues(help_label, XmNlabelString, label, NULL);
	    XmTextSetString(help_text, widget_info_list[j].resource_string);
	    XmListSelectPos(topics_list, j + 1, False);
	    XmStringFree(label);
#else
	    char *label = create_label_string(widget_info_list[j].title, widget_info_list[j].summary);
	    XtVaSetValues(help_label, XtNlabel, label, NULL);
	    XtVaSetValues(help_text, XtNstring, widget_info_list[j].resource_string, NULL);
	    XawListHighlight(topics_list, j);
	    free(label);
#endif
	    return;
	}
    }
    XBell(DISP, 0);
    popup_message(MSG_WARN,
		  NULL,
		  "Shouldn't happen: Could not find topic `%s' in help list!\n"
		  REPORT_XDVI_BUG_TEMPLATE,
		  topic);
}

static void
select_help_topic_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
#if MOTIF
    XmListCallbackStruct *cbs = (XmListCallbackStruct *)call_data;
    int new = cbs->item_position;

    UNUSED(w);
    UNUSED(client_data);
    
    select_help_topic(widget_info_list[new - 1].title);
#else
    XawListReturnStruct *item = (XawListReturnStruct *)call_data;
    int new = item->list_index;

    UNUSED(w);
    UNUSED(client_data);

    select_help_topic(widget_info_list[new].title);
#endif
}


#if !MOTIF
static void
xaw_popdown(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    UNUSED(event);
    UNUSED(params);
    UNUSED(num_params);
    popdown_help_callback(w, NULL, NULL);
}
#endif

/*
 * helper routines
 */

#if MOTIF
static XmString
#else
static char *
#endif
create_label_string(_Xconst char *title, _Xconst char *subtitle)
{
#if MOTIF
    XmString label;
    if (subtitle == NULL) { /* simple bold label */
	label = XmStringCreate((char *)title, (char *)TAG_TOPIC_LABEL);
    }
    else { /* two-part title with subheading */
	XmString s1, s2, s3, s4;
	s1 = XmStringCreate((char *)title, (char *)TAG_TOPIC_HEADING);
	s2 = XmStringCreate((char *)"   ", (char *)TAG_NORMAL);
	s3 = XmStringCreate((char *)subtitle, (char *)TAG_NORMAL);
	s4 = XmStringConcat(s1, s2);
	label = XmStringConcat(s4, s3);
	XmStringFree(s1);
	XmStringFree(s2);
	XmStringFree(s3);
	XmStringFree(s4);
    }
#else
    char *label = xstrdup(title);
    if (subtitle != NULL && strlen(subtitle) > 0) {
	label = xstrcat(label, "   -   ");
	label = xstrcat(label, subtitle);
    }
#endif
    return label;
}

static Widget
create_label_widget(Widget parent, _Xconst char *name, _Xconst char *title, _Xconst char *subtitle)
{
    Widget label;
#if MOTIF
    XmString label_string = create_label_string(title, subtitle);
    label = XtVaCreateManagedWidget(name, xmLabelGadgetClass, parent,
				    XmNtopAttachment, XmATTACH_FORM,
				    XmNleftAttachment, XmATTACH_FORM,
				    XmNrightAttachment, XmATTACH_FORM,
				    XmNlabelString, label_string,
				    XmNalignment, XmALIGNMENT_BEGINNING,
				    NULL);
    XmStringFree(label_string);
#else /* MOTIF */
    char *label_string = create_label_string(title, subtitle);
    label = XtVaCreateManagedWidget(name, labelWidgetClass, parent,
				    XtNlabel, label_string,
				    XtNborderWidth, 0,
/* 				    XtNinternalHeight, 2, */
				    XtNjustify, XtJustifyLeft,
				    NULL);
    free(label_string);
#endif /* MOTIF */
    XtManageChild(label);
    return label;
}

static Widget
create_list_widget(Widget parent, _Xconst char *name)
{
    Widget list;
    size_t topics_num;
#if MOTIF
    XmString *help_topics = NULL;
#else
    static char **help_topics = NULL;
#endif

    for (topics_num = 0; widget_info_list[topics_num].title != NULL; topics_num++) {
	help_topics = xrealloc(help_topics, (topics_num + 2) * sizeof help_topics);
#if MOTIF
	help_topics[topics_num] = XmStringCreateLocalized(widget_info_list[topics_num].title);
#else
	help_topics[topics_num] = xstrdup(widget_info_list[topics_num].title);
#endif
    }
    
#if MOTIF
    {
	size_t i;
	Arg args[20];
	int n = 0;
	XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
/*  	XtSetArg(args[n], XmNtopOffset, 10);			n++; */
	XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
	XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
	XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
	XtSetArg(args[n], XmNlistSpacing, 4); n++;
	XtSetArg(args[n], XmNlistMarginHeight, 4); n++;
/*  	XtSetArg(args[n], XmNtopOffset, 10); n++; */
/*  	XtSetArg(args[n], XmNbottomOffset, 10); n++; */
	XtSetArg(args[n], XmNlistMarginWidth, 4); n++;
	XtSetArg(args[n], XmNhighlightThickness, 0); n++;
	XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
	
	list = XmCreateScrolledList(parent, (char *)name, args, n);

#ifdef __GNUC__
#warning TODO: make up/down scroll the list, and PgUp/PgDown scroll help text
#warning TODO: add wheel mouse bindings
#endif
	
	XmListDeleteAllItems(list);
	XmListAddItems(list, help_topics, topics_num, 0);
	
	for (i = 0; i < topics_num; ++i) {
	    XmStringFree(help_topics[i]);
	}
	free(help_topics);
    }
    XtAddCallback(list, XmNbrowseSelectionCallback, select_help_topic_callback, NULL);
    XmListSelectPos(list, 1, False); /* default position */
#else /* MOTIF */
    {
	int i = 0;
	static char **help_topics = NULL;
	for (i = 0; widget_info_list[i].title != NULL; i++) {
	    help_topics = xrealloc(help_topics, (i + 2) * sizeof *help_topics); /* 1 more for terminating NULL */
	    help_topics[i] = xstrdup(widget_info_list[i].title);
	}
	help_topics[i] = NULL;
	list = XtVaCreateWidget(name, listWidgetClass, parent,
				XtNlist, help_topics,
				XtNdefaultColumns, 1,
				XtNforceColumns, True,
				XtNscrollVertical, XawtextScrollAlways,
				XtNverticalList, True,
				XtNrowSpacing, 4,
				XtNheight, 429,
				NULL);
    }
    XtAddCallback(list, XtNcallback, select_help_topic_callback, NULL);
    XawListHighlight(list, 0); /* default position */
#endif /* MOTIF */
    XtManageChild(list);
    return list;
}

static Widget
create_help_text(Widget parent, Widget top_widget, _Xconst char *name, _Xconst char *value)
{
    Widget text;
#if MOTIF
    Arg args[20];
    int n = 0;

    XtSetArg(args[n], XmNeditable, False);				n++;
    XtSetArg(args[n], XmNcursorPositionVisible, False);		n++;
    XtSetArg(args[n], XmNvalue, value);	n++;
    XtSetArg(args[n], XmNeditMode, XmMULTI_LINE_EDIT);		n++;
    XtSetArg(args[n], XmNwordWrap, True);				n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET);		n++;
    XtSetArg(args[n], XmNtopWidget, top_widget);			n++;
    XtSetArg(args[n], XmNtopOffset, 10);			n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM);		n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM);		n++;
    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM);		n++;
    XtSetArg(args[n], XmNscrollingPolicy, XmAUTOMATIC);		n++;
    XtSetArg(args[n], XmNscrollBarDisplayPolicy, XmAS_NEEDED);	n++;
    
    text = XmCreateScrolledText(parent, (char *)name, args, n);

    XtManageChild(text);
#else /* MOTIF */
    text = XtVaCreateManagedWidget(name, asciiTextWidgetClass, parent,
				   XtNfromVert, top_widget,
				   XtNvertDistance, 10,
				   XtNstring, value,
				   XtNheight, 400,
				   XtNwidth, 500,
				   /* resizing of pane by user isn't needed */
				   XtNshowGrip, False,
				   XtNscrollVertical, XawtextScrollAlways,
				   XtNscrollHorizontal, XawtextScrollNever,
				   XtNeditType, XawtextRead,
				   XtNleftMargin, 5,
				   NULL);
    XawTextDisplayCaret(text, False);
#endif /* MOTIF */
    return text;
}

static Widget
create_button(Widget parent, _Xconst char *name, Boolean show_as_default, Boolean left_position)
{
    Widget button;
#if MOTIF
    XmString s1 = XmStringCreateLocalized((char *)name);
    button = XtVaCreateWidget(name, xmPushButtonWidgetClass, parent,
			      XmNlabelString, s1,
			      XmNdefaultButtonShadowThickness, 1,
			      XmNtopAttachment, XmATTACH_FORM,
			      XmNbottomAttachment, XmATTACH_FORM,
			      XmNmarginWidth, 6,
			      XmNmarginHeight, 4,
			      XmNtopOffset, 10,
			      XmNbottomOffset, 10,
			      NULL);
    if (left_position)
	XtVaSetValues(button,
		      XmNleftAttachment, XmATTACH_FORM,
		      XmNleftOffset, 10,
		      NULL);
    else
	XtVaSetValues(button,
		      XmNrightAttachment, XmATTACH_FORM,
		      XmNrightOffset, 10,
		      NULL);
    if (show_as_default)
	XtVaSetValues(button, XmNshowAsDefault, True, NULL);

    XtManageChild(button);
    XmStringFree(s1);
#else /* MOTIF */
    UNUSED(show_as_default);
    button = XtVaCreateManagedWidget(name, commandWidgetClass, parent,
				     XtNtop, XtChainTop,
				     XtNbottom, XtChainBottom,
				     NULL);
    if (left_position)
	XtVaSetValues(button,
		      XtNleft, XtChainLeft,
		      XtNright, XtChainLeft,
		      NULL);
    else
	XtVaSetValues(button,
		      XtNleft, XtChainRight,
		      XtNright, XtChainRight,
		      NULL);
#endif /* MOTIF */
    return button;
}


static void
get_title_and_summary(_Xconst char *str, int len, char **title, char **summary)
{
    char *tmp = xmalloc(len + 1);
    char *ptr;

    memcpy(tmp, str, len);
    tmp[len] = '\0';

    if ((ptr = strchr(tmp, '\t')) == NULL) {
	XDVI_WARNING((stderr, "Help resource label `%s' doesn't contain a tab character - ignoring it.", tmp));
	*title = tmp;
	*summary = NULL;
    }
    else {
	*ptr++ = '\0';
	*title = tmp;
	*summary = ptr;
    }
    
    TRACE_GUI((stderr, "Title, Summary: |%s|%s|", *title, *summary));
}

static void
get_resource_setting(_Xconst char *resource, _Xconst char **resource_default,
		      char **widget_text,
		      char **title, char **summary)
{
    _Xconst char *ptr;
    
    if (resource != NULL) {
	if ((ptr = strchr(resource, '\n')) == NULL) {
	    XDVI_WARNING((stderr, "Help resource text `%s' doesn't contain a newline character.", resource));
	    ptr = resource;
	}
	else
	    ptr++;
	get_title_and_summary(resource, ptr - resource - 1, title, summary);
	*widget_text = (char *)ptr;
    }
    else { /* resource not set; copy resource_default into malloc()ed widget_text: */
	size_t size = 0, alloc_len = 0, offset;
	_Xconst size_t alloc_step = 1024;
	int i;

	for (i = 0; resource_default[i] != NULL; i++) {
	    if (i == 0) { /* special case */
		get_title_and_summary(resource_default[i], strlen(resource_default[i]) - 1, title, summary);
	    }
	    else {
		offset = size;
		size += strlen(resource_default[i]);
		/*
		 * allocate chunks of `alloc_step' to avoid frequent calls to malloc.
		 * `alloc_len' is always 1 more than `size', for the terminating NULL character.
		 */
		while (size + 1 > alloc_len) {
		    alloc_len += alloc_step;
		    *widget_text = xrealloc(*widget_text, alloc_len);
		}
		memcpy(*widget_text + offset, resource_default[i], size - offset);
	    }
	}
	/* null-terminate string */
	(*widget_text)[size] = '\0';
    }
}

static void
initialize_resources(void)
{
    int k;
    
    static Boolean initialized = False;

    /*
     * Define fallbacks: default_xyz is used as fallback text if
     * X resource xyz isn't specified.
     *
     * We use arrays of strings rather than simple strings because of C's
     * limitations on maximum string length; but since the resource needs to
     * be a simple `char *', these have to be copied into larger buffers
     * later on (which is a bit wasteful to space). OTOH, splitting the
     * strings into smaller pieces would make them hard to deal with as
     * X resources. They are defined as static so that they are initialized
     * only once.
     *
     * Last elem of each array is NULL for ease of looping through it.
     *
     * Advantages of this method vs. putting the help texts into a file:
     * - couldn't use #ifdef's as easily
     * - would need to invent our own file format
     * - file searching is more error-prone (needs to be installed etc.)
     */

    static _Xconst char *default_help_general[] = {
	"Introduction\tAbout this version of Xdvi\n",
	"This is xdvik, version ",
	XDVI_TERSE_VERSION_INFO
	".\nThe xdvik project homepage is located at:\n",
	"http://sourceforge.net/projects/xdvi\n",
	"where you can find updates, report bugs and submit feature requests.\n",
	"\n",
	"\n",
	"Getting help and exiting xdvi:\n",
	"\n",
	"h or ? or `Help' button\n",
	"     Displays this help text.\n",
	"\n",
	"q or Ctrl-z (VAX VMS)\n",
	"     Quits the program.\n",
	"\n",
	"Xdvi accepts many command-line options, which are actually too numerous\n",
	"to be listed here; please see the man page for a full description.\n",
	"\n",
	"\n",
	"Built on ",
	__DATE__,
	" using these configure options:\n",
#if MOTIF
	"- Motif toolkit (",
	XmVERSION_STRING,
	")\n",
#else
	"- Athena toolkit\n",
#endif
#ifdef HTEX
	"- htex enabled\n",
#endif
#ifdef A4
	"- paper: a4, units cm\n",
#else
	"- paper: letter, units inches\n",
#endif
#ifdef GREY
	"- anti-aliasing (grey) enabled\n",
#endif
#ifdef T1LIB
	"- T1lib (direct rendering of PS fonts) enabled\n",
#endif
#ifdef TEXXET
	"- left-to-right typesetting (TeXXeT) support enabled\n",
#endif
#ifdef USE_GF
	"- gf file support enabled\n",
#endif
	NULL
    };

#ifdef HTEX
    static _Xconst char *default_help_hypertex[] = {
	"Hyperlinks\tNavigating links\n",
	"Whenever the mouse is positioned on a link, the cursor changes\n",
	"to a `hand' shape and the target of the link is displayed\n",
	"in the statusline at the bottom of the window.\n",
	"\n",
	"The following keybindings are pre-configured:\n",
	"\n",
        "Mouse-1\n",
	"	Follow the link at the cursor position.\n",
	"	If the link target is not a DVI file, try to launch\n",
	"	an application to view the file.\n",
        "Mouse-2\n",
	"	Open a new xdvi window displaying the link\n",
        "	at the cursor position if the link is a DVI file;\n",
	"	else, try to launch an application to view the file.\n",
        "B",
#if MOTIF
	", toolbar button 9\n",
#else
	"\n",
#endif	
        "	Go back to the previous hyperlink in the history.\n",
        "F",
#if MOTIF
	", toolbar button 10\n",
#else
	"\n",
#endif	
	"	Go forward to the next hyperlink in the history.\n",
	"\n",
	"By default, the hyperlinks are displayed in the colors \n",
	"`linkColor' and `visitedLinkColor' (for visited links) and \n",
	"underlined in the same colors. This can be customized \n",
	"by setting the resource or command-line option `linkstyle' \n",
	"to a value between 0 and 3, which have the following meaning:\n",
	"    0: no highlighting of links,\n",
	"    1: underline links,\n",
	"    2: color links,\n",
	"    3: color and underline links.\n\n",
	NULL
    };
#endif

    static _Xconst char *default_help_othercommands[] = {
	"Other Commands\tMiscellaneous other commands\n",
	"Note: Unless a key binding also has an uppercase version,\n",
	"all bindings are case-insensitive.\n\n",
#if defined(SELFILE)
	"Ctrl-f",
#if MOTIF
	", toolbar button 1\n",
#else
	"\n",
#endif	
	"     Opens a popup window to select another DVI file.\n",
	"\n",
#endif
	"Ctrl-l\n",
	"     Toggles fullscreen mode (which may not work with your\n",
	"     window manager/desktop).\n",
	"\n",
	"Ctrl-r or Clear\n",
	"     Redisplays the current page.\n",
	"\n",
	"Ctrl-p",
#if MOTIF
	", toolbar button 11\n",
#else
	"\n",
#endif	
	"     Opens a popup window for printing the DVI file, or parts of it.\n",
	"\n",
	"Ctrl-s\n",
	"     Opens a popup window for saving the DVI file, or parts of it.\n",
	"\n",
	"G    Toggles the use of greyscale anti-aliasing for\n",
	"     displaying shrunken bitmaps.  In addition, the key\n",
	"     sequences `0G' and `1G' clear and set this flag,\n",
	"     respectively.  See also the -nogrey option.\n",
	"\n",
	"k    Normally when xdvi switches pages, it moves to the home\n",
	"     position as well.  The `k' keystroke toggles a `keep-\n",
	"     position' flag which, when set, will keep the same\n",
	"     position when moving between pages.  Also `0k' and `1k'\n",
	"     clear and set this flag, respectively.  See also the\n",
	"     -keep option.\n",
	"\n",
	"M    Sets the margins so that the point currently under the\n",
	"     cursor is the upper left-hand corner of the text in the\n",
	"     page.  Note that this command itself does not move the\n",
	"     image at all.  For details on how the margins are used,\n",
	"     see the -margins option.\n",
	"\n",
	"P    ``This is page number n.''  This can be used to make\n",
	"     the `g' keystroke refer to a different page number.\n",
	"     (See also `Options->Use TeX pages').\n",
	"\n",
	"R",
#if MOTIF
	", toolbar button 2\n",
#else
	"\n",
#endif
	"     Forces the DVI file to be reread.\n",
	"\n",
	"s    Changes the shrink factor to the given number.\n",
	"     If no number is given, the smallest factor that makes the\n",
	"     entire page fit in the window will be used.  (Margins\n",
	"     are ignored in this computation.)\n",
	"\n",
	"S    Sets the density factor to be used when shrinking\n",
	"     bitmaps.  This should be a number between 0 and 100;\n",
	"     higher numbers produce lighter characters.\n",
	"\n",
	"t    Switches to the next unit in a sorted list of TeX dimension\n",
	"     units for the popup magnifier ruler and `Ruler mode'\n",
	"\n",
	"V    Toggles Ghostscript anti-aliasing.  Also `0V' and `1V' clear\n",
	"     and enables this mode, respectively.  See also the the\n",
	"     -gsalpha option.\n",
	"\n",
	"\n",
	"v    Toggles between several modes of displaying postscript specials:\n",
	"     Display specials, display specials with their bounding box\n",
	"     (if available), and display bounding boxes only (if available).\n",
	"     The prefix arguments 1, 2 and 0 also allow you to select one of\n"
	"     these states directly.\n",
	"\n",
	"x    Toggles expert mode (in which ",
#if MOTIF
	"the menu bar, the toolbar\n",
#else
	"the menu buttons,\n",
#endif
	"     the page list and the statusline do not appear).\n",
	"     `1x' toggles the display of the statusline at the bottom of\n",
	"     the window.\n",
	"     `2x' toggles the scrollbars,\n",
#if MOTIF
	"     `3x' toggles the page list,\n",
	"     `4x' toggles the toolbar,\n",
	"     `5x' toggles the menu bar.\n",
#else
	"     `3x' toggles the page list and menu buttons.\n",
#endif
	"\n",
	"Ctrl-+",
#if MOTIF
	", toolbar button 7\n",
#else
	"\n",
#endif
	"     Makes the display of the page larger (zooms in).\n",
	"\n",
	"Ctrl-+",
#if MOTIF
	", toolbar button 8\n",
#else
	"\n",
#endif
	"     Makes the display of the page smaller (zooms out).\n",
	"\n",
	"Alt-Ctrl-+",
#if MOTIF
	", toolbar button 16\n",
#else
	"\n",
#endif
	"     Makes the fonts darker (by adding to the gamma value).\n",
	"\n",
	"Alt-Ctrl--",
#if MOTIF
	", toolbar button 17\n",
#else
	"\n",
#endif
	"     Makes the fonts lighter (by substracting from the gamma\n",
	"     value).\n",
	"\n",
	NULL
    };

    static _Xconst char *default_help_marking[] = {
	"Printing and Saving\tMarking, printing and saving pages\n",
	"Note: Unless a key binding also has an uppercase version,\n",
	"all bindings are case-insensitive.\n\n",
	"The `Save' and `Print' dialogs allow you to save or print all pages,\n",
	"a range of pages, or all marked pages from a DVI file. To mark\n",
        "a page or a range of pages, use one of the folllowing methods:\n",
	"- Click on the page in the page list with Mouse Button 2 to mark\n",
	"  a single page, or drag the mouse while holding down Button 2\n",
	"  to mark a range of pages.\n",
	"- Use one of the following key combinations:\n",
	"     m:  toggle the mark of the current page,\n",
	"     1m  toggle the marks of all odd pages,\n",
	"     2m  toggle the marks of all even pages,\n",
	"     0m: unmark all pages,\n",
	"     Ctrl-n: toggle mark of current page, then move one page forward,\n",
	"     Ctrl-u: move one page back, then toggle mark of that page.\n",
#if MOTIF
        "- Use one of the toobar buttons 12 to 15 to toggle the marks\n",
        "  of odd pages, toggle the marks of even pages, toggle the mark\n",
	"  of the current page, or unmark all pages, respectively.\n",
#endif
	"\n",
	"If the X resource or command line option `paper' has been used, its\n",
	"value is inserted into the `Dvips Options' field of the printing\n",
	"dialog so that the appropriate options can be passed to dvips.\n",
	"This doesn't happen if the paper size has been specified explicitly\n",
	"in the DVI file (e.g. by using the LaTeX `geometry' package).\n",
	"Note that  not all of the paper options used by xdvi\n",
	"may be understood by dvips; dvips will ignore the option\n",
	"in that case, and will use its default paper setting.\n",
	NULL
    };
    
    static _Xconst char *default_help_pagemotion[] = {
	"Page Motion\tMoving around in the document\n",
	"Note: Unless a key binding also has an uppercase version,\n",
	"all bindings are case-insensitive.\n\n",
	"n or f or Space or Return or LineFeed or PgDn",
#if MOTIF
	", toolbar button 5\n",
#else
	"\n",
#endif
	"     Moves to the next page (or to the nth next page if a\n",
	"     number is given).\n",
	"\n",
	"p or b or Ctrl-h or BackSpace or Del or PgUp",
#if MOTIF
	", toolbar button 4\n",
#else
	"\n",
#endif
	"     Moves to the previous page (or back n pages).\n",
	"\n",
	"Up-arrow\n",
	"     Scrolls page up.\n",
	"\n",
	"Down-arrow\n",
	"     Scrolls page down.\n",
	"u\n",
	"     Moves page up two thirds of a window-full.\n",
	"\n",
	"d\n",
	"     Moves page down two thirds of a window-full.\n",
	"\n",
	"Left-arrow\n",
	"     Scrolls page left.\n",
	"\n",
	"Right-arrow\n",
	"     Scrolls page right.\n",
	"\n",
	"l\n",
	"     Moves page left two thirds of a window-full.\n",
	"\n",
	"r\n",
	"     Moves page right two thirds of a window-full.\n",
	"\n",
	"g or j or End\n",
	"     Moves to the page with the given number.  Initially,\n",
	"     the first page is assumed to be page number 1, but this\n",
	"     can be changed with the `P' keystroke, below.  If no\n",
	"     page number is given, then it goes to the last page.\n",
	"\n",
	"<",
#if MOTIF
	", toolbar button 3\n",
#else
	"\n",
#endif
	"     Moves to first page in the document.\n",
	">",
#if MOTIF
	", toolbar button 6\n",
#else
	"\n",
#endif
	"     Moves to last page in the document.\n",
	"\n",
	"^ or Home\n",
	"     Move to the ``home'' position of the page.  This is\n",
	"     normally the upper left-hand corner of the page,\n",
	"     depending on the margins as described in the -margins\n",
	"     option, above.\n",
	"\n",
	"c    Moves the page so that the point currently beneath the\n",
	"     cursor is moved to the middle of the window.  It also\n",
	"     warps the cursor to the same place.\n",
	"\n",
	NULL
    };


    static _Xconst char *default_help_mousebuttons[] = {
	"Mouse Buttons\tActions bound to the mouse buttons\n",
	"The mouse buttons can be customized just like the keys;\n",
	"however the bindings cannot be intermixed (since\n",
	"a mouse event always requires the cursor location\n",
	"to be present, which a key event doesn't).\n",
        "The default bindings are as follows:\n"
	"\n",
	"Buttons 1-3\n",
	"     Pops up magnifier windows of different sizes.\n",
	"     When the mouse is over a hyperlink, the link overrides\n",
	"     the magnifier. In that case, Button 1 jumps to the link\n",
	"     in the current Xdvi window, Button 2 opens the link target\n",
	"     in a new instance of Xdvik.\n"
	"     In `Ruler Mode', Button1 shows/drags the ruler instead.\n"
	"\n",
	"Shift-Button1 to Shift-Button3\n",
	"     Drag the page in each direction (Button 1), vertically\n",
	"     only (Button 2) or horizontally only (Button 3).\n",
	"\n",
	"Ctrl-Button1\n",
	"     Invoke a reverse search for the text on the cursor\n",
	"     location (see the section SOURCE SPECIALS for more\n",
	"     information on this).\n",
	"\n",
	"The buttons 4 and 5 (wheel up and down for wheel mice)\n",
	"scroll the page up and down respectively, or jump to the\n",
	"next/previous page when the mouse is over the page list.",
	"\n",
	"In `Ruler Mode', Button 1 sets the ruler; moving the mouse\n",
	"while holding Button 1 down drags the ruler.\n",
	"\n",
	"In the page list, Button 2 toggles the mark a page (see\n",
	"section `Marking Pages'); moving the mouse while holding\n",
	"Button 2 lets you toggle a range of pages.\n",
	"\n",
	NULL
    };

    static _Xconst char *default_help_sourcespecials[] = {
	"Source Specials\tNavigating between the TeX and the DVI file\n",
	"Note: Unless a key binding also has an uppercase version,\n",
	"all bindings are case-insensitive.\n\n",
	"Some TeX implementations have an option to automatically\n",
	"include so-called `source specials' into a DVI file. These\n",
	"contain the line number and the filename of the TeX source\n",
	"and make it possible to go from a DVI file to the\n",
	"(roughly) corresponding place in the TeX source and back\n",
	"(this is also called `reverse search' and `forward search').\n",
	"\n",
	"On the TeX side, you need a TeX version that supports the `-src'\n",
	"option (e.g. teTeX >= 2.0) or a macro package like srcltx.sty\n",
	"to insert the specials into the DVI file.\n",
	"\n",
	"Source special mode can be customized for various editors\n",
	"by using the command line option \"-editor\" or one of the\n",
	"environment variables \"XEDITOR\", \"VISUAL\" or \"EDITOR\".\n",
	"See the xdvi man page on the \"-editor\" option for details\n",
	"and examples.\n",
	"\n",
	"Forward search can be performed by a program (i.e. your editor)\n",
	"invoking xdvi with the \"-sourceposition\" option like this:\n",
	"xdvi -sourceposition \"<line> <filename>\" <main file>\n",
	"If there is already an instance of xdvi running that displays\n",
	"<main file>, it will try to open the page specified by\n",
	"<line> and <filename> an highlight this location on the page.\n",
	"Else, a new instance of xdvi will be started that will try to\n",
	"do the same.\n",
	"\n",
	"The following keybindings are pre-configured:\n",
	"\n",
	"Ctrl-Mouse1\n",
	"     [source-special()] Invoke the editor (the value\n",
	"     of the \"editor\" resource ) to display the line in the\n",
	"     TeX file corresponding to special at cursor position.\n",
	"\n",
	"Ctrl-v\n",
	"     [show-source-specials()]  Show bounding boxes for every\n",
	"     source special on the current page, and print the strings\n",
	"     contained in these specials to  stderr. With prefix 1,\n",
	"     show every bounding box on the page (for debugging purposes).\n",
	"\n",
	"Ctrl-x\n",
	"     [source-what-special()]  Display information about the\n",
	"     source special next to the cursor, similar to\n",
	"     \"source-special()\", but without actually invoking\n",
	"     the editor (for debugging purposes).\n",
	"\n",
	NULL
    };

    static _Xconst char *default_help_rulermode[] = {
	"Ruler Mode\tMeasuring distances in the DVI file\n",
	"Note: Unless a key binding also has an uppercase version,\n",
	"all bindings are case-insensitive.\n\n",
	"This is a mode that can be toggled either via the `Options'\n",
	"menu or via the keystroke Ctrl-m. It provides a simple way of\n",
	"measuring distances on the page.\n",
	"\n",
	"When this mode is activated, the mouse cursor changes into a\n",
	"thin cross, and a larger, cross-shaped ruler is drawn in the\n",
	"highlight color at the mouse location. The ruler doesn't have\n",
	"units attached to it; instead, the current distance between the\n",
	"ruler and the mouse cursor is continously printed to the\n",
	"statusline.\n",
	"\n",
	"When activating Ruler Mode, the ruler is at first attached to\n",
	"the mouse and can be moved around. It can then be positioned at\n",
	"a fixed place by clicking Mouse-1. After that, the mouse cursor\n",
	"can be moved to measure the horizontal (dx), vertical (dy) and\n",
	"direct/shortest (dr) distance between the ruler center point\n",
	"and the mouse.\n",
	"\n",
	"Clicking Mouse-1 again will snap the ruler to the current mouse\n",
	"position, and holding down Mouse-1 will move the ruler around.\n",
	"\n",
	"In Ruler Mode, the following special keybindings extend or\n",
	"replace the default bindings:\n",
	"\n",
	"t	Toggles between various ruler units.\n",
	"o	Snaps the ruler back to the origin (the upper-left\n",
	"	corner).\n",
	"P	Prints the distances shown in the statusline to standard\n",
	"	output.\n",
	"\n",
	NULL
    };

    static _Xconst char *default_help_search[] = {
	"String Search\tSearching for strings in the DVI file\n",
        "The keystroke Ctrl-f or the menu entry File->Find ...\n",
        "opens a dialog window to search for a text string or a\n",
        "regular expression in the DVI file.\n",
#ifdef MOTIF
	"(With Motif, you can also click on the `Binoculars' symbol in the toolbar.)\n",
#endif
        "\n",
        "The search term should be a string in the encoding specified\n",
        "by the X resource/command-line option textEncoding;\n",
        "currently, only the values iso_8859-1 and utf-8 are suported.\n",
        "Internally, the text in the DVI file is represented in\n",
        "UTF-8 encoding (you can view the text by saving the DVI\n",
        "file to a text file in UTF-8 encoding via the `File -> Save As ...'\n",
        "dialog).\n",
	"To match a newline character, use `\\n' in the search string;\n",
	"to match the string `\\n', use `\\\\n'.\n",
        "\n",
        "If the checkbox Regular Expression is activated, the\n",
        "string is teated as a regular expression in extended POSIX\n",
        "format, with the following properties:\n",
        "\n",
        " - a? matches a zero or one times\n",
        "\n",
        " - a* matches a zero or more times\n",
        "\n",
        " - a+ matches a one or more times. Note that * and + are\n",
        "   greedy, i.e. they match the longest possible\n",
        "   sub string.\n",
        "\n",
        " - a{n} matches a exactly n times\n",
        "\n",
        " - a{n,m} matches a at least n and no more than m times\n",
        "\n",
        " - a|b matches a or b. Brackets can be used for grouping,\n",
        "   e.g.: (a|b)|c.\n",
        "\n",
        " - The string matched by the nth group can be referenced\n",
        "   by \\n, e.g. \\1 refers to the first match.\n",
        "\n",
        " - The characters ^ and $ match the beginning and the end\n",
        "   of a line, respectively.\n",
        "\n",
        " - [abc] matches any of the letters a, b, c, and [a-z]\n",
        "   matches all characters from a to z.\n",
        "\n",
        " - The patterns . and [...] without an explicit newline\n",
        "   don't match a newline character.\n",
        "\n",
        " - Each item in a regular expression can also be one of\n",
        "   the following POSIX character classes:\n",
        "   [[:alnum:]] [[:alpha:]] [[:blank:]] [[:cntrl:]] [[:digit:]]\n",
        "   [[:graph:]] [[:lower:]] [[:print:]] [[:space:]] [[:upper:]]\n",
        "\n",
        "   These can be negated by inserting a ^ symbol after the\n",
        "   first bracket: [^[:alpha:]]\n",
        "\n",
        "   For more details on POSIX regular expressions, see\n",
        "   e.g. the IEEE Std 1003.1 available online from:\n",
        "\n",
        "   http://www.opengroup.org/onlinepubs/007904975/basedefs/xbd_chap09.html\n",
	"\n",
        " - As a non-standard extension, the following Perl-like\n",
        "   abbreviations can be used instead of the POSIX classes:\n",
        "\n",
        "\n",
        "      Symbol Meaning                       POSIX Class\n",
        "\n",
        "        \\w   an alphanumeric character     [[:alnum:]]\n",
        "        \\W   a non-alphanumeric character  [^[:alnum:]]\n",
        "        \\d   a digit character             [[:digit:]]\n",
        "        \\D   a non-digit character         [^[:digit:]]\n",
        "        \\s   a whitespace character        [[:space:]]\n",
        "        \\S   a non-whitespace character    [^[:space:]]\n",
        "\n",
        " - The following characters are special symbols; they\n",
        "   need to be escaped with \\ in order to match them\n",
        "   literally: ( ) [ ] . * ? + ^ $ \\.\n",
        "\n",
        "The dialog also provides checkboxes to search backwards\n",
        "and match in a case-sensitive manner (the default is to\n",
        "ignore case, i.e. a search string Test will match both\n",
        "the strings test and TEST in the DVI file).\n",
	"\n",
	NULL
    };
    
    if (initialized)
	return;

    k = 0;


    get_resource_setting(resource.help_general, default_help_general,
			 &(widget_info_list[k].resource_string),
			 &(widget_info_list[k].title),
			 &(widget_info_list[k].summary));
    widget_info_list[k].name = widget_info_list[k].title;
    k++;

    get_resource_setting(resource.help_pagemotion, default_help_pagemotion,
			 &(widget_info_list[k].resource_string),
			 &(widget_info_list[k].title),
			 &(widget_info_list[k].summary));
    widget_info_list[k].name = widget_info_list[k].title;
    k++;

    get_resource_setting(resource.help_othercommands, default_help_othercommands,
			 &(widget_info_list[k].resource_string),
			 &(widget_info_list[k].title),
			 &(widget_info_list[k].summary));
    widget_info_list[k].name = widget_info_list[k].title;
    k++;

#ifdef HTEX
    get_resource_setting(resource.help_hypertex, default_help_hypertex,
			 &(widget_info_list[k].resource_string),
			 &(widget_info_list[k].title),
			 &(widget_info_list[k].summary));
    widget_info_list[k].name = widget_info_list[k].title;
    k++;
#endif

    get_resource_setting(resource.help_mousebuttons, default_help_mousebuttons,
			 &(widget_info_list[k].resource_string),
			 &(widget_info_list[k].title),
			 &(widget_info_list[k].summary));
    widget_info_list[k].name = widget_info_list[k].title;
    k++;

    get_resource_setting(resource.help_rulermode, default_help_rulermode,
			 &(widget_info_list[k].resource_string),
			 &(widget_info_list[k].title),
			 &(widget_info_list[k].summary));
    widget_info_list[k].name = widget_info_list[k].title;
    k++;

    get_resource_setting(resource.help_search, default_help_search,
			 &(widget_info_list[k].resource_string),
			 &(widget_info_list[k].title),
			 &(widget_info_list[k].summary));
    widget_info_list[k].name = widget_info_list[k].title;
    k++;

    get_resource_setting(resource.help_pagemotion, default_help_marking,
			 &(widget_info_list[k].resource_string),
			 &(widget_info_list[k].title),
			 &(widget_info_list[k].summary));
    widget_info_list[k].name = widget_info_list[k].title;
    k++;

    get_resource_setting(resource.help_sourcespecials, default_help_sourcespecials,
			 &(widget_info_list[k].resource_string),
			 &(widget_info_list[k].title),
			 &(widget_info_list[k].summary));
    widget_info_list[k].name = widget_info_list[k].title;
    k++;

    widget_info_list[k].name = NULL;
    widget_info_list[k].title = NULL;
    widget_info_list[k].resource_string = NULL;
    widget_info_list[k].default_resource_string = NULL;

    ASSERT(k < WIDGET_INFO_LIST_LEN, "Too many elements for widget_info_list[]");

    initialized = True;
}



#if !MOTIF
static XtActionsRec popdown_actions[] = {
    {"WM_popdown",		xaw_popdown },
};
#endif

static char *text_translations = "#override \n"
"<Key>q: close-help()\n"
#if MOTIF
/*
 * We are overriding this since otherwise the Text widget would attempt to
 * first move the invisible cursor instead of the scroll bars.
 */
/* FIXME: This gives a warning:
   Warning: Actions not found: scroll-one-line-down, scroll-one-line-up
*/
/* "<Key>osfUp: scroll-one-line-down()\n" */
/* "<Key>osfDown: scroll-one-line-up()\n" */
#endif
;

/*
 * Pops up help window with menu bar.
 */
void
show_help(Widget toplevel)
{
    Widget helppaned, form, left_form, right_form;
    Widget action_area, ok_button, cancel_button;
    
    static Boolean help_created = False;
    Atom WM_DELETE_WINDOW;
    
    initialize_resources();
    
    if (!help_created) {
	/* called 1st time; create all widgets. Note: widgets are not
	 * destroyed by popdown_help_callback. */
#if !MOTIF
	XtTranslations wm_translations;
	XtAddActions(popdown_actions, XtNumber(popdown_actions));
	wm_translations = XtParseTranslationTable("<Message>WM_PROTOCOLS: WM_popdown()");
#endif
	
	help_shell = XtVaCreatePopupShell("helpWindow", SHELL_WIDGET_CLASS, toplevel,
					  XtNvisual, our_visual,
					  XtNtitle, "Xdvik: Help",
					  XtNcolormap, our_colormap,
					  HELP_SHELL_ARGS);

#if MOTIF
	/* make the window manager destroy action pop down the window */
	WM_DELETE_WINDOW = XmInternAtom(XtDisplay(help_shell), "WM_DELETE_WINDOW", False);
	XmAddWMProtocolCallback(help_shell, WM_DELETE_WINDOW, popdown_help_callback, NULL);
#else
	WM_DELETE_WINDOW = XInternAtom(XtDisplay(help_shell), "WM_DELETE_WINDOW", False);
#endif
	helppaned = XtVaCreateWidget("helppaned", PANED_WIDGET_CLASS, help_shell, HELP_PANED_ARGS);
	XtOverrideTranslations(helppaned, XtParseTranslationTable(text_translations));

	form = XtVaCreateWidget("form", FORM_WIDGET_CLASS, helppaned, FORM_ARGS);

	/* left pane for topics selection */
	left_form = XtVaCreateWidget("left_form", FORM_WIDGET_CLASS, form, LEFT_FORM_ARGS);
	topics_list = create_list_widget(left_form, "topics_list");
	
	XtManageChild(left_form);
	
	/* right form for help text display */
	right_form = XtVaCreateWidget("right_form", FORM_WIDGET_CLASS, form, RIGHT_FORM_ARGS);

	help_label = create_label_widget(right_form, "help_label",
					 widget_info_list[0].title, widget_info_list[0].summary);
	
	help_text = create_help_text(right_form, help_label, "help_text", widget_info_list[0].resource_string);
	XtOverrideTranslations(help_text, XtParseTranslationTable(text_translations));

#if !MOTIF
	{
	    Dimension w;
	    XtVaGetValues(help_text, XtNwidth, &w, NULL);
	    XtVaSetValues(help_label, XtNwidth, w, NULL);
	}
#endif
	XtManageChild(right_form);

	/* action area at bottom */
	action_area = XtVaCreateWidget("action_area", FORM_WIDGET_CLASS, helppaned, ACTION_AREA_ARGS);
	
	ok_button = create_button(action_area, "Close", True, True);
	cancel_button = create_button(action_area, "Cancel", False, False);

	adjust_width(ok_button, cancel_button);

	/* don't show the cancel button for help windows */
	XtUnmanageChild(cancel_button);
	
	XtManageChild(action_area);

	XtManageChild(form);

	XtManageChild(helppaned);
	
	XtRealizeWidget(help_shell);

#if MOTIF
	/* enable OK button */
	XmProcessTraversal(ok_button, XmTRAVERSE_CURRENT);
	XtOverrideTranslations(ok_button,
			       XtParseTranslationTable("<Key>Return:ArmAndActivate()\n"
						       "<Key>q:ArmAndActivate()"));
	XtAddCallback(ok_button, XmNactivateCallback, popdown_help_callback, NULL);
	XmAddWMProtocolCallback(help_shell, WM_DELETE_WINDOW, popdown_help_callback, NULL);
	{ /* disable resizing of lower part of pane (and button) */
	    Dimension h;
	    XtVaGetValues(ok_button, XmNheight, &h, NULL);
	    XtVaSetValues(action_area,
			  XmNpaneMaximum, h + 20,
			  XmNpaneMinimum, h + 20,
			  NULL);
	}
#else
	XSetWMProtocols(XtDisplay(help_shell), XtWindow(help_shell), &WM_DELETE_WINDOW, 1);
	XtAddCallback(ok_button, XtNcallback, popdown_help_callback, NULL);
#endif
	center_window(help_shell, top_level);
	help_created = True;
    }
    
    /* finally, pop up help window */
    if (help_active)
	return;
    
#if MOTIF
    { /* check if resources are set properly */
	Dimension w, h;
	XtVaGetValues(help_shell, XtNwidth, &w, XtNheight, &h, NULL);
	if (h < 200 || w < 400) {
	    XDVI_WARNING((stderr, "Initial help window size too small (%dx%d); overriding size.\n"
			  "Please check/update your application defaults file, and set both of\n"
			  "`XDvi*help_text.rows' and `XDvi*help_text.columns' to a realistic value.",
			  h, w));
	    XtVaSetValues(help_shell, XtNwidth, 620, XtNheight, 520, NULL);
	}
    }
#endif
    help_active = 1;
    XtPopup(help_shell, XtGrabNone);
}
