/*------------------------------------------------------------
message-window.c: message popups for xdvi.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL PAUL VOJTA OR ANY OTHER AUTHOR OF THIS SOFTWARE BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

------------------------------------------------------------*/



/*

============================================
Suggested Policy for using the GUI messages:
============================================

- Use the statusline for shorter messages, a message window for more
  important messages or such where you'd like to give further help info
  (see the `helptext' argument of popup_message()). When in doubt,
  prefer the statusline (excessive use of popup windows is a nuisance
  for the user).

- Don't use any of the GUI messages to report internal workings of
  the program; for important internal information, there should be a
  debugging setting to print it to stderr. Use the GUI messages
  only in situations such as the following:

  - to give the user feedback on actions that (s)he initiated

  - to indicate that an internal action causes a delay perceptible
    by the user (as a rough guide: a delay of more than half a second)

  - to report situations that might require new actions by the user.

*/

#include <ctype.h>

#include "xdvi-config.h"
#include "xdvi.h"
#include "string-utils.h"

/* Xaw specific stuff */
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#ifdef MOTIF
# include <Xm/DialogS.h>
# include <Xm/MessageB.h>
# include <Xm/Label.h>
# include <Xm/Form.h>
# include <Xm/MenuShell.h>
# include <Xm/Protocols.h>
#else
# include <X11/Xaw/Paned.h>
# include <X11/Xaw/Box.h>
# include <X11/Xaw/MenuButton.h>
# include <X11/Xaw/SimpleMenu.h>
# include <X11/Xaw/Sme.h>
# include <X11/Xaw/SmeBSB.h>
# include <X11/Xaw/AsciiText.h>
# include <X11/Xaw/Dialog.h>
#endif

#include "kpathsea/c-vararg.h"
#include "xdvi.h"
#include "util.h"
#include "string-utils.h"
#include "x_util.h"
#include "message-window.h"

/* have no more than MAX_POPUPS open simultaneously */
#define MAX_POPUPS 10

/* offset for cascading popups */
#define POPUP_OFFSET ((my_popup_num * 20))

#ifdef MOTIF
/* wrap messages after MSG_WRAP_LEN characters at whitespace */
#define MSG_WRAP_LEN 60
#endif

/* array of active popups: */
static int g_popup_array[MAX_POPUPS];

static Atom WM_DELETE_WINDOW;

/* arrays for saving the widgets of multiple popups;
 * same index as in g_popup_array:
 */
#ifdef MOTIF
static Widget popup_window[MAX_POPUPS], dialog[MAX_POPUPS];
#else
static Widget popup_window[MAX_POPUPS], message_box[MAX_POPUPS],
    message_text[MAX_POPUPS], message_paned[MAX_POPUPS],
    message_ok[MAX_POPUPS], message_help[MAX_POPUPS], message_not_ok[MAX_POPUPS];
#endif

/* map popup_message_t's to strings/motif dialog elements */
static struct message_map {
    char *window_title;
#ifdef MOTIF
    int motif_msg_type;
#endif
} my_msg_map[] = {
    { "Xdvi Question"
#ifdef MOTIF
      , XmDIALOG_QUESTION
#endif
    }, /* MSG_QUESTION */
    { "Xdvi Help"
#ifdef MOTIF
      , XmDIALOG_INFORMATION
#endif
    }, /* MSG_HELP */
    { "Xdvi Info"
#ifdef MOTIF
      , XmDIALOG_INFORMATION
#endif
    }, /* MSG_INFO */
    { "Xdvi Warning"
#ifdef MOTIF
      , XmDIALOG_WARNING
#endif
    }, /* MSG_WARN */
    { "Xdvi Error"
#ifdef MOTIF
      , XmDIALOG_ERROR
#endif
    }, /* MSG_ERR */
};

typedef struct ok_or_cancel_cb_t_ {
    void (*callback)(); /* a callback function with signature `void callback(XtPointer arg)' */
    XtPointer arg;	/* arg for callback function */
    int idx;		/* index of this popup */
} ok_or_cancel_cb_t;

static ok_or_cancel_cb_t ok_callbacks[MAX_POPUPS], cancel_callbacks[MAX_POPUPS];


#ifndef MOTIF
static void
xaw_popdown_cancel(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    size_t idx;
    ok_or_cancel_cb_t cb;
    
    UNUSED(w);
    UNUSED(event);
    UNUSED(params);
    UNUSED(num_params);

    ASSERT(*num_params == 1, "Wrong number of parameters in callback");
    idx = strtoul(*params, (char **)NULL, 10);

    /*
      NOTE: first pop down window and mark its position as free, then
      invoke the callback.  The reason for this is that the callback
      may need to wait for open windows.
    */
    XtPopdown(popup_window[idx]);
    XtDestroyWidget(popup_window[idx]);
    g_popup_array[idx] = 0;
    XSync(DISP, True);
    
    cb = cancel_callbacks[idx];
    if (cb.callback != NULL) {
	cb.callback(cb.arg);
    }
}

static void
xaw_popdown(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    size_t idx;

    UNUSED(w);
    UNUSED(event);
    UNUSED(params);
    UNUSED(num_params);

    ASSERT(*num_params == 1, "Wrong number of parameters in callback");
    idx = strtoul(*params, (char **)NULL, 10);

    /*
      NOTE: first pop down window and mark its position as free, then
      invoke the callback.  The reason for this is that the callback
      may need to wait for open windows.
    */
    XtPopdown(popup_window[idx]);
    XtDestroyWidget(popup_window[idx]);
    g_popup_array[idx] = 0;
    XSync(DISP, True);
}

static XtActionsRec popdown_actions[] = {
    {"WM_popdown",		xaw_popdown_cancel },
    {"close-popup-cancel",	xaw_popdown_cancel },
    {"close-popup",		xaw_popdown	   }
};
#endif

static void
ok_action(Widget w, XtPointer client_data, XtPointer call_data)
{
    ok_or_cancel_cb_t *cb = (ok_or_cancel_cb_t *)client_data;
    int cnt = cb->idx;
    
    UNUSED(w);
    UNUSED(call_data);
    
#if DEBUG
    fprintf(stderr, "quit_action called for popup %d\n", cnt);
#endif
    ASSERT(cnt >= 0 && cnt < MAX_POPUPS, "Invalid widget index in ok_action()");

    /*
      NOTE: first pop down window and mark its position as free, then
      invoke the callback.  The reason for this is that the callback
      may need to wait for open windows.
    */
    XtPopdown(popup_window[cnt]);
    XtDestroyWidget(popup_window[cnt]);
    g_popup_array[cnt] = 0;
    XSync(DISP, True);
    
    if (cb->callback != NULL) {
	cb->callback(cb->arg);
    }
}

/*------------------------------------------------------------
 *  help_action
 *
 *  Arguments:
 *	Widget w, XtPointer call_data
 *		- (ignored)
 *	XtPointer client_data
 *		- the help string
 *
 *  Returns:
 *	void
 *
 *  Purpose:
 *	Callback for the `Help' button; opens another window
 *	containing the help text. The new window won't have
 *	another `Help' button.
 *------------------------------------------------------------*/

static void
help_action(Widget w, XtPointer client_data, XtPointer call_data)
{
    UNUSED(w);
    UNUSED(call_data);

    /* open another window with the help text */
    popup_message(MSG_HELP, NULL, client_data);
}


/*
  callback for cancel button in ok-or-cancel dialog. It's assumed
  that call_data is a ptr to Boolean that is True by default; this
  callback sets it to false.
*/
static void
cancel_action(Widget w, XtPointer client_data, XtPointer call_data)
{
    ok_or_cancel_cb_t *cb = (ok_or_cancel_cb_t *)client_data;
    int cnt = cb->idx;
    
    UNUSED(w);
    UNUSED(call_data);
    ASSERT(cnt >= 0 && cnt < MAX_POPUPS, "Invalid widget index in cancel_action()");

    /*
      NOTE: first pop down window and mark its position as free, then
      invoke the callback.  The reason for this is that the callback
      may need to wait for open windows.
    */
    XtPopdown(popup_window[cnt]);
    XtDestroyWidget(popup_window[cnt]);
    g_popup_array[cnt] = 0;
    XSync(DISP, True);

    /* invoke the callback if present */
    if (cb->callback != NULL) {
	(cb->callback)(cb->arg);
    }
}

/*------------------------------------------------------------
 *  create_dialogs
 *
 *  Arguments:
 *	Widget toplevel - parent for the dialog window
 *
 *	helptext	- if not-NULL, create an additional
 *			  `help' button
 *
 *	cnt		- number of current popup dialog
 *
 *  Returns:
 *	void
 *
 *  Purpose:
 *	Create message dialog widgets
 *------------------------------------------------------------*/

static void
create_dialogs(popup_message_sizehint_t size,
	       Widget toplevel,
	       int cnt,
	       const char *helptext,
	       const char *ok_button, void(*ok_function)(), XtPointer ok_arg,
	       const char *cancel_button, void(*cancel_function)(), XtPointer cancel_arg)
{
    Widget new_popup_window;
#ifdef MOTIF
    Widget new_dialog;
    UNUSED(size);
#else
    char *translations_str = NULL, *key_translations_str = NULL;
    int msg_w = 400, msg_h = 100;
    Widget new_message_paned, new_message_text, new_message_box, new_message_ok,
	new_message_help = 0, new_message_not_ok;
    XtTranslations wm_translations, key_translations;
#endif
    
    /* save callbacks for `OK' and `Cancel' buttons to global array */
    ok_callbacks[cnt].callback = ok_function;
    ok_callbacks[cnt].arg = ok_arg;
    ok_callbacks[cnt].idx = cnt;

    cancel_callbacks[cnt].callback = cancel_function;
    cancel_callbacks[cnt].arg = cancel_arg;
    cancel_callbacks[cnt].idx = cnt;

#ifndef MOTIF
    /* get index into WM_popdown arg */
    translations_str = get_string_va("<Message>WM_PROTOCOLS: WM_popdown(%d)", cnt);
    XtAddActions(popdown_actions, XtNumber(popdown_actions));
    wm_translations = XtParseTranslationTable(translations_str);
    free(translations_str);
#endif

    new_popup_window = XtVaCreatePopupShell("xdvi",
#ifdef MOTIF
					     xmDialogShellWidgetClass, toplevel,
					     XmNdeleteResponse, XmDO_NOTHING, /* we'll take care of that ourselves */
#else
					     transientShellWidgetClass, toplevel,
					     XtNx, 60,
					     XtNy, 80,
					     XtNtranslations, wm_translations,
					     XtNtransientFor, top_level,
					     XtNaccelerators, accels_cr,
#endif
					     XtNvisual, our_visual,
					     XtNcolormap, our_colormap,
					     XtNmappedWhenManaged, False,
					     NULL);

    
#ifdef MOTIF

    WM_DELETE_WINDOW = XmInternAtom(XtDisplay(new_popup_window), "WM_DELETE_WINDOW", False);
    XmAddWMProtocolCallback(new_popup_window, WM_DELETE_WINDOW, cancel_action, (XtPointer)&(cancel_callbacks[cnt]));
    
    new_dialog = XtVaCreateWidget("dialog", xmMessageBoxWidgetClass, new_popup_window,
				  XmNdialogType, XmDIALOG_WARNING,
				  XmNtraversalOn, True,
				  XmNhighlightOnEnter, True,
				  NULL);
    XtAddCallback(new_dialog, XmNokCallback, ok_action, (XtPointer)&(ok_callbacks[cnt]));
    
    if (cancel_button != NULL) {
	XmString cancel_label = XmStringCreateLtoR((char *)cancel_button, g_charset);
	XtVaSetValues(XmMessageBoxGetChild(new_dialog, XmDIALOG_OK_BUTTON),
		      XmNlabelString, cancel_label, NULL);
	XmStringFree(cancel_label);
	XtAddCallback(new_dialog, XmNcancelCallback, cancel_action, (XtPointer)(&cancel_callbacks[cnt]));
    }
    else {
	XtUnmanageChild(XmMessageBoxGetChild(new_dialog, XmDIALOG_CANCEL_BUTTON));
    }
    XtInstallAllAccelerators(new_dialog,
			     XmMessageBoxGetChild(new_dialog, XmDIALOG_OK_BUTTON));

    if (helptext != NULL) {
	XtAddCallback(new_dialog, XmNhelpCallback, help_action, (XtPointer)helptext);
    }
    else {
	XtUnmanageChild(XmMessageBoxGetChild(new_dialog, XmDIALOG_HELP_BUTTON));
    }

    if (ok_button != NULL) { /* change `OK' button label */
	XmString ok_label;
	ok_label = XmStringCreateLtoR((char *)ok_button, g_charset);
	XtVaSetValues(XmMessageBoxGetChild(new_dialog, XmDIALOG_OK_BUTTON),
		      XmNlabelString, ok_label, NULL);
	XmStringFree(ok_label);
    }
    
    /* insert the new widgets into the global arrays */
    dialog[cnt] = new_dialog;

#else /* MOTIF */
    switch (size) {
    case SIZE_SMALL:
	msg_w = 400;
	msg_h = 100;
	break;
    case SIZE_MEDIUM:
	msg_w = 430;
	msg_h = 160;
	break;
    case SIZE_LARGE:
	msg_w = 450;
	msg_h = 180;
	break;
    }
    WM_DELETE_WINDOW = XInternAtom(XtDisplay(new_popup_window), "WM_DELETE_WINDOW", False);
    
    new_message_paned = XtVaCreateManagedWidget("message_paned", panedWidgetClass, new_popup_window,
						XtNaccelerators, accels_cr,
						NULL);

    new_message_text = XtVaCreateManagedWidget("message_text", asciiTextWidgetClass, new_message_paned,
/* 					       XtNheight, 100, */
/* 					       XtNwidth, 400, */
					       XtNwidth, msg_w,
					       XtNheight, msg_h,
					       /* wrap horizontally instead of scrolling
						* TODO: this won't work for the first widget instance?
						*/
					       XtNwrap, XawtextWrapWord,
					       XtNscrollVertical, XAW_SCROLL_ALWAYS,
					       XtNeditType, XawtextRead,
					       XtNinput, True,
					       XtNdisplayCaret, False,
					       XtNleftMargin, 5,
					       XtNaccelerators, accels_cr,
					       NULL);
    /* box for the OK/Cancel button */
    new_message_box = XtVaCreateManagedWidget("message_box", formWidgetClass, new_message_paned,
					      /* resizing by user isn't needed */
					      XtNshowGrip, False,
					      XtNdefaultDistance, 6, /* some padding */
					      /* resizing the window shouldn't influence this box,
					       * but only the text widget
					       */
					      XtNskipAdjust, True,
					      XtNaccelerators, accels_cr,
					      NULL);

    new_message_ok = XtVaCreateManagedWidget(ok_button == NULL ? "OK" : ok_button,
					     commandWidgetClass, new_message_box,
					     XtNtop, XtChainTop,
 					     XtNbottom, XtChainBottom,
 					     XtNleft, XtChainLeft,
 					     XtNright, XtChainLeft,
					     XtNaccelerators, accels_cr,
					     NULL);
    /* add quit_action callback for the "OK" button */
    /* FIXME: how to make accelerators be accepted by new_popup_window as well? */
    key_translations_str = get_string_va("#override\n<Key>q:close-popup-cancel(%d)\n"
					 "<Key>Return:close-popup-cancel(%d)\n", cnt, cnt);
    key_translations = XtParseTranslationTable(key_translations_str);
    free(key_translations_str);
    XtOverrideTranslations(new_popup_window, key_translations);
    XtOverrideTranslations(new_message_paned, key_translations);
    XtOverrideTranslations(new_message_text, key_translations);
    
    XtInstallAllAccelerators(new_message_box, new_message_ok);
    XtAddCallback(new_message_ok, XtNcallback, ok_action, (XtPointer)&(ok_callbacks[cnt]));

    /* we create additional buttons in any case,
       to make the sizing more consistent */
    new_message_help = XtVaCreateManagedWidget("Help", commandWidgetClass, new_message_box,
					       XtNtop, XtChainTop,
					       XtNfromHoriz, new_message_ok,
					       XtNbottom, XtChainBottom,
					       XtNleft, XtChainRight,
					       XtNright, XtChainRight,
					       XtNaccelerators, accels_cr,
					       NULL);
    message_help[cnt] = new_message_help;
    
    /* add cancel button */
    new_message_not_ok = XtVaCreateManagedWidget(cancel_button == NULL ? "Cancel" : cancel_button,
						 commandWidgetClass, new_message_box,
						 XtNtop, XtChainTop,
						 XtNfromHoriz, new_message_ok,
						 XtNbottom, XtChainBottom,
						 XtNleft, helptext == NULL ? XtChainRight : XtChainLeft,
						 XtNright, helptext == NULL ? XtChainRight : XtChainLeft,
						 XtNaccelerators, accels_cr,
						 NULL);
    message_not_ok[cnt] = new_message_not_ok;

    { /* set all buttons to same size */
	Dimension w1, w2, w3, max;
	XtVaGetValues(new_message_ok, XtNwidth, &w1, NULL);
	XtVaGetValues(new_message_help, XtNwidth, &w2, NULL);
	XtVaGetValues(new_message_not_ok, XtNwidth, &w3, NULL);
	max = w1;
	if (w2 > max)
	    max = w2;
	if (w3 > max)
	    max = w3;
	XtVaSetValues(new_message_ok, XtNwidth, max, NULL);
	XtVaSetValues(new_message_help, XtNwidth, max, NULL);
	XtVaSetValues(new_message_not_ok, XtNwidth, max, NULL);
    }
    /* if helptext argument is not-NULL, add help_action callback,
       else unmanage help button */
    if (helptext != NULL) {
	XtAddCallback(new_message_help, XtNcallback, help_action, (XtPointer)helptext);
    }
    else {
	XtUnmanageChild(new_message_help);
    }

    if (cancel_button != NULL) {
	XtAddCallback(new_message_not_ok, XtNcallback, cancel_action, (XtPointer)&(cancel_callbacks[cnt]));
    }
    else {
	XtUnmanageChild(new_message_not_ok);
    }
    /* insert the new widgets into the global arrays */
    message_box[cnt] = new_message_box;
    message_paned[cnt] = new_message_paned;
    message_text[cnt] = new_message_text;
    message_ok[cnt] = new_message_ok;

#endif /* MOTIF */
    popup_window[cnt] = new_popup_window;
}


/*
 * Popup a window with wrapped text in it.
 * For Motif, the text is explicitly wrapped inside this method.
 */
static void
internal_popup_window(popup_message_sizehint_t size,
		      popup_message_t type,
		      int x_coord, int y_coord,
		      const char *helptext,
		      char *msg_buf,
#ifndef MOTIF
		      const char *xaw_ret_action_str,
#endif
		      const char *ok_button, void(*ok_callback)(), const XtPointer ok_arg,
		      const char *cancel_button, void(*cancel_callback)(), const XtPointer cancel_arg)
{
    char win_title[64];
    int my_popup_num = 0;
#ifdef MOTIF
    XmString str;
#endif
    ASSERT(type < (sizeof my_msg_map / sizeof my_msg_map[0]), "too few elements in my_msg_map");
    sprintf(win_title, my_msg_map[type].window_title);

#if DEBUG
    fprintf(stderr, "internal_popup_window called with prompt: \"%s\"\n", msg_buf);
#endif

    if (!XtIsRealized(top_level)) {
	/*
	  Additionally dump messages to STDERR, since in some cases
	  such as non-found T1 resources, it might cause mktexpk to be
	  run before the main window comes up. All texts must be
	  passed as arguments to fprintf (NOT puts), since they are
	  supposed to be printf-format strings (e.g. with doubled `%'
	  to escape them):
	*/
	fprintf(stderr, "\n%s:\n", my_msg_map[type].window_title);
	fprintf(stderr, msg_buf);
	fprintf(stderr, "\n");
	if (helptext) {
	    fprintf(stderr, "---------- helptext ----------\n");
	    fprintf(stderr, helptext);
	    fprintf(stderr, "\n---------- end of helptext ----------\n");
	}
	fprintf(stderr, "\n");
    }
    /* search for first free position in g_popup_array */
    while (my_popup_num < MAX_POPUPS && (g_popup_array[my_popup_num] == 1)) {
	my_popup_num++;
    }
    if (my_popup_num == MAX_POPUPS) {
	/* already enough popups on screen, just dump it to stderr */
	fprintf(stderr, "%s:\n", win_title);
	fprintf(stderr, msg_buf);
	/* Note: If a mad function continues to open popups, this will
	 * stop after MAX_POPUPS, but open a new window for each
	 * window the user pops down. Maybe we ought to do something
	 * about this.
	 */
	return;
    }
    else {
	/* mark it as non-free */
	g_popup_array[my_popup_num] = 1;
    }
#if DEBUG
    fprintf(stderr, "first free position in g_popup_array: %d\n", my_popup_num);
#endif

    /* create a new set of widgets for the additional popup. */
    create_dialogs(size, top_level,
		   my_popup_num,
		   helptext,
		   ok_button, ok_callback, ok_arg,
		   cancel_button, cancel_callback, cancel_arg);
#ifdef MOTIF
    XtVaSetValues(popup_window[my_popup_num], XmNtitle, win_title, NULL);
    XtVaSetValues(dialog[my_popup_num], XmNdialogType, my_msg_map[type].motif_msg_type, NULL);
    { /* wrap message at space before MSG_WRAP_LEN */
	char *testwrap = msg_buf;
	int ctr;
	for (ctr = 0; *testwrap++; ctr++) {
	    if (*testwrap == '\n') {
		ctr = 0;
	    }
	    else if (ctr > MSG_WRAP_LEN) {
		size_t before_len = 0, after_len = 0;
		char *before_ptr, *after_ptr;
		before_ptr = after_ptr = testwrap;
		/* try to find shortest sequence before or after point to wrap at;
		   this seems to give the most pleasing results.
		 */
		while (before_ptr > msg_buf && !isspace((int)*--before_ptr)) {
		    before_len++;
		}
		while (*after_ptr != '\0' && !isspace((int)*++after_ptr)) {
		    after_len++;
		}

		if (before_len < after_len && isspace((int)*before_ptr)) {
		    /* use last in sequence of multiple spaces */
		    while (isspace((int)*++before_ptr)) { ; }
		    /* back up, and wrap */
		    *--before_ptr = '\n';
		    ctr = 0;
		}
		else if (isspace((int)*after_ptr)) {
		    /* use last in sequence of multiple spaces */
		    while (isspace((int)*++after_ptr)) { ; }
		    /* back up, and wrap */
		    *--after_ptr = '\n';
		    ctr = 0;
		}
	    }
	}
    }
    str = XmStringCreateLtoR((char *) msg_buf, g_charset);
    XtVaSetValues(dialog[my_popup_num],
		  XmNmessageString, str,
		  XmNtraversalOn, True,
		  XmNhighlightOnEnter, True,
		  NULL);
    XmStringFree(str);

    XtManageChild(dialog[my_popup_num]);
    
    if (x_coord > 0 && y_coord > 0) {
	position_window(XtParent(dialog[my_popup_num]), (Position)x_coord, (Position)y_coord);
    }
    
    XtPopup(XtParent(dialog[my_popup_num]), XtGrabNone);

#else /* MOTIF */

    /* add a binding of xaw_ret_action_str to <Return> to relevant widgets */
    if (xaw_ret_action_str != NULL) {
	XtTranslations xlats;
	char *translation_str = get_string_va("<Key>Return:close-popup(%d)%s(%p)",
					      my_popup_num, xaw_ret_action_str, (void *)ok_arg);
	xlats = XtParseTranslationTable(translation_str);
	free(translation_str);
	XtOverrideTranslations(popup_window[my_popup_num], xlats);
	XtOverrideTranslations(message_paned[my_popup_num], xlats);
	XtOverrideTranslations(message_text[my_popup_num], xlats);
    }
    
    XtVaSetValues(popup_window[my_popup_num], XtNtitle, win_title, NULL);
    XtVaSetValues(message_text[my_popup_num], XtNstring, msg_buf, NULL);
    XtRealizeWidget(popup_window[my_popup_num]);

    XSetWMProtocols(XtDisplay(popup_window[my_popup_num]), XtWindow(popup_window[my_popup_num]),
		    &WM_DELETE_WINDOW, 1);

    if (x_coord <= 0 || y_coord <= 0)
	center_window(popup_window[my_popup_num], top_level);
    else
	position_window(popup_window[my_popup_num], (Position)x_coord, (Position)y_coord);
    
    if (my_popup_num > 0) {
	/* some window managers position new windows exactly above the
	   existing one; to prevent this, move it with some offset
	   from the previous one: */
	Dimension x = 0, y = 0;
	XtVaGetValues(popup_window[my_popup_num-1], XtNx, &x, XtNy, &y, NULL);
	XtVaSetValues(popup_window[my_popup_num], XtNx, x + POPUP_OFFSET, XtNy, y + POPUP_OFFSET, NULL);

    }
    XtPopup(popup_window[my_popup_num], XtGrabNone);
    if (XtIsManaged(message_not_ok[my_popup_num]) && XtIsManaged(message_help[my_popup_num])) {
	/* center the help button. This is something of a sham, since it won't
	   survive resizing; but in general most users won't resize dialogs ;-) */
	Position x1, x2, bw;
	int w, dist;
	
	XtVaGetValues(message_ok[my_popup_num], XtNx, &x1, XtNwidth, &w, XtNborderWidth, &bw, NULL);
	XtVaGetValues(message_help[my_popup_num], XtNx, &x2, NULL);
	/* following formula is measured, not calculated -
	   I have no idea why it's e.g. 2 * w, not 1.5 * w ... */
	dist = (x2 - x1 - 2 * w) / 2 - 2 * bw;
	XtVaSetValues(message_not_ok[my_popup_num], XtNhorizDistance, dist, NULL);
    }
    
#endif /* MOTIF */
}


/*------------------------------------------------------------
 *  popup_message
 *
 *  Arguments:
 *	popup_message_t - info, warning, error etc; see message-window.h for details
 *
 *	char *helptext
 *	      - if not-null, this will add a `Help'
 *		button to the message widget that pops
 *		up another message widget containing
 *		<helptext>.
 *
 *
 *	char *msg, ...
 *	      - format string followed by a variable
 *		number of arguments to be formatted.
 *
 *  Returns:
 *	void
 *
 *  Purpose:
 *	Pop up a message window containing <msg>.
 *	If there are already n popups open, will open
 *	a new one unless n >= MAX_POPUPS.
 *------------------------------------------------------------*/

void
popup_message(popup_message_t type, const char *helptext, const char *format, ...)
{
    char *msg_buf = NULL;
    
    va_list argp;
    va_start(argp, format);
    msg_buf = get_string_argp(format, argp);
    va_end(argp);
    internal_popup_window(SIZE_SMALL,
			  type,
			  -1, -1, /* just center it */
			  helptext, msg_buf,
#ifndef MOTIF
			  NULL,
#endif
			  /* no special callbacks here */
			  NULL, NULL, NULL,
			  NULL, NULL, NULL);
    free(msg_buf);
}

void
popup_message_medium(popup_message_t type, const char *helptext, const char *format, ...)
{
    char *msg_buf = NULL;
    
    va_list argp;
    va_start(argp, format);
    msg_buf = get_string_argp(format, argp);
    va_end(argp);
    internal_popup_window(SIZE_MEDIUM,
			  type,
			  -1, -1, /* just center it */
			  helptext, msg_buf,
#ifndef MOTIF
			  NULL,
#endif
			  /* empty callbacks */
			  NULL, NULL, NULL,
			  NULL, NULL, NULL);
    free(msg_buf);
}

void
positioned_popup_message(popup_message_t type,
			 int x, int y,
			 const char *helptext, const char *format, ...)
{
    char *msg_buf = NULL;
    
    va_list argp;
    va_start(argp, format);
    msg_buf = get_string_argp(format, argp);
    va_end(argp);
    internal_popup_window(SIZE_SMALL,
			  type,
			  x, y, /* position at these coordinates */
			  helptext, msg_buf,
#ifndef MOTIF
			  NULL,
#endif
			  /* empty callbacks */
			  NULL, NULL, NULL,
			  NULL, NULL, NULL);
    free(msg_buf);
}

void
choice_dialog(popup_message_t type,
	      const char *helptext,
#ifndef MOTIF
	      const char *xaw_ret_action_str,
#endif
	      const char *ok_label, void(*ok_callback)(), const XtPointer ok_arg,
	      const char *cancel_label, void(*cancel_callback)(), const XtPointer cancel_arg,
	      const char *format, ...)
{
    char *msg_buf = NULL;
    
    va_list argp;
    va_start(argp, format);
    msg_buf = get_string_argp(format, argp);
    va_end(argp);
    internal_popup_window(SIZE_SMALL,
			  type,
			  -1, -1, /* just center it */
			  helptext, msg_buf,
#ifndef MOTIF
			  xaw_ret_action_str,
#endif
			  ok_label, ok_callback, ok_arg,
			  cancel_label, cancel_callback, cancel_arg);
    free(msg_buf);
}

void
choice_dialog_medium(popup_message_t type,
		     const char *helptext,
#ifndef MOTIF
		     const char *xaw_ret_action_str,
#endif
		     const char *ok_label, void(*ok_callback)(), const XtPointer ok_arg,
		     const char *cancel_label, void(*cancel_callback)(), const XtPointer cancel_arg,
		     const char *format, ...)
{
    char *msg_buf = NULL;
    
    va_list argp;
    va_start(argp, format);
    msg_buf = get_string_argp(format, argp);
    va_end(argp);
    internal_popup_window(SIZE_MEDIUM,
			  type,
			  -1, -1, /* just center it */
			  helptext, msg_buf,
#ifndef MOTIF
			  xaw_ret_action_str,
#endif
			  ok_label, ok_callback, ok_arg,
			  cancel_label, cancel_callback, cancel_arg);
    free(msg_buf);
}

void
positioned_choice_dialog(popup_message_t type,
			 int x_pos, int y_pos,
			 const char *helptext,
#ifndef MOTIF
			 const char *xaw_ret_action_str,
#endif
			 const char *ok_label, void(*ok_callback)(), const XtPointer ok_arg,
			 const char *cancel_label, void(*cancel_callback)(), const XtPointer cancel_arg,
			 const char *format, ...)
{
    char *msg_buf = NULL;
    
    va_list argp;
    va_start(argp, format);
    msg_buf = get_string_argp(format, argp);
    va_end(argp);
    internal_popup_window(SIZE_SMALL,
			  type,
			  x_pos, y_pos,
			  helptext, msg_buf,
#ifndef MOTIF
			  xaw_ret_action_str,
#endif
			  ok_label, ok_callback, ok_arg,
			  cancel_label, cancel_callback, cancel_arg);
    free(msg_buf);
}

void
warn_overstrike(void)
{
    static Boolean warned_overstrike = False;

    if (!warned_overstrike) {
	popup_message(MSG_WARN,
			 /* helptext */
			 "Greyscaling is running in copy mode; this will cause overstrike characters to "
			 "appear incorrectly, and may result in poor display quality.  "
			 "Possible fixes are:\n"
			 "- Use the ``-thorough'' command-line option.\n"
			 "- Quit some other color-hungry applications (e.g. Netscape).\n"
			 "- Use the ``-install'' command-line option.\n"
			 "See the section ``GREYSCALING AND COLORMAPS'' in the "
			 "xdvi manual page for more details.",
			 /* text */
			 "Couldn't allocate enough colors - expect low display quality.");
	warned_overstrike = True;
    }
}

/*
  Raise any popups that currently exist; return True iff such popups
  exist, else False.
*/
Boolean
raise_message_windows(void)
{
    int i;
    Boolean have_popups = False;
    
    for (i = 0; i < MAX_POPUPS; i++) {
	if (g_popup_array[i] != 0) {
	    XRaiseWindow(DISP, XtWindow(popup_window[i]));
	    have_popups = True;
	}
    }
    
    return have_popups;
}
