#include "xdvi-config.h"

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>

#include "x_util.h"
#include "util.h"
#include "statusline.h"
#include "events.h" /* for set_bar_value() */

#define DEBUG_SCROLL_IF_NEEDED 0

Boolean do_autoscroll = False;


/* Center window wa over window wb */
void
center_window(Widget wa, Widget wb)
{
    Position x, y;
    Dimension w1, h1, w2, h2;

    if (!XtIsRealized(wa) || !XtIsRealized(wb))
	return;
    
    XtVaGetValues(wa,
		  XtNwidth, &w1,
		  XtNheight, &h1,
		  NULL);
    XtVaGetValues(wb,
		  XtNwidth, &w2,
		  XtNheight, &h2,
		  XtNx, &x,
		  XtNy, &y,
		  NULL);
    XtVaSetValues(wa, XtNx, x + (w2 - w1) / 2,
		  XtNy, y + (h2 - h1) / 2,
		  NULL);
}

/* Position window w at coordinates x, y */
void
position_window(Widget w, Position x, Position y)
{
    if (!XtIsRealized(w))
	return;

    TRACE_GUI((stderr, "positioning %ld at %d, %d", (unsigned long)w, x, y));
    XtVaSetValues(w, XtNx, x, XtNy, y, NULL);
}

/*
  Used for the hyperref and forward search markers: scroll the page
  to make the marker visible.
*/
void
scroll_page_if_needed(int x_min, int x_max, int y_min, int y_max)
{
    Position drawing_x, drawing_y, drawing_h, clip_x, clip_y, clip_h, clip_w;
    int test_scroll, need_v_scroll = 0, need_h_scroll = 0;

    if (!do_autoscroll)
	return;
    
    XtVaGetValues(clip_widget,
		  XtNx, &clip_x, XtNy, &clip_y,
		  XtNheight, &clip_h, XtNwidth, &clip_w,
		  NULL);
    XtVaGetValues(draw_widget,
		  XtNx, &drawing_x, XtNy, &drawing_y,
		  XtNheight, &drawing_h,
		  NULL);

#if DEBUG_SCROLL_IF_NEEDED
    fprintf(stderr, "y: %d, drawing_y: %d, clip_h: %d\n", y, drawing_y, clip_h);
#endif
    /* check if we need to scroll vertically; first, down for y_min */
    test_scroll = y_min + drawing_y - clip_h;
    if ((expert_flag & XPRT_SHOW_STATUSLINE) != 0)
	test_scroll += global_statusline_h;

    TRACE_SRC((stderr, "test_scroll vertically: %d", test_scroll));
#if DEBUG_SCROLL_IF_NEEDED
    fprintf(stderr, "%d + %d > %d?\n", drawing_y, y_min, clip_h);
#endif
    if (test_scroll > 0) { /* need to scroll down? */
	need_v_scroll = test_scroll;
	TRACE_SRC((stderr, "need_v_scroll down: %d", need_v_scroll));
    }
    else if (abs(drawing_y) + 2 > y_max) { /* need to scroll up? */
	need_v_scroll = -((abs(drawing_y) - y_max) + 1);
	TRACE_SRC((stderr, "need_v_scroll up: %d (%d > %d; %d)", need_v_scroll, abs(drawing_y), y_max, clip_y));
    }

    /* check if we need to scroll horizontally; x_min < 0 blocks this
       (e.g. for hyperref, where we don't want it) */
    if (x_min >= 0) {
	test_scroll = x_min + drawing_x - clip_w + 1;
	TRACE_SRC((stderr, "test_scroll horizontally: %d", test_scroll));
	
	if (test_scroll > 0) {
	    /* need to scroll to right (i.e. make stuff on right-hand side visible)? */
	    need_h_scroll = test_scroll;
	    TRACE_SRC((stderr, "need_h_scroll right: %d", need_h_scroll));
	}
	else if (abs(drawing_x) > x_max) {
	    /* need to scroll to left (i.e. make stuff on left-hand side visible)? */
	    need_h_scroll =  -(abs(drawing_x) - x_max);
	    TRACE_SRC((stderr, "need_h_scroll left: %d", need_h_scroll));
	}
    }

    /* FIXME: should we not scroll if keep_flag is active? */
    if (need_v_scroll != 0 && y_bar != NULL) {
#ifdef MOTIF
	/* need to add new value to current one */
	XtVaGetValues(y_bar, XmNvalue, &test_scroll, NULL);
	(void)set_bar_value(y_bar, test_scroll + need_v_scroll, (int)(page_h - mane.height));
#else
	XtCallCallbacks(y_bar, XtNscrollProc, (XtPointer)need_v_scroll);
#endif
    }

    if (need_h_scroll != 0 && x_bar != NULL) {
#ifdef MOTIF
	/* need to add new value to current one */
	XtVaGetValues(x_bar, XmNvalue, &test_scroll, NULL);
	(void)set_bar_value(x_bar, test_scroll + need_h_scroll, (int)(page_w - mane.width));
#else
	XtCallCallbacks(x_bar, XtNscrollProc, (XtPointer)need_h_scroll);
#endif
    }

    do_autoscroll = False;
}

void adjust_width(Widget a, Widget b)
{
    Dimension w1, w2;
    
    XtVaGetValues(a, XtNwidth, &w1, NULL);
    XtVaGetValues(b, XtNwidth, &w2, NULL);
    
    if (w1 < w2)
	XtVaSetValues(a, XtNwidth, w2, NULL);
    else if (w2 > w1)
	XtVaSetValues(b, XtNwidth, w1, NULL);
}


GC make_gc(int function, Pixel fg, Pixel bg)
{
    XGCValues values;
    values.function = function;
    values.foreground = fg;
    values.background = bg;
    return XCreateGC(DISP, XtWindow(top_level), GCFunction | GCForeground | GCBackground, &values);
}

/*
 *	Atom handling.
 */

static char *atom_names[] = { "XDVI_WINDOWS", "DVI_NAME", "SRC_GOTO", "NEWDOC", "NEWPAGE", "RAISE" };
static Atom atoms[XtNumber(atom_names)];

Window
get_window_id(unsigned char *window_p)
{
    Window w;
    
#ifndef WORD64
    w = (*((xuint32 *) window_p));
#else
# if WORDS_BIGENDIAN
    w = ((unsigned long)wp[0] << 24) | ((unsigned long)wp[1] << 16)
	| ((unsigned long)wp[2] << 8) | (unsigned long)wp[3];
# else
    w = ((unsigned long)wp[3] << 24) | ((unsigned long)wp[2] << 16)
	| ((unsigned long)wp[1] << 8) | (unsigned long)wp[0];
# endif
#endif
    return w;
}

size_t
property_get_data(Window w, Atom a, unsigned char **ret_buf,
		  int (*x_get_property)(Display *, Window, Atom, long,
					long, Bool, Atom, Atom *, int *, unsigned long *,
					unsigned long *, unsigned char **))
{
    /* all of these are in 8-bit units */
    unsigned long byte_offset = 0;
    Atom type_ret;
    int format_ret;
    unsigned long nitems_ret;
    unsigned long bytes_after_ret = 0;
    unsigned char *prop_ret = NULL;
    
    /*
     * buffer for collecting returned data; this is static to
     * avoid expensive malloc()s at every call (which is often!)
     */
    static unsigned char *buffer = NULL;
    static size_t buffer_len = 0;

    while (x_get_property(DISP, w,
			  a, byte_offset / 4, (bytes_after_ret + 3) / 4, False,
			  a, &type_ret, &format_ret, &nitems_ret,
			  &bytes_after_ret, &prop_ret)
	   == Success) {

	if (type_ret != a || format_ret == 0)
	    break;

	nitems_ret *= (format_ret / 8);	/* convert to bytes */

	while ((byte_offset + nitems_ret) >= buffer_len) {
	    buffer_len += 256;
	    buffer = xrealloc(buffer, buffer_len);
	}

	/* the +1 captures the extra '\0' that Xlib puts after the end.  */
	memcpy(buffer + byte_offset, prop_ret, nitems_ret + 1);
	byte_offset += nitems_ret;

	XFree(prop_ret);
	prop_ret = NULL;

	if (bytes_after_ret == 0)	/* got all data */
	    break;
    }

    if (prop_ret != NULL)
	XFree(prop_ret);

    *ret_buf = buffer;
    return byte_offset;
}

size_t
property_get_window_list(unsigned char **window_list)
{
    size_t len = property_get_data(DefaultRootWindow(DISP),
				   atom_xdvi_windows(), window_list,
				   XGetWindowProperty);
    if (len == 0) {
	TRACE_CLIENT((stderr, "No \"xdvi windows\" property found"));
	return 0;
    }
    
    if (len % 4 != 0) {
	TRACE_CLIENT((stderr, "\"XDVI_WINDOWS\" property had incorrect size; deleting it."));
	XDeleteProperty(DISP, DefaultRootWindow(DISP), atom_xdvi_windows());
	return 0;
    }
    return len;
}

void
set_window_id(Window w, unsigned char *data)
{
#if WORDS_BIGENDIAN
    data[0] = (unsigned int)w >> 24;
    data[1] = (unsigned int)w >> 16;
    data[2] = (unsigned int)w >> 8;
    data[3] = (unsigned int)w;
#else
    data[0] = (unsigned int)w;
    data[1] = (unsigned int)w >> 8;
    data[2] = (unsigned int)w >> 16;
    data[3] = (unsigned int)w >> 24;
#endif
}

/**
 **	set_dvi_property sets the appropriate property for the main
 **	window (used in source special handoff).
 **/

void
set_dvi_property(void)
{
    XChangeProperty(DISP, XtWindow(top_level), atom_dvi_file(), atom_dvi_file(),
		    8, PropModeReplace, dvi_property, dvi_property_length);
}


/*
 * Delete all occurences of window w from the window list property. Then,
 * if `prepend' is true, prepend the window ID to the existing list.
 */
void
update_window_property(Window w, Boolean prepend)
{
    unsigned char *window_list;
    size_t window_list_len;
    unsigned char *window_list_end;
    unsigned char *wp;


    /* this allocates window_list */
    if ((window_list_len = property_get_window_list(&window_list)) == 0)
	return;

    /* Loop over list of windows.  */
    window_list_end = window_list + window_list_len;

#if 0
    if (prepend && window_list_len == 4) /* if ours is the only window */
	return;
#endif /* 0 */
    
    for (wp = window_list; wp < window_list_end; wp += 4) {
	if (get_window_id(wp) == w) { /* match, remove our ID */
	    window_list_len -= 4;
	    window_list_end -= 4;
	    memmove(wp, wp + 4, window_list_end - wp);

	    if (prepend) { /* add our ID again to front */
#ifdef WORD64
		unsigned char data[4];
		set_window_id(w, data);
#else
		xuint32 data = w;
#endif
/* 		if (debug & DBG_CLIENT) { */
/* 		    unsigned char *wp; */
/* 		    int i; */
/* 		    fprintf(stderr, "======== BEFORE replacing:\n"); */
/* 		    for (i = 0, wp = window_list; wp < window_list_end; wp += 4, i++) { */
/* 			Window w = get_window_id(wp); */
/* 			fprintf(stderr, "Window %d: %08lx\n", i, w); */
/* 		    } */
/* 		} */
		/* Note: no need to realloc window_list, since the original length
		   was sufficient for all elements.
		*/
		window_list_len += 4;
		memmove(window_list + 4, window_list, window_list_len - 4);
		memcpy(window_list, &data, 4);
/* 		if (debug & DBG_CLIENT) { */
/* 		    unsigned char *wp; */
/* 		    int i; */
/* 		    fprintf(stderr, "======== after replacing:\n"); */
/* 		    for (i = 0, wp = window_list; wp < window_list_end; wp += 4, i++) { */
/* 			Window w = get_window_id(wp); */
/* 			fprintf(stderr, "Window %d: %08lx\n", i, w); */
/* 		    } */
/* 		} */
	    }
	    
	    if (window_list_len == 0)
		XDeleteProperty(DISP, DefaultRootWindow(DISP),
				atom_xdvi_windows());
	    else
		XChangeProperty(DISP, DefaultRootWindow(DISP),
				atom_xdvi_windows(), atom_xdvi_windows(), 32,
				PropModeReplace, (unsigned char *)window_list,
				window_list_len / 4);

	    XFlush(DISP);
	    return;
	}
    }
    TRACE_CLIENT((stderr, "Couldn't find window %08lx in list when exiting.", w));

    return;
}

void
property_initialize(void)
{
    if (
#if XlibSpecificationRelease >= 6
	!XInternAtoms(DISP, atom_names, XtNumber(atom_names), False, atoms)
#else
	(atoms[0] = XInternAtom(DISP, atom_names[0], False)) == None
	|| (atoms[1] = XInternAtom(DISP, atom_names[1], False)) == None
	|| (atoms[2] = XInternAtom(DISP, atom_names[2], False)) == None
	|| (atoms[3] = XInternAtom(DISP, atom_names[3], False)) == None
	|| (atoms[4] = XInternAtom(DISP, atom_names[4], False)) == None
	|| (atoms[5] = XInternAtom(DISP, atom_names[5], False)) == None
#endif
	) {
	XDVI_FATAL((stderr, "XtInternAtoms failed."));
    }

    if (debug & DBG_CLIENT) {
	Cardinal k;
	for (k = 0; k < XtNumber(atom_names); ++k)
	    TRACE_CLIENT((stderr, "Atom(%s) = %lu", atom_names[k], atoms[k]));
    }
}


Atom atom_xdvi_windows(void)
{
    return atoms[0];
}

Atom atom_dvi_file(void)
{
    return atoms[1];
}

Atom atom_src_goto(void)
{
    return atoms[2];
}

Atom atom_newdoc(void)
{
    return atoms[3];
}

Atom atom_newpage(void)
{
    return atoms[4];
}

Atom atom_raise(void)
{
    return atoms[5];
}

/*
 * Syntetisize a button press object for the `button' passed as
 * the second argument.
 */
void
syntetisize_event(XEvent *ev, Widget button)
{
    memset(ev, 0, sizeof(XButtonPressedEvent));
    ev->type = ButtonPress;
    ev->xbutton.serial = 1;
    ev->xbutton.send_event = True;
    ev->xbutton.button = 1;
    ev->xbutton.display = XtDisplayOfObject(button);
    ev->xbutton.window = XtWindowOfObject(button);
}

#ifdef MOTIF
int xm_get_height(Widget w)
{
    int ret_h = 0;
    static Arg args = { XtNheight, (XtArgVal)0 };
    ASSERT(w != NULL, "widget in xm_get_width mustn't be NULL!");

    args.value = (XtArgVal)&ret_h;
    XtGetValues(w, &args, 1);
    TRACE_GUI((stderr, "xm_get_height: %d", ret_h));
    return ret_h;
}

int xm_get_width(Widget w)
{
    int ret_w = 0;
    static Arg args = { XtNwidth, (XtArgVal)0 };
    ASSERT(w != NULL, "widget in xm_get_width mustn't be NULL!");

    args.value = (XtArgVal)&ret_w;
    XtGetValues(w, &args, 1);
    TRACE_GUI((stderr, "xm_get_width: %d", ret_w));
    return ret_w;
}
#endif /* MOTIF */
