/*
 * Copyright (c) 2001 Marcin Dalecki
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * PAUL VOJTA OR ANY OTHER AUTHOR OF THIS SOFTWARE BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 */

/*
 * Implementation of the File selection dialogue for the Motif widget set
 * based GUI.
 */

#include "xdvi-config.h"
#include "xdvi.h"
#include "sfSelFile.h"

#include "dvi.h"
#include "string-utils.h"
#include "util.h"

#if defined(SELFILE) && defined(MOTIF) /* entire file */

#include <Xm/FileSB.h>

#include "xdvi.h"

static Widget dialog = NULL;
static char *browse_fname = NULL;

/*
 * Process callback from Dialog cancel actions.
 */
static void
cancel_callback(Widget w,
	XtPointer client_data,
	XtPointer call_data)
{
    UNUSED(w);
    UNUSED(client_data);
    UNUSED(call_data);
    /* don't reset browse_fname so that user gets the current
       value as default next time when he cancels now */
    /*
    if (browse_fname != NULL)
    {
	XtFree(browse_fname);
	browse_fname = NULL;
    }
    */
    XtUnmanageChild(dialog);
}

/*
 * Process callback from Dialog actions.
 */

static void
accept_callback(Widget w,
	XtPointer client_data,
	XtPointer call_data)
{
    XmFileSelectionBoxCallbackStruct *fcb;

    UNUSED(w);
    UNUSED(client_data);

    if (browse_fname != NULL)
    {
	XtFree(browse_fname);
	browse_fname = NULL;
    }
    fcb = (XmFileSelectionBoxCallbackStruct *)call_data;

    /* get the filename from the file selection box */
    XmStringGetLtoR(fcb->value, g_charset, &browse_fname);

    /* popdown the file selection box */
    XtUnmanageChild(dialog);
}

char *
XsraSelFile(Widget toplevel,
	    const char *title, const char *prompt,
	    const char *ok, const char *cancel,
	    const char *failed,
	    const char *init_path,
	    const char *mask,
	    Boolean must_exist)
{
#define ARG_CNT 4    
    XmString filemask = NULL;
    XmString directory = NULL;
    Arg args[ARG_CNT];
    int i = 0;
    
    UNUSED(prompt);
    UNUSED(ok);
    UNUSED(cancel);
    UNUSED(failed);
    UNUSED(init_path);
    UNUSED(must_exist);

    dialog = XmCreateFileSelectionDialog(toplevel,
					 "file", NULL, 0);
    XtVaSetValues(XtParent(dialog), XmNtitle, title, NULL);

    XtAddCallback(dialog, XmNokCallback, accept_callback, (XtPointer)0);
    XtAddCallback(dialog, XmNcancelCallback, cancel_callback, (XtPointer)0);
    /* We have no help in this window, so hide help button */
    XtUnmanageChild(XmFileSelectionBoxGetChild(dialog,
					       (unsigned char)XmDIALOG_HELP_BUTTON));

    /* only show files matching our mask */
    filemask = XmStringCreateLtoR((char *)mask, g_charset);
    XtSetArg(args[i], XmNpattern, filemask); i++;

    /* set directory to last directory used */
    if (browse_fname != NULL) {
	char *path = xstrdup(browse_fname);
	char *ptr = strrchr(path, '/');
	if (ptr != NULL)
	    *ptr = '\0';
	directory = XmStringCreateLtoR(path, g_charset);
	XtSetArg(args[i], XmNdirectory, directory); i++;
	free(path);
    }

    ASSERT(i < ARG_CNT, "args list too short");
    XtSetValues(dialog, args, (Cardinal)i);

    free(filemask);
    free(directory);
    
    XtManageChild(dialog);

    /* sit in a loop until the dialog box has gone away */
    do {
	XtAppProcessEvent(XtWidgetToApplicationContext(dialog),
	    (XtInputMask)XtIMAll);
    } while (XtIsManaged(dialog));

    XtDestroyWidget(dialog);

    if (browse_fname == NULL)
	return NULL;

    return xstrdup(browse_fname);
#undef ARG_CNT
}

#else
/* silence `empty compilation unit' warnings */
static void bar(void); static void foo() { bar(); } static void bar(void) { foo(); }
#endif /* defined(SELFILE) && defined(MOTIF) */
