/*
 * Copyright (c) 2001 Marcin Dalecki
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * PAUL VOJTA OR ANY OTHER AUTHOR OF THIS SOFTWARE BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 */

#include "xdvi-config.h"
#include "xdvi.h"

#include "xm_menu.h"
#include "xm_toolbar.h"
#include "util.h"
#include "version.h"
#include "statusline.h"
#include "c-openmx.h"

#ifdef MOTIF /* needed for `make depend' */

#include <Xm/RowColumn.h>
#include <Xm/Frame.h>
#include <Xm/PushB.h>
#include <Xm/PushBG.h>
#include <Xm/CascadeBG.h>
#include <Xm/SeparatoG.h>
#include <Xm/ToggleB.h>
#include <Xm/ToggleBG.h>
#include <Xm/Separator.h>


/*
 * There's an apparent bug in Motif related to the interaction between
 * the menubar menus and the magnifier.
 *
 * If you click on a menu on the menubar and then click on the drawing
 * widget to pop up a magnifier, the keyboard and pointer are still
 * grabbed, leading to a weird situation in which the magnifier stays
 * around even after you release the pointer.  The ungrab_serial
 * counter works around this bug by ignoring such pointer events.
 *
 * This bug occurs in Motif 1.2 and OpenMotif 2.2.2 (at least).  It
 * does not occur in Lesstif.
 */
static unsigned int ungrab_serial = 0;

static void
popdown_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
    UNUSED(w);
    UNUSED(client_data);
    /* Lesstif gives call_data == NULL */
    if (call_data != NULL && *((XtGrabKind *) call_data) != XtGrabNone)
	ungrab_serial = LastKnownRequestProcessed(DISP);
}

unsigned int
get_last_ungrab(void) {
    return ungrab_serial;
}


void
toggle_menubar(void)
{
    if ((expert_flag & XPRT_SHOW_MENUBAR) == 0)
	XtUnmanageChild(menu_bar);
    else
	XtManageChild(menu_bar);
}    

Widget
xm_create_menu(Widget parent, char *title, char mnemonic, struct button_info *item)
{
    Widget menu = 0, cascade = 0, w = 0;
    size_t i;
    XmString str;

    menu = XmCreatePulldownMenu(parent, "_pulldown", NULL, 0);
    str = XmStringCreateLocalized(title);
    cascade =  XtVaCreateManagedWidget(title, xmCascadeButtonGadgetClass, parent,
 				       XmNsubMenuId, menu,
 				       XmNlabelString, str,
 				       XmNmnemonic, mnemonic,
 				       NULL);
    XmStringFree(str);
    
    /* add the menu items */
    for (i = 0; item != NULL && i < item->size; i++) {
	item->elems[i].widget = 0;
	/* if there's a subitem, call this function recursively */
	if (item->elems[i].submenu != NULL) {
	    w = xm_create_menu(menu, item->elems[i].title,
			       item->elems[i].mnemonic, item->elems[i].submenu);
	    /*
	      workaround for pointer grabbing bug; add additional callback to all menus
	      (they have the same parent)
	    */
	    XtAddCallback(XtParent(menu), XtNpopdownCallback,
			  popdown_callback, (XtPointer) 0);
	}
	else {
	    switch(item->elems[i].type) {
	    case BT_PUSH:
		w = XtVaCreateManagedWidget(item->elems[i].title, xmPushButtonGadgetClass, menu,
					    NULL);
		break;
	    case BT_RADIO:
		w = XtVaCreateManagedWidget(item->elems[i].title, xmToggleButtonGadgetClass, menu,
					    XmNindicatorType, XmONE_OF_MANY,
					    NULL);
		break;
	    case BT_CHECK:
		w = XtVaCreateManagedWidget(item->elems[i].title, xmToggleButtonGadgetClass, menu,
					    XmNindicatorType, XmN_OF_MANY,
					    NULL);
		break;
	    case BT_SEP:
		w = XtVaCreateManagedWidget(item->elems[i].title, xmSeparatorGadgetClass, menu,
					    NULL);
		break;
	    default:
		XDVI_WARNING((stderr, "unrecognized button type %d in menu %s (skipping this item).\n",
			      item->elems[i].type, item->elems[i].title));
		break;
	    }
	    item->elems[i].widget = w;
	}

	if (item->elems[i].mnemonic != 0)
	    XtVaSetValues(w, XmNmnemonic, item->elems[i].mnemonic, NULL);
	
	if (item->elems[i].accelerator != NULL) {
	    str = XmStringCreateLocalized(item->elems[i].accelerator);
	    XtVaSetValues(w, XmNacceleratorText, str, NULL);
	    XmStringFree(str);
	}
	
	if (item->elems[i].action != NULL) {
	    String cb_type;
	    if (XmIsToggleButton(w) || XmIsToggleButtonGadget(w))
		cb_type = XmNvalueChangedCallback;
	    else
		cb_type = XmNactivateCallback;
	    XtAddCallback(w, cb_type, handle_command, item->elems[i].action);
	}
    }
    return cascade;
}

#else
/* silence `empty compilation unit' warnings */
static void bar(void); static void foo() { bar(); } static void bar(void) { foo(); }
#endif /* MOTIF */

